/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.build;

import com.google.common.annotations.VisibleForTesting;
import com.google.common.base.Predicate;
import com.google.common.collect.Lists;
import com.google.common.collect.Multimap;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import lombok.val;
import org.apache.commons.io.FileUtils;
import org.apache.commons.io.IOUtils;
import org.apache.commons.io.filefilter.AbstractFileFilter;
import org.apache.commons.io.filefilter.DirectoryFileFilter;
import org.apache.commons.io.filefilter.IOFileFilter;
import org.apache.commons.io.output.NullOutputStream;
import org.apache.commons.lang.StringUtils;
import org.apache.http.HttpHeaders;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.build.BuildAddon;
import org.artifactory.addon.build.artifacts.ProducedBy;
import org.artifactory.addon.build.artifacts.UsedBy;
import org.artifactory.addon.ha.HaCommonAddon;
import org.artifactory.addon.plugin.PluginAction;
import org.artifactory.addon.plugin.PluginsAddon;
import org.artifactory.addon.plugin.build.AfterBuildSaveAction;
import org.artifactory.addon.plugin.build.BeforeBuildSaveAction;
import org.artifactory.addon.retention.ArtifactRetentionAddon;
import org.artifactory.addon.xray.XrayAddon;
import org.artifactory.api.build.BuildProps;
import org.artifactory.api.build.BuildRetentionWorkItem;
import org.artifactory.api.build.BuildRunComparators;
import org.artifactory.api.build.BuildWrapper;
import org.artifactory.api.build.GeneralBuild;
import org.artifactory.api.build.ImportableExportableBuild;
import org.artifactory.api.build.ModuleArtifact;
import org.artifactory.api.build.ModuleDependency;
import org.artifactory.api.build.PublishedModule;
import org.artifactory.api.build.model.BuildGeneralInfo;
import org.artifactory.api.build.model.diff.BuildParams;
import org.artifactory.api.build.request.BuildArtifactoryRequest;
import org.artifactory.api.common.BasicStatusHolder;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.api.repo.exception.RepoRejectException;
import org.artifactory.api.request.ArtifactoryResponse;
import org.artifactory.api.request.UploadService;
import org.artifactory.api.rest.artifact.PromotionResult;
import org.artifactory.api.rest.build.ContinueBuildFilter;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.api.security.SecurityService;
import org.artifactory.aql.AqlService;
import org.artifactory.aql.api.domain.sensitive.AqlApiBuild;
import org.artifactory.aql.api.domain.sensitive.AqlApiItem;
import org.artifactory.aql.api.internal.AqlBase;
import org.artifactory.aql.model.AqlComparatorEnum;
import org.artifactory.aql.result.AqlEagerResult;
import org.artifactory.aql.result.AqlLazyResult;
import org.artifactory.aql.result.rows.AqlBaseFullRowImpl;
import org.artifactory.aql.result.rows.AqlBuild;
import org.artifactory.checksum.ChecksumType;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.MutableStatusHolder;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.config.CentralConfigKey;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.descriptor.repo.RepoBaseDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.factory.JaxbAdapter;
import org.artifactory.factory.jaxb.JaxbAdapterFactory;
import org.artifactory.fs.FileInfo;
import org.artifactory.io.NullResourceStreamHandle;
import org.artifactory.model.common.RepoPathImpl;
import org.artifactory.projects.ProjectsService;
import org.artifactory.projects.ProjectsUtils;
import org.artifactory.repo.ReloadableRepositoryConfigService;
import org.artifactory.repo.RepoConfigReload;
import org.artifactory.repo.RepoDetailsType;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.RepoPathFactory;
import org.artifactory.repo.service.InternalRepositoryService;
import org.artifactory.repo.service.deploy.ArtifactoryDeployRequest;
import org.artifactory.repo.service.deploy.ArtifactoryDeployRequestBuilder;
import org.artifactory.request.InternalCapturingResponse;
import org.artifactory.resource.ResourceStreamHandle;
import org.artifactory.rest.exception.BadRequestException;
import org.artifactory.rest.exception.ConflictException;
import org.artifactory.rest.exception.ForbiddenException;
import org.artifactory.rest.exception.NotFoundException;
import org.artifactory.sapi.common.ExportSettings;
import org.artifactory.sapi.common.ImportKind;
import org.artifactory.sapi.common.ImportSettings;
import org.artifactory.sapi.interceptor.context.DeleteContext;
import org.artifactory.schedule.Task;
import org.artifactory.schedule.TaskService;
import org.artifactory.schedule.TaskUtils;
import org.artifactory.security.AccessLogger;
import org.artifactory.security.SystemAuthenticationToken;
import org.artifactory.spring.Reloadable;
import org.artifactory.storage.StorageException;
import org.artifactory.storage.db.InternalDbService;
import org.artifactory.storage.db.build.aql.BuildsAql;
import org.artifactory.storage.db.build.entity.BuildEntity;
import org.artifactory.storage.db.build.service.BuildIdImpl;
import org.artifactory.storage.db.build.service.BuildRunImpl;
import org.artifactory.storage.db.build.service.InternalBuildStoreService;
import org.artifactory.storage.db.fs.dao.NodesDao;
import org.artifactory.storage.db.fs.entity.Node;
import org.artifactory.storage.db.util.JdbcHelper;
import org.artifactory.storage.jobs.migration.buildinfo.BuildInfoCalculationFatalException;
import org.artifactory.storage.jobs.migration.buildinfo.BuildInfoCalculationWorkItem;
import org.artifactory.storage.jobs.migration.buildinfo.BuildInfoMigrationJob;
import org.artifactory.storage.jobs.migration.buildinfo.BuildInfoMigrationJobDelegate;
import org.artifactory.ui.rest.service.builds.search.BuildsSearchFilter;
import org.artifactory.util.CollectionUtils;
import org.artifactory.util.date.DateUtils;
import org.artifactory.version.CompoundVersionDetails;
import org.iostreams.streams.in.StringInputStream;
import org.jetbrains.annotations.NotNull;
import org.jfrog.access.client.project.ProjectSlimEntity;
import org.jfrog.build.api.Artifact;
import org.jfrog.build.api.Build;
import org.jfrog.build.api.BuildRetention;
import org.jfrog.build.api.Dependency;
import org.jfrog.build.api.Issue;
import org.jfrog.build.api.Issues;
import org.jfrog.build.api.release.Promotion;
import org.jfrog.build.api.release.PromotionStatus;
import org.jfrog.common.JsonUtils;
import org.jfrog.common.StreamSupportUtils;
import org.jfrog.common.concurrency.ConflictGuard;
import org.jfrog.common.concurrency.ConflictsGuard;
import org.jfrog.common.config.diff.DataDiff;
import org.jfrog.storage.binstore.common.ChecksumInputStream;
import org.jfrog.storage.binstore.exceptions.BinaryRejectedException;
import org.jfrog.storage.binstore.utils.Checksum;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.stereotype.Service;

import javax.annotation.Nullable;
import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.Response;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.sql.SQLException;
import java.util.Collection;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.TimeUnit;
import java.util.function.Function;
import java.util.function.Supplier;
import java.util.stream.Collector;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static com.google.common.collect.Iterables.filter;
import static com.google.common.collect.Iterables.getLast;
import static com.google.common.collect.Lists.newArrayList;
import static java.lang.String.format;
import static java.nio.charset.StandardCharsets.UTF_8;
import static org.apache.commons.lang.StringUtils.isNotBlank;
import static org.apache.commons.lang.StringUtils.isNotEmpty;
import static org.apache.commons.lang.StringUtils.startsWith;
import static org.apache.commons.lang3.math.NumberUtils.min;
import static org.apache.http.HttpStatus.SC_BAD_REQUEST;
import static org.apache.http.HttpStatus.SC_FORBIDDEN;
import static org.artifactory.addon.build.BuildAddon.BUILD_INFO_REPO_EXTENSION;
import static org.artifactory.addon.build.BuildAddon.PROP_BUILD_NAME;
import static org.artifactory.addon.build.BuildAddon.PROP_BUILD_NUMBER;
import static org.artifactory.addon.build.BuildAddon.PROP_BUILD_TIMESTAMP;
import static org.artifactory.aql.api.domain.sensitive.AqlApiBuild.module;
import static org.artifactory.aql.api.internal.AqlBase.and;
import static org.artifactory.build.BuildInfoUtils.formatBuildTime;
import static org.artifactory.build.BuildServiceUtils.VerifierLogLevel;
import static org.artifactory.build.BuildServiceUtils.addIncludeExcludeRepos;
import static org.artifactory.build.BuildServiceUtils.getBuildAsJsonString;
import static org.artifactory.build.BuildServiceUtils.isBuildModelInvalid;
import static org.artifactory.build.BuildServiceUtils.iterateFetchFromDb;
import static org.artifactory.build.BuildServiceUtils.matchArtifactsToFileInfos;
import static org.artifactory.build.BuildServiceUtils.matchDependenciesToFileInfos;
import static org.artifactory.build.BuildServiceUtils.matchUnmatchedArtifactsNonStrict;
import static org.artifactory.build.BuildServiceUtils.prepareBuildsFolder;
import static org.artifactory.repo.system.SystemRepos.BUILDINFO_REPO_KEY;
import static org.artifactory.storage.db.build.service.BuildStoreServiceImpl.formatDateToString;
import static org.artifactory.storage.jobs.migration.buildinfo.BuildInfoCalculationWorkItem.BUILD_INFO_CALCULATION_KEY_PREFIX;
import static org.jfrog.storage.util.DbUtils.foreignKeyExists;
import static org.jfrog.storage.util.DbUtils.tableExists;

/**
 * Build service main implementation
 *
 * @author Noam Y. Tenne
 */
@Slf4j
@Service
@Reloadable(beanClass = InternalBuildService.class, initAfter = {InternalRepositoryService.class})
public class BuildServiceImpl implements InternalBuildService {

    private static final String EXPORTABLE_BUILD_VERSION = "v2";
    private static final String BUILD_JSONS_TABLE = "build_jsons";

    //Signifies the state of the v6.6.0 build info migration[
    boolean buildInfoReady = false;
    private Map<String, String> projectToBuildRepoMap;
    private Set<String> allBuildsRepositories;
    private Builds builds;
    private AddonsManager addonsManager;
    private AuthorizationService authService;
    private InternalRepositoryService repoService;
    private ReloadableRepositoryConfigService repoConfigService;
    private InternalBuildStoreService buildStoreService;
    private InternalDbService dbService;
    private AqlService aqlService;
    private UploadService uploadService;
    private TaskService taskService;
    private JdbcHelper jdbcHelper;
    private JaxbAdapter buildJaxbAdapter;
    private BuildsAql buildsAql;
    private ProjectsService projectsService;
    private NodesDao nodesDao;
    private final JsonUtils jsonUtils = JsonUtils.getInstance();

    @Autowired
    public void setAddonsManager(AddonsManager addonsManager) {
        this.addonsManager = addonsManager;
    }

    @Autowired
    public void setDbService(InternalDbService dbService) {
        this.dbService = dbService;
    }

    @Autowired
    public void setAqlService(AqlService aqlService) {
        this.aqlService = aqlService;
    }

    @Autowired
    public void setUploadService(UploadService uploadService) {
        this.uploadService = uploadService;
    }

    @Autowired
    public void setAuthService(AuthorizationService authService) {
        this.authService = authService;
    }

    @Autowired(required = false)
    public void setBuilds(Builds builds) {
        this.builds = builds;
    }

    @Autowired
    public void setRepoService(InternalRepositoryService repoService) {
        this.repoService = repoService;
    }

    @Autowired
    public void setRepoConfigService(ReloadableRepositoryConfigService repoConfigService) {
        this.repoConfigService = repoConfigService;
    }

    @Autowired
    public void setBuildStoreService(InternalBuildStoreService buildStoreService) {
        this.buildStoreService = buildStoreService;
    }

    @Autowired
    public void setTaskService(TaskService taskService) {
        this.taskService = taskService;
    }

    @Autowired
    public void setJdbcHelper(JdbcHelper jdbcHelper) {
        this.jdbcHelper = jdbcHelper;
    }

    @Autowired
    public void setBuildsAql(BuildsAql buildsAql) {
        this.buildsAql = buildsAql;
    }

    @Autowired
    public void setProjectsService(ProjectsService projectsService) {
        this.projectsService = projectsService;
    }

    @Autowired
    public void setNodesDao(NodesDao nodesDao) {
        this.nodesDao = nodesDao;
    }

    @Override
    public void init() {
        buildJaxbAdapter = JaxbAdapterFactory.create(ImportableExportableBuild.class);
        verifyBuildInfoState();
        projectToBuildRepoMap = new ConcurrentHashMap<>();
        refreshAllBuildsRepositories();
    }

    @Override
    public void reload(CentralConfigDescriptor oldDescriptor, List<DataDiff<?>> configDiff) {
        buildJaxbAdapter = JaxbAdapterFactory.create(ImportableExportableBuild.class);
        convertIfNeeded(configDiff);
        verifyBuildInfoState();
        projectToBuildRepoMap.clear();
    }

    @Override
    public void onRepoConfigReload(Map<RepoDetailsType, List<RepoConfigReload>> repoConfigReload) {
        refreshAllBuildsRepositories();
    }

    @Override
    public void handleBuildUploadRedirect(BuildArtifactoryRequest request, ArtifactoryResponse response,
            @NonNull String targetRepoKey) throws IOException {
        try {
            modifyRequestForProperDeployment(request, RepoPathFactory.create(targetRepoKey));
            uploadService.upload(request, response);
        } catch (IOException e) {
            log.error("", e);
            if (!response.isCommitted()) {
                response.sendError(SC_BAD_REQUEST, e.getMessage(), log);
            } else {
                throw new BadRequestException(e.getMessage());
            }
        } catch (RepoRejectException | BinaryRejectedException rej) {
            log.error("Deployment of build info failed: {}", rej.getMessage());
            log.debug("Deployment of build info failed", rej);
            if (!response.isCommitted()) {
                response.sendError(rej.getErrorCode(), rej.getMessage(), log);
            } else {
                throw new ConflictException(rej.getMessage());
            }
        }
    }

    /**
     * This flow is mildly different from the {@link org.artifactory.api.build.BuildService#addBuild} method since we
     * need to preserve the
     * headers that came in with the {@param request} (and other meta-fields that the {@link UploadService} cares
     * about).
     */
    private void modifyRequestForProperDeployment(BuildArtifactoryRequest request, RepoPath deploymentPath) {
        if (isReplicationRequest(request)) {
            String path = request.getRepoPath().getPath();
            // RepoFilter changes the request in case of replication we should handle the original request.
            String originalRequest = StringUtils.substringAfter(path, "build/buildUploadRedirect/");
            RepoPath repoPath = RepoPathFactory.create(originalRequest);
            request.setRepoPath(repoPath);
        } else {
            Build build = getBuildModelFrom(request);
            preAddBuildSteps(build, deploymentPath.getRepoKey());
            request.replaceContent(JsonUtils.getInstance().valueToByteArray(build));
            RepoPath correctUploadPath =
                    BuildServiceUtils.generateBuildJsonRepoPath(build, deploymentPath.getRepoKey());
            if (!correctUploadPath.equals(request.getRepoPath())) {
                if (log.isDebugEnabled()) {
                    log.debug("Wrong path given for deployment of build {}:{}:{}, adjusting to path '{}'",
                            build.getName(), build.getNumber(), deploymentPath.getRepoKey(),
                            correctUploadPath.toPath());
                }
                request.setRepoPath(correctUploadPath);
            }
        }
    }

    private boolean isReplicationRequest(BuildArtifactoryRequest request) {
        Enumeration<String> headerValues = request.getHeaderValues(HttpHeaders.USER_AGENT);
        String userAgent = headerValues.hasMoreElements() ? headerValues.nextElement() : null;
        // In case the userAgent is Artifactory/ we can assume that the request is valid and not need to do modification
        return isNotBlank(userAgent) && startsWith(userAgent, "Artifactory/");
    }

    @VisibleForTesting
    void refreshAllBuildsRepositories() {
        allBuildsRepositories = repoConfigService.getRepoKeysByTypeAndPackageType(
                Set.of(RepoDetailsType.LOCAL, RepoDetailsType.FEDERATED), RepoType.BuildInfo);
    }

    @NonNull
    @Override
    public String addBuild(Build build, @NonNull String buildRepo) {
        validateFeatureAllowed();
        if (log.isDebugEnabled()) {
            log.debug("Adding build '{}' to the repo '{}'", build, buildRepo);
        }
        createBuildRepoIfNeeded(buildRepo);
        InternalCapturingResponse response = getTransactionalMe().persistBuildInTransaction(build, buildRepo);
        // Checking response here in order to return status to REST API deployment :(
        if (response.getStatus() == SC_FORBIDDEN) {
            throw new ForbiddenException(response.getStatusMessage());
        } else if (response.getStatus() >= SC_BAD_REQUEST) {
            Response errorResponse = Response.status(response.getStatus())
                    .entity(response.getStatusMessage()).build();
            throw new WebApplicationException(errorResponse);
        }
        return parseBuildSha256(response);
    }

    @Override
    @NotNull
    public InternalCapturingResponse persistBuildInTransaction(Build build, String buildRepo) {
        //These pre-steps modify the actual build so they must be done here.
        preAddBuildSteps(build, buildRepo);
        InternalCapturingResponse artifactoryResponse = new InternalCapturingResponse();
        deployBuildJson(build, artifactoryResponse, buildRepo);
        return artifactoryResponse;
    }

    @VisibleForTesting
    @NonNull
    String parseBuildSha256(@NonNull InternalCapturingResponse response) {
        String sha256 = "";
        try {
            String responseJson = response.getResultAsString();
            if (StringUtils.isNotBlank(responseJson)) {
                sha256 = jsonUtils.readTree(responseJson).path("originalChecksums").path("sha256").asText("");
            }
        } catch (Exception e) {
            log.error("Failed to parse SHA256 from internal artifactory response");
            log.debug("Exception:", e);
        }
        return sha256;
    }

    private void createBuildRepoIfNeeded(@NonNull String buildRepoKey) {
        if (getAllBuildInfoRepos().contains(buildRepoKey)) {
            return;
        }
        if (!buildRepoKey.endsWith(BuildAddon.BUILD_INFO_REPO_EXTENSION)) {
            log.error("Illegal Build info repo '{}' name. Repo name is missing '{}' suffix.", buildRepoKey,
                    BUILD_INFO_REPO_EXTENSION);
            throwRepoNotExistException(buildRepoKey);
        }
        String projectKey = ProjectsUtils.getProjectKeyFromRepoKey(buildRepoKey);
        if (projectKey == null) {
            log.error("Cannot parse project name out of '{}' repository name.", buildRepoKey);
            throwRepoNotExistException(buildRepoKey);
        }
        ProjectSlimEntity project = projectsService.getProjectSpec(projectKey);
        if (project == null) {
            log.error("Project '{}' was not found.", projectKey);
            throwRepoNotExistException(buildRepoKey);
        }

        addonsManager.addonByType(BuildAddon.class).createProjectRelatedBuildInfoRepo(project.getProjectKey());
    }

    @NonNull
    private void throwRepoNotExistException(@NonNull String buildRepoKey) {
        throw new IllegalArgumentException(String.format("Build info repository '%s' does not exist", buildRepoKey));
    }

    private void preAddBuildSteps(Build build, @NonNull String buildRepo) {
        build.setArtifactoryPrincipal(authService.currentUsername());
        buildStoreService.populateMissingChecksums(build);
        aggregatePreviousBuildIssues(build, buildRepo);
        notifyInterceptors(build, buildRepo, BeforeBuildSaveAction.class);
    }

    /**
     * Deploys the {@param build} as a json into the repo, thus triggering the
     * {@link org.artifactory.repo.interceptor.BuildInfoInterceptor})
     * which will in turn trigger {@link InternalBuildService#addBuildInternal} that manages the entire add build flow.
     */
    private void deployBuildJson(Build build, ArtifactoryResponse artifactoryResponse, String buildsRepoKey) {
        String buildJson = getBuildAsJsonString(build);
        RepoPath targetPath = BuildServiceUtils.generateBuildJsonRepoPath(build, buildsRepoKey);
        log.debug("Creating build json for '{}:{}' at '{}'", build.getName(), build.getNumber(), targetPath);
        try (StringInputStream buildInStream = new StringInputStream(buildJson, UTF_8)) {
            ArtifactoryDeployRequest deployRequest = createDeployRequest(buildJson.getBytes().length, targetPath,
                    buildInStream);
            uploadService.upload(deployRequest, artifactoryResponse);
        } catch (BinaryRejectedException | RepoRejectException rej) {
            log.debug("Rejected error occurred", rej);
            throw new ConflictException(rej.getMessage());
        } catch (Exception e) {
            log.error("Error during build json deploy", e);
            throw new BadRequestException(e.getMessage());
        }
    }

    private ArtifactoryDeployRequest createDeployRequest(long contentLength, RepoPath targetPath,
            StringInputStream buildAsStream) {
        return new ArtifactoryDeployRequestBuilder(targetPath)
                .inputStream(buildAsStream)
                .contentLength(contentLength)
                .trustServerChecksums(true)
                .build();
    }

    @Override
    public void addBuildInternal(final BuildWrapper build) {
        String buildName = build.getName();
        String buildNumber = build.getNumber();
        String buildStarted = build.getStarted();
        String buildRepo = build.getBuildRepo();
        // must have it here to verify the user has permission to the final destination (if move will be needed)
        if (!authService.canUploadBuild(buildName, buildNumber, buildStarted, buildRepo)) {
            AccessLogger.buildCreateDenied(buildName + ':' + buildNumber);
            throw new ForbiddenException(ACCESS_DENIED_MSG);
        }

        //Shouldn't really happen but users might reach this state via the plugins so why not save them?
        //Projects update: Or they are trying to deploy a build into 2 different repositories
        if (exists(build)) {
            String buildString = format("Build %s:%s-%s at %s", build, buildNumber, buildStarted, buildRepo);
            String oldBuildRepoKey = buildStoreService.getBuildStoringRepo(build.getName(), buildNumber, buildStarted);
            if (!buildRepo.equals(oldBuildRepoKey)) {
                throw new ConflictException(buildString + "already exists in repository " + oldBuildRepoKey + ".");
            }
            if (isBuildInfoChangedFromDb(build.getBuild(), buildRepo)) {
                throw new ConflictException(buildString + "already exists.");
            } else {
                if (log.isDebugEnabled()) {
                    log.debug(format("Trying to add the same build twice: %s", buildString), new RuntimeException());
                }
                return;
            }
        }
        log.debug("Adding info for build '{}:{}' date:{} stored in {}", buildName, buildNumber, buildStarted,
                buildRepo);
        buildStoreService.addBuild(build, buildInfoReady);
        log.debug("Added info for build '{}:{}' at {} repo", buildName, buildNumber, buildRepo);
        AccessLogger.buildCreate(build.getName() + ':' + build.getNumber());
        log.debug("Calling Xray build scan (if required) on build '{}:{}' at {} repo", buildName, buildNumber,
                buildRepo);
        addonsManager.addonByType(XrayAddon.class).callAddBuildInterceptor(build);
        BuildRetention buildRetention = build.getBuild().getBuildRetention();
        if (buildRetention != null) {
            log.debug("Executing build retention for build {} at {} repo", buildName, buildRepo);
            addonsManager.addonByType(BuildAddon.class)
                    .discardOldBuilds(build.getName(), buildRepo, buildRetention, false);
        }
        notifyInterceptors(build.getBuild(), buildRepo, AfterBuildSaveAction.class);
    }

    private String calcBuildJsonSha1(Build build) {
        Checksum sha1 = new Checksum(ChecksumType.sha1.alg(), ChecksumType.sha1.length());
        String buildAsJsonString = getBuildAsJsonString(build);
        if (StringUtils.isEmpty(buildAsJsonString)) {
            return "";
        }
        try (InputStream in = IOUtils.toInputStream(buildAsJsonString, UTF_8); ChecksumInputStream calcStream =
                new ChecksumInputStream(in, sha1)) {
            log.debug("Calculating sha1 for build {} {} {}", build.getName(), build.getNumber(), build.getStarted());
            // waste the stream to have it calculate the checksum
            IOUtils.copy(calcStream, NullOutputStream.NULL_OUTPUT_STREAM);
        } catch (Exception e) {
            String msg = "Can't calculate '" + sha1 + "' for build: " + build.getName() + " " + build.getNumber();
            log.error(msg, e);
            return "";
        }
        return sha1.getChecksum();
    }

    private boolean isBuildInfoChangedFromDb(Build build, String repo) {
        // In case the migration is done there is no need to do the validation
        if (buildInfoReady) {
            return true;
        }
        Build buildFromDb = buildStoreService
                .getBuild(new BuildRunImpl(build.getName(), build.getNumber(), build.getStarted(), repo));
        // Build missing equals so we will use the 2 string
        return !calcBuildJsonSha1(build).equals(calcBuildJsonSha1(buildFromDb));
    }

    /**
     * Check if the latest build available has issues, add them all the our newly created build
     * only if the previous build status is not in "Released" status (configurable status from our plugins).
     * This way we collect all previous issues related to the same version which is not released yet.
     *
     * @param newBuild the newly created build to add previous issues to
     */
    private void aggregatePreviousBuildIssues(Build newBuild, String buildRepo) {
        Issues newBuildIssues = newBuild.getIssues();
        if (newBuildIssues == null) {
            return;
        }

        if (!newBuildIssues.isAggregateBuildIssues()) {
            return;
        }

        BuildWrapper latestBuild = getLatestBuildByNameAndStatus(newBuild.getName(), buildRepo, LATEST_BUILD);
        if (latestBuild == null) {
            return;
        }

        // Only aggregate if the previous build does not equal to the user requested status (e.g: "Released")
        // this way we only aggregate the issues related to the current release
        List<PromotionStatus> statuses = latestBuild.getBuild().getStatuses();
        if (statuses != null) {
            String aggregationBuildStatus = newBuildIssues.getAggregationBuildStatus();
            for (PromotionStatus status : statuses) {
                if (status.getStatus().equalsIgnoreCase(aggregationBuildStatus)) {
                    return;
                }
            }
        }

        // It is important to create new Issue instance so we won't mess up previous ones
        Issues previousIssues = latestBuild.getBuild().getIssues();
        if (previousIssues != null) {
            Set<Issue> affectedIssues = previousIssues.getAffectedIssues();
            if (affectedIssues != null) {
                for (Issue issue : affectedIssues) {
                    Issue issueToAdd = new Issue(issue.getKey(), issue.getUrl(), issue.getSummary());
                    issueToAdd.setAggregated(true);
                    newBuildIssues.addIssue(issueToAdd);
                }
            }
        }
    }

    @Override
    public BuildRun getBuildRun(String buildName, String buildNumber, String buildStarted,
            @NonNull String buildRepo) {
        BuildRun buildRun = getBuildRunInternally(buildName, buildNumber, buildStarted);
        if (buildRun == null) {
            assertBasicReadPermissions(buildName, buildNumber, buildStarted, buildRepo);
        } else {
            assertBasicReadPermissions(buildName, buildNumber, buildStarted, buildRun.getBuildRepo());
        }
        return buildRun;
    }

    @Override
    public BuildRun getBuildRunInternally(String buildName, String buildNumber, String buildStarted) {
        return buildStoreService.getBuildRun(buildName, buildNumber, buildStarted);
    }

    @Override
    @Nullable
    public BuildWrapper getBuild(BuildRun buildRun) {
        BuildWrapper buildWrapper = getBuildInternally(buildRun);
        if (buildWrapper == null || !buildWrapper.isBuildExist()) {
            assertReadPermissions(buildRun.getName(), buildRun.getNumber(), buildRun.getStarted(),
                    getBuildInfoRepoForDefaultProject());
            return null;
        }
        assertReadPermissions(buildRun.getName(), buildRun.getNumber(), buildRun.getStarted(),
                buildWrapper.getBuildRepo());
        return buildWrapper;
    }

    @Override
    public BuildWrapper getBuildInternally(BuildRun buildRun) {
        if (isBuildModelInvalid(buildRun.getName(), buildRun.getNumber(), buildRun.getStarted())) {
            return null;
        }
        return getBuildAccordingToCoordinates(buildRun);
    }

    @Override
    public String getBuildAsJson(BuildRun buildRun) {
        if (isBuildModelInvalid(buildRun.getName(), buildRun.getNumber(), buildRun.getStarted())) {
            return null;
        }
        assertReadPermissions(buildRun.getName(), buildRun.getNumber(), buildRun.getStarted(), buildRun.getBuildRepo());
        BuildWrapper buildWrapper = getBuildAccordingToCoordinates(buildRun);
        if (buildWrapper != null) {
            return getBuildAsJsonString(buildWrapper.getBuild());
        }
        return null;
    }

    private BuildWrapper getBuildAccordingToCoordinates(BuildRun build) {
        if (buildInfoReady) {
            String buildRepo = build.getBuildRepo();
            if (StringUtils.isEmpty(buildRepo)) {
                throw new IllegalArgumentException("Build info file does not exist");
            }
            return getBuildModelFromFile(BuildServiceUtils
                    .generateBuildJsonRepoPath(build.getName(), build.getNumber(), build.getStarted(), buildRepo));
        } else {
            //Don't you love legacy considerations? because I sure do
            return new BuildWrapper(buildStoreService.getBuild(build), BUILDINFO_REPO_KEY);
        }
    }

    /**
     * Reads the build json from the repository and converts the stream into {@link Build model}
     *
     * @return Build model
     * null in case of an error
     */
    @SuppressWarnings("java:S2139") // logging and re-throwing exception.
    @Override
    public BuildWrapper getBuildModelFromFile(RepoPath buildJsonPath) {
        try (ResourceStreamHandle streamHandle = repoService.getResourceStreamHandle(buildJsonPath)) {
            if (streamHandle instanceof NullResourceStreamHandle) {
                throw new IllegalArgumentException("Build info file does not exist: " + buildJsonPath);
            }
            return new BuildWrapper(
                    getBuildAs(streamHandle::getInputStream, in -> JsonUtils.getInstance().readValue(in, Build.class)),
                    buildJsonPath.getRepoKey());
        } catch (IOException e) {
            log.error("Unable to retrieve build info json from '{}': {}", buildJsonPath, e.getMessage());
            log.debug("Unable to retrieve build info json from '{}'", buildJsonPath, e);
            throw new IllegalStateException("Could not get JSON string for build " + buildJsonPath, e);
        }
    }

    private Build getBuildModelFrom(BuildArtifactoryRequest request) {
        try (InputStream requestStream = request.getInputStream()) {
            return getBuildAs(() -> requestStream, in -> JsonUtils.getInstance().readValue(in, Build.class));
        } catch (Exception e) {
            String message = request.getRepoPath() + " could not be deployed. " +
                    "Build Info repositories only supports build-info.json files.";
            log.error(message, request.getRepoPath(), e);
            throw new BadRequestException(message);
        }
    }

    private <T> T getBuildAs(Supplier<InputStream> inputSupplier, Function<InputStream, T> whatToDo)
            throws IOException {
        try (InputStream in = inputSupplier.get()) {
            return whatToDo.apply(in);
        }
    }

    @Override
    public void deleteAllBuildsByName(List<BuildCoordinate> buildsCoordinates, boolean deleteArtifacts,
            BasicStatusHolder status) {
        for (BuildCoordinate buildCoordinates : buildsCoordinates) {
            if (!authService.canDeleteBuildName(buildCoordinates.getBuildName(), buildCoordinates.getBuildRepo())) {
                log.warn("Build name {} under {} was not found. Skip on deleting.", buildCoordinates.getBuildName(),
                        buildCoordinates.getBuildRepo());
                throw buildNotFoundException();
            }
            // double check for permission, also check whether anonymous
            // can read build info by testing configuration settings
            deleteAllBuildsByName(buildCoordinates.getBuildName(), buildCoordinates.getBuildRepo(), deleteArtifacts,
                    status);
        }
    }

    private ForbiddenException buildNotFoundException() {
        return new ForbiddenException(userNotAuthorizedMsg() + " to delete build info. Delete permission is needed.");
    }

    @Override
    public void deleteAllBuildsByName(String buildName, String buildRepo, boolean deleteArtifacts,
            BasicStatusHolder status) {
        searchBuildsByNameInternally(buildName, buildRepo)
                .forEach(build -> deleteBuild(build, deleteArtifacts, status));
    }

    @Override
    public void deleteBuildNumberByRetention(BuildRun buildRun, boolean deleteArtifacts, boolean async,
            BasicStatusHolder status) {
        BuildRetentionWorkItem buildRetentionWorkItem = new BuildRetentionWorkItem(buildRun, deleteArtifacts);
        if (async) {
            getTransactionalMe().deleteBuildAsync(buildRetentionWorkItem);
        } else {
            deleteBuild(buildRun, deleteArtifacts, status);
        }
    }

    @Override
    public void deleteBuild(BuildRun buildRun, boolean deleteArtifacts, BasicStatusHolder status) {
        if (buildRun == null) {
            status.error("Couldn't find requested build", log);
            return;
        }
        String name = buildRun.getName();
        String number = buildRun.getNumber();
        String started = buildRun.getStarted();
        //Its either you're admin (if we're converting, so we're supposed to keep this admin-only) or you have
        //new-mechanism permissions
        assertDeletePermissions(name, number, started, buildRun.getBuildRepo());
        if (buildInfoReady) {
            performDeleteBuild(buildRun, deleteArtifacts, status);
        } else {
            deleteBuildDuringMigration(buildRun, deleteArtifacts, status);
        }
        log.debug("Deleting build {} : {} : {} from {} repo", name, number, started, buildRun.getBuildRepo());
    }

    /**
     * 2 flows are possible here - one is legacy-supporting for conversion time, the other is the clean interceptor flow
     */
    private void performDeleteBuild(BuildRun buildRun, boolean deleteArtifacts, BasicStatusHolder status) {
        //deleting artifacts must come first because of foreign keys and the interceptor can't propagate the flag
        //from the request
        if (deleteArtifacts) {
            removeBuildArtifacts(buildRun, status);
        }
        String name = buildRun.getName();
        String number = buildRun.getNumber();
        String started = buildRun.getStarted();
        RepoPath targetBuildJsonPath =
                BuildServiceUtils.generateBuildJsonRepoPath(name, number, started, buildRun.getBuildRepo());
        if (repoService.exists(targetBuildJsonPath)) {
            // Will trigger the build info interceptor, or will throw a runtime error in case of permissions failure
            status.merge(repoService.undeploy(targetBuildJsonPath, false, true));
        } else {
            if (buildInfoReady) {
                throw new NotFoundException(
                        "Build '" + name + ":" + number + "' cannot be found in path " + targetBuildJsonPath);
            } else {
                //Must support the old logic for conversion-quasi-time
                Build build = buildStoreService.getBuild(buildRun);
                deleteBuildInternal(build, buildRun.getBuildRepo(), status);
            }
        }
    }

    private void deleteBuildDuringMigration(BuildRun buildRun, boolean deleteArtifacts, BasicStatusHolder status) {
        long buildId = buildStoreService.findIdFromBuild(buildRun);
        // Locking the action in order to keep the Build Info migration job safe in case it is currently running
        ConflictGuard lock = getConflictsGuard().getLock(BUILD_INFO_CALCULATION_KEY_PREFIX + buildId);
        try {
            boolean lockAcquired = lock.tryToLock(120, TimeUnit.SECONDS);
            if (lockAcquired) {
                performDeleteBuild(buildRun, deleteArtifacts, status);
            } else {
                log.warn("Could not obtain lock for Build id '{}', cannot delete build", buildId);
            }
        } catch (InterruptedException e) {
            log.error("Failed to acquire lock for Build id: '{}'. {}'", buildId, e.getMessage());
            Thread.currentThread().interrupt();
        } finally {
            lock.unlock();
        }
    }

    @Override
    public void deleteBuildInternal(Build build, String buildRepo, MutableStatusHolder status) {
        String buildName = build.getName();
        String buildNumber = build.getNumber();
        String buildStarted = build.getStarted();
        String buildData = buildName + "' #" + buildNumber + " from " + buildRepo + " repo";
        status.debug("Starting to remove build '" + buildData, log);
        buildStoreService.deleteBuild(buildName, buildNumber, buildStarted, buildRepo, buildInfoReady);
        AccessLogger.buildDelete(buildName);
        status.debug("Finished removing build '" + buildData, log);
        triggerXrayBuildDelete(buildName, buildNumber, buildRepo);
    }

    @Override
    public void deleteBuildAsync(BuildRetentionWorkItem buildRetentionWorkItem) {
        BuildRun buildRun = buildRetentionWorkItem.getBuildId();
        log.info("Async delete of build {} number: {}", buildRun.getName(), buildRun.getNumber());
        boolean deleteArtifacts = buildRetentionWorkItem.isDeleteArtifacts();
        BasicStatusHolder status = new BasicStatusHolder();
        getTransactionalMe().deleteBuild(buildRun, deleteArtifacts, status);
    }

    @Override
    public void removeBuildArtifacts(BuildRun buildRun, BasicStatusHolder status) {
        //Would be a shame to go through all of this just to find out you're missing build delete permissions later
        assertDeletePermissions(buildRun.getName(), buildRun.getNumber(), buildRun.getStarted(),
                buildRun.getBuildRepo());
        BuildWrapper build = getBuildInternally(buildRun);
        String buildName = build.getName();
        String buildNumber = build.getNumber();
        String buildRepo = build.getBuildRepo();
        String buildData = buildName + "' #" + buildNumber + " from " + buildRepo + " repo";
        status.debug("Starting to remove the artifacts of build '" + buildData, log);
        Set<ArtifactoryBuildArtifact> buildArtifactsInfos = getBuildArtifactsFileInfos(build, false, null, null);
        for (ArtifactoryBuildArtifact artifact : buildArtifactsInfos) {
            if (artifact.getFileInfo() != null) {
                RepoPath repoPath = artifact.getFileInfo().getRepoPath();
                BasicStatusHolder undeployStatus = repoService.undeploy(repoPath, true, true);
                status.merge(undeployStatus);
            }
        }
        status.debug("Finished removing the artifacts of build '" + buildData, log);
    }

    @Nullable
    @Override
    public BuildWrapper getLatestBuildByNameAndStatus(String buildName, String buildRepo, final String buildStatus) {
        if (StringUtils.isBlank(buildName)) {
            return null;
        }
        if (StringUtils.isBlank(buildStatus)) {
            return null;
        }
        //let's find all builds
        Set<BuildRun> buildsByName = searchBuildsByNameInternally(buildName, buildRepo);
        if (buildsByName == null || buildsByName.isEmpty()) { //no builds - no glory
            return null;
        }
        List<BuildRun> buildRuns = newArrayList(buildsByName);
        buildRuns.sort(BuildRunComparators.getComparatorFor(buildRuns));
        BuildRun latestBuildRun;

        if (buildStatus.equals(LATEST_BUILD)) {
            latestBuildRun = getLast(buildRuns, null);
        } else {
            latestBuildRun = getLast(filter(buildRuns, new Predicate<BuildRun>() {
                @Override
                public boolean apply(BuildRun buildRun) {
                    // Search for the latest build by the given status
                    return buildStatus.equals(buildRun.getReleaseStatus());
                }
            }), null);

        }
        return latestBuildRun == null ? null :
                getBuildInternally(latestBuildRun);
    }

    @Override
    @Nullable
    public BuildWrapper getLatestBuildByNameAndNumber(String buildName, String buildNumber, String buildRepo) {
        BuildWrapper buildWrapper = getLatestBuildByNameAndNumberInternally(buildName, buildNumber, buildRepo);
        if (buildWrapper == null || buildWrapper.getBuild() == null) {
            return null;
        }
        assertReadPermissions(buildName, buildNumber, buildWrapper.getBuild().getStarted(),
                buildWrapper.getBuildRepo());
        return buildWrapper;
    }

    @Override
    @Nullable
    public BuildWrapper getLatestBuildByNameAndNumberInternally(String buildName, String buildNumber,
            String buildRepo) {
        BuildRun latestBuildRun = getLatestBuildRunByNameAndNumberInternally(buildName, buildNumber, buildRepo);
        if (latestBuildRun != null) {
            return getBuildInternally(latestBuildRun);
        }
        return null;
    }

    private BuildRun getLatestBuildRunByNameAndNumberInternally(String buildName, String buildNumber, String repoKey) {
        if (StringUtils.isBlank(buildName) || StringUtils.isBlank(buildNumber) || StringUtils.isBlank(repoKey)) {
            return null;
        }
        return buildStoreService.getLatestBuildRun(buildName, buildNumber, repoKey);
    }

    @Override
    public SortedSet<BuildRun> searchBuildsByName(String buildName, @NonNull String buildRepo) {
        return searchBuildsByNameInternally(buildName, buildRepo).stream()
                .filter(build -> authService
                        .isBuildBasicRead(build.getName(), build.getNumber(), build.getStarted(), build.getBuildRepo()))
                .collect(sortedSetReverseStartDateCollector());
    }

    @Override
    public Set<BuildRun> searchBuildsByNameInternally(String buildName, @NonNull String buildRepo) {
        return buildStoreService.findBuildsByName(buildName, buildRepo);
    }

    @Override
    public List<String> getBuildNames(String buildRepo) {
        return getBuildsInfoInternally(buildRepo).stream()
                .filter(buildInfo -> authService.isBuildBasicRead(buildInfo.getName(), buildInfo.getRepo()))
                .map(BuildNameAndRepo::getName)
                .collect(Collectors.toList());
    }

    @Override
    public List<BuildNameAndRepo> getBuildsInfoInternally(String buildRepo) {
        return buildStoreService.getAllBuildsInfo(buildRepo).stream()
                .distinct()
                .sorted()
                .collect(Collectors.toList());
    }

    @Override
    public Set<BuildRun> searchBuildsByNameAndNumber(String buildName, String buildNumber, @NonNull String buildRepo) {
        return buildStoreService.findBuildsByNameAndNumber(buildName, buildNumber, buildRepo).stream()
                .filter(build -> authService
                        .isBuildBasicRead(build.getName(), build.getNumber(), build.getStarted(), build.getBuildRepo()))
                .collect(Collectors.toSet());
    }

    @Override
    public boolean isBuildInfoReady() {
        return buildInfoReady;
    }

    @Override
    public Set<BuildRun> searchBuildsByNameAndNumberInternal(String buildName, String buildNumber,
            @NonNull String buildRepo) {
        return buildStoreService.findBuildsByNameAndNumber(buildName, buildNumber, buildRepo);
    }

    @NonNull
    @Override
    public String getBuildInfoRepoForDefaultProject() {
        return getBuildInfoRepoKeyByProject(ProjectsService.DEFAULT_PROJECT);
    }

    @Override
    public String getBuildInfoRepoKeyByProject(@Nullable String projectKey) {
        if (ProjectsUtils.isDefaultProject(projectKey)) {
            projectKey = ProjectsService.DEFAULT_PROJECT;
        }
        final String finalProject = projectKey;
        String repoKey =
                projectToBuildRepoMap.computeIfAbsent(projectKey, s -> extractBuildInfoRepoKeyFromRepos(finalProject));
        log.debug("Build-info repository of project {} is {}", projectKey, repoKey);
        return repoKey;
    }


    @Override
    public Set<String> getAllBuildInfoRepos() {
        return new HashSet<>(allBuildsRepositories);
    }

    @Override
    public int getBuildsCountByRepoKey(String buildInfoRepoKey) {
        return buildStoreService.getBuildsCountFromSpecificRepo(buildInfoRepoKey);
    }

    @Override
    public int getUniqueBuildsCountByRepoKey(String buildInfoRepoKey) {
        return buildStoreService.getUniqueBuildsCountFromSpecificRepo(buildInfoRepoKey);
    }

    @Override
    public int getTotalBuildsCount() {
        return buildStoreService.getTotalBuildsCount();
    }

    @Override
    public Set<ArtifactoryBuildArtifact> getBuildArtifactsFileInfos(BuildWrapper buildWrapper, boolean useFallBack,
            List<String> sourceRepos, List<String> excludedRepos) {
        Build build = buildWrapper.getBuild();
        String buildName = build.getName();
        String buildNumber = build.getNumber();
        AqlBase.AndClause and = getBuildArtifactsByNameAndNumberAql(sourceRepos, excludedRepos, buildName, buildNumber);

        boolean onlyDefaultBuildRepoExists = isOnlyDefaultBuildRepoExists();
        if (onlyDefaultBuildRepoExists) {
            log.debug("Executing Artifacts search for build {}:{}", buildName, buildNumber);
        } else {
            String startedTimestamp = String.valueOf(BuildInfoUtils.parseBuildTime(build.getStarted()));
            log.debug("Executing Artifacts search for build {}:{}:{}", buildName, buildNumber, startedTimestamp);
            and.append(
                    AqlApiItem.property().property(PROP_BUILD_TIMESTAMP, AqlComparatorEnum.equals, startedTimestamp));
        }

        AqlEagerResult<AqlBaseFullRowImpl> aqlResult;
        AqlBase buildArtifactsQuery;
        buildArtifactsQuery = AqlApiItem.create().filter(and);
        aqlResult = aqlService.executeQueryEager(buildArtifactsQuery);

        if (!onlyDefaultBuildRepoExists && aqlResult.getSize() == 0) {
            log.debug("Artifacts search returned 0 results. Trying to fallback search for build {}:{}", buildName,
                    buildNumber);
            AqlBase.AndClause andFallback =
                    getBuildArtifactsByNameAndNumberAql(sourceRepos, excludedRepos, buildName, buildNumber);
            buildArtifactsQuery = AqlApiItem.create().filter(andFallback);
            aqlResult = aqlService.executeQueryEager(buildArtifactsQuery);
        }

        log.debug("Search returned {} artifacts", aqlResult.getSize());
        Multimap<String, Artifact> buildArtifacts = BuildServiceUtils.getBuildArtifacts(build);

        log.debug("This build contains {} artifacts (taken from build info)", buildArtifacts.size());
        List<String> virtualRepoKeys = getVirtualRepoKeys();
        Set<ArtifactoryBuildArtifact> matchedArtifacts = matchArtifactsToFileInfos(aqlResult.getResults(),
                buildArtifacts, virtualRepoKeys);
        log.debug("Matched {} build artifacts to actual paths returned by search", matchedArtifacts.size());

        //buildArtifacts contains all remaining artifacts that weren't matched - match them with the weak search
        //only if indicated and if such remaining unmatched artifacts still exist in the map.
        if (useFallBack && !buildArtifacts.isEmpty()) {
            log.debug("Unmatched artifacts exist and 'use weak match fallback' flag is lit - executing weak match");
            Set<ArtifactoryBuildArtifact> weaklyMatchedArtifacts =
                    matchUnmatchedArtifactsNonStrict(buildWrapper, sourceRepos,
                            excludedRepos, buildArtifacts, virtualRepoKeys, aqlService);
            log.debug("Weak match has matched {} additional artifacts", weaklyMatchedArtifacts);
            matchedArtifacts.addAll(weaklyMatchedArtifacts);
        }
        //Lastly, populate matchedArtifacts with all remaining unmatched artifacts with null values to help users of
        //this function know if all build artifacts were found.
        log.debug("{} artifacts were not matched to actual paths", buildArtifacts.size());
        matchedArtifacts.addAll(buildArtifacts.values().stream()
                .map(artifact -> new ArtifactoryBuildArtifact(artifact, null))
                .collect(Collectors.toList()));
        return matchedArtifacts;
    }

    @Override
    @SuppressWarnings({"unchecked", "java:S2139"})
    public Map<Dependency, FileInfo> getBuildDependenciesFileInfos(BuildWrapper buildWrapper) {
        AqlBase.AndClause<AqlApiBuild> and = and(
                AqlApiBuild.name().equal(buildWrapper.getName()),
                AqlApiBuild.number().equal(buildWrapper.getNumber()),
                AqlApiBuild.buildRepo().equal(buildWrapper.getBuildRepo())
        );
        log.debug("Executing dependencies search for build {}:{} from {} repo", buildWrapper.getName(),
                buildWrapper.getNumber(), buildWrapper.getBuildRepo());

        AqlBase buildDependenciesQuery = AqlApiBuild.create().filter(and);
        //Ordering by the last updated field, in case of duplicates with the same checksum.
        buildDependenciesQuery.include(
                module().dependecy().name(),
                module().dependecy().item().sha1Actual(),
                module().dependecy().item().md5Actual(),
                module().dependecy().item().sha1Original(),
                module().dependecy().item().md5Orginal(),
                module().dependecy().item().created(),
                module().dependecy().item().modifiedBy(),
                module().dependecy().item().createdBy(),
                module().dependecy().item().updated(),
                module().dependecy().item().repo(),
                module().dependecy().item().path(),
                module().dependecy().item().name(),
                module().dependecy().item().size()
        ).addSortElement(module().dependecy().item().updated()).asc();
        Map<Dependency, FileInfo> matchedDependencies;
        Multimap<String, Dependency> buildDependencies;
        try (AqlLazyResult<AqlBuild> aqlLazyResult = aqlService.executeQueryLazy(buildDependenciesQuery)) {
            buildDependencies = BuildServiceUtils.getBuildDependencies(buildWrapper.getBuild());
            log.debug("This build contains {} dependencies (taken from build info)", buildDependencies.size());
            matchedDependencies = matchDependenciesToFileInfos(aqlLazyResult, buildDependencies,
                    getVirtualRepoKeys(), repoService);
        } catch (Exception e) {
            log.error("Failed to close dependencies lazy query, {}", e.getMessage());
            log.debug("Failed to close dependencies lazy query", e);
            throw new BuildDependenciesCloseException("Failed to close dependencies lazy query", e);
        }

        log.debug("Matched {} build dependencies to actual paths returned by search", matchedDependencies.size());
        //Lastly, populate matchedDependencies with all remaining unmatched dependencies with null values to help users
        //of this function know if all build artifacts were found.
        log.debug("{} dependencies were not matched to actual paths", buildDependencies.size());
        for (Dependency dependency : buildDependencies.values()) {
            if (!matchedDependencies.containsKey(dependency)) {
                matchedDependencies.put(dependency, null);
            }
        }
        return matchedDependencies;
    }

    @Override
    public void exportTo(ExportSettings settings) {
        MutableStatusHolder multiStatusHolder = settings.getStatusHolder();
        multiStatusHolder.debug("Starting build info export", log);

        File buildsFolder = new File(settings.getBaseDir(), BUILDS_EXPORT_DIR);
        prepareBuildsFolder(settings, multiStatusHolder, buildsFolder);
        if (multiStatusHolder.isError()) {
            return;
        }

        try {
            long exportedBuildCount = 1;
            List<BuildNameAndRepo> buildsInfo = buildStoreService.getAllBuildsInfo(null);
            for (BuildNameAndRepo buildInfo : buildsInfo) {
                Set<BuildRun> buildsByName =
                        buildStoreService.findBuildsByName(buildInfo.getName(), buildInfo.getRepo());
                for (BuildRun buildRun : buildsByName) {
                    String buildNumber = buildRun.getNumber();
                    try {
                        exportBuild(settings, buildRun, exportedBuildCount, buildsFolder);
                        exportedBuildCount++;
                    } catch (Exception e) {
                        String errorMessage = format("Failed to export build info: %s:%s", buildInfo, buildNumber);
                        if (settings.isFailFast()) {
                            throw new Exception(errorMessage, e);
                        }
                        multiStatusHolder.error(errorMessage, e, log);
                    }
                }
            }
        } catch (Exception e) {
            multiStatusHolder.error("Error occurred during build info export.", e, log);
        }

        if (settings.isIncremental() && !multiStatusHolder.isError()) {
            try {
                log.debug("Cleaning previous builds backup folder.");

                File[] backupDirsToRemove = settings.getBaseDir().listFiles(
                        (dir, name) -> name.startsWith(BACKUP_BUILDS_FOLDER));
                if (backupDirsToRemove != null) {
                    for (File backupDirToRemove : backupDirsToRemove) {
                        log.debug("Cleaning previous build backup folder: {}", backupDirToRemove.getAbsolutePath());
                        FileUtils.forceDelete(backupDirToRemove);
                    }
                }
            } catch (IOException e) {
                multiStatusHolder.error("Failed to clean previous builds backup folder.", e, log);
            }
        }

        multiStatusHolder.debug("Finished build info export", log);
    }

    @NotNull
    private AqlBase.AndClause getBuildArtifactsByNameAndNumberAql(List<String> sourceRepos, List<String> excludedRepos,
            String buildName, String buildNumber) {
        AqlBase.AndClause and = and();
        addIncludeExcludeRepos(sourceRepos, excludedRepos, and, true);
        and.append(AqlApiItem.property().property(PROP_BUILD_NAME, AqlComparatorEnum.equals, buildName));
        and.append(AqlApiItem.property().property(PROP_BUILD_NUMBER, AqlComparatorEnum.equals, buildNumber));
        return and;
    }

    private boolean isOnlyDefaultBuildRepoExists() {
        return getAllBuildInfoRepos().size() == 1;
    }

    /**
     * Do not use this. Use {@link #getBuildInfoRepoKeyByProject(String projectKey)}
     */
    private String extractBuildInfoRepoKeyFromRepos(@Nullable String projectKey) {
        String prefix = StringUtils.EMPTY;
        if (ProjectsUtils.isDefaultProject(projectKey)) {
            prefix = "artifactory";
        } else {
            ProjectSlimEntity projectSpec = projectsService.getProjectSpec(projectKey);
            if (projectSpec != null) {
                prefix = projectSpec.getProjectKey();
            }
        }
        if (StringUtils.isEmpty(prefix)) {
            throw new IllegalArgumentException("Project " + projectKey + " was not found.");
        }
        final String finalPrefix = prefix;
        log.debug("Looking for build-info repository for project {}.", projectKey);
        String buildRepoKey = getAllBuildInfoRepos().stream()
                .filter(repo -> repo.startsWith(finalPrefix + BuildAddon.BUILD_INFO_REPO_EXTENSION))
                .findAny().orElse(null);
        if (buildRepoKey == null) {
            return addonsManager.addonByType(BuildAddon.class).createProjectRelatedBuildInfoRepo(projectKey);
        }
        return buildRepoKey;
    }

    /**
     * Removes the entire Build Info repo files, while avoiding the storage interceptors
     */
    private void deleteAllBuildRepos() {
        // Needs to be taken care of when adding support for projects related builds import\export
        Authentication currentAuth = SecurityContextHolder.getContext().getAuthentication();
        try {
            for (String buildRepo : getAllBuildInfoRepos()) {
                RepoPath targetBuild = RepoPathFactory.create(buildRepo);
                SecurityContextHolder.getContext().setAuthentication(new SystemAuthenticationToken());
                BasicStatusHolder statusHolder = repoService
                        .undeploy(targetBuild, true, new DeleteContext(targetBuild).avoidBuildDeleteInterceptor());
                if (statusHolder.isError()) {
                    throw new StorageException(
                            "Could not delete build json from " + targetBuild + ": "
                                    + statusHolder.getLastError().getMessage(),
                            statusHolder.getLastError().getException());
                }
            }
        } finally {
            // Restore previous permissions
            SecurityContextHolder.getContext().setAuthentication(currentAuth);
        }
    }

    @Override
    public void importFrom(ImportSettings settings) {
        val status = settings.getStatusHolder();
        status.status("Starting build info import", log);

        clearCaches();
        val importKind = settings.getImportKind();
        if (ImportKind.HYBRID_MIGRATION.equals(importKind)) {
            log.debug("Import kind is {} builds an repositories will not be deleted", importKind);
        } else {
            log.debug("Import kind is {} builds and repositories will be deleted", importKind);
            deleteBuildsAndRepositories(status);
        }
        importBuildFiles(settings, status);
    }

    @VisibleForTesting
    void clearCaches() {
        log.debug("Restarting the Build Info repositories map");
        projectToBuildRepoMap.clear();
        refreshAllBuildsRepositories();
    }

    @VisibleForTesting
    void deleteBuildsAndRepositories(MutableStatusHolder status) {
        dbService.invokeInTransaction("BuildImport-deleteAllBuilds", () -> {
            try {
                // First we should undeploy all builds in the builds repo
                deleteAllBuildRepos();
                // This might can be removed, we can keep it
                buildStoreService.deleteAllBuilds(buildInfoReady);
            } catch (Exception e) {
                status.error("Failed to delete builds root node", e, log);
            }
            return null;
        });
    }

    @VisibleForTesting
    void importBuildFiles(ImportSettings settings, MutableStatusHolder status) {
        File buildsFolder = new File(settings.getBaseDir(), BUILDS_EXPORT_DIR);
        String buildsFolderPath = buildsFolder.getPath();
        if (!buildsFolder.exists()) {
            status.debug("'" + buildsFolderPath + "' folder is either non-existent or not a " +
                    "directory. old-style build info import was not performed", log);
            return;
        }

        IOFileFilter buildExportFileFilter = new AbstractFileFilter() {
            @Override
            public boolean accept(File file) {
                String fileName = file.getName();
                return fileName.startsWith("build") && fileName.endsWith(".xml");
            }
        };

        Collection<File> buildExportFiles =
                FileUtils.listFiles(buildsFolder, buildExportFileFilter, DirectoryFileFilter.DIRECTORY);

        if (buildExportFiles.isEmpty()) {
            status.status("'" + buildsFolderPath + "' folder does not contain build export files. " +
                    "Build info import was not performed", log);
            return;
        }

        importBuildFiles(settings, buildExportFiles);
        status.warn("Old-style import builds witnessed at location '"
                + buildsFolder.getParent() + File.separator + buildsFolder.getName()
                +
                "'.  If a build-info repository content export exists as well in this import it will be disregarded " +
                "for backwards compatibility.", log);
        status.warn("The correct way to import and export builds is to enable content export and let " +
                "Artifactory handle the build info repo import/export", log);
        settings.setExcludeBuildInfoRepo(true);
        status.status("Finished build info import", log);
    }

    private void importBuildFiles(ImportSettings settings, Collection<File> buildExportFiles) {
        for (File buildExportFile : buildExportFiles) {
            try (FileInputStream inputStream = new FileInputStream(buildExportFile)) {
                ImportableExportableBuild importableBuild = (ImportableExportableBuild) buildJaxbAdapter
                        .fromXML(inputStream);
                // import each build in a separate transaction
                getTransactionalMe().importBuild(settings, importableBuild);
            } catch (Exception e) {
                settings.getStatusHolder().error("Error occurred during build info import", e, log);
                if (settings.isFailFast()) {
                    break;
                }
            }
        }
    }

    @Override
    public void importBuild(ImportSettings settings, ImportableExportableBuild importableBuild) {
        String buildName = importableBuild.getBuildName();
        MutableStatusHolder multiStatusHolder = settings.getStatusHolder();
        String buildNumber = importableBuild.getBuildNumber();
        String buildStarted = importableBuild.getBuildStarted();
        String buildRepo = normaliseBuildRepoName(importableBuild.getBuildRepo());
        try {
            multiStatusHolder.debug(format("Beginning import of build: %s:%s:%s from %s repo", buildName, buildNumber,
                    buildStarted, buildRepo), log);
            Build build = JsonUtils.getInstance().readValue(importableBuild.getJson(), Build.class);
            // Needs to be taken care of when adding support for projects related builds import\export
            addBuild(build, buildRepo);
        } catch (Exception e) {
            String msg =
                    String.format("Could not import build %s:%s:%s from %s repo", buildName, buildNumber, buildStarted,
                            buildRepo);
            // Print stack trace in debug
            log.debug(msg, e);
            multiStatusHolder.error(msg, e, log);
        }
        multiStatusHolder
                .debug(format("Finished import of build: %s:%s:%s from %s repo", buildName, buildNumber, buildStarted,
                        buildRepo), log);
    }

    @Override
    public PromotionResult promoteBuild(BuildRun buildRun, Promotion promotion) {
        validateFeatureAllowed();
        // Assertion is done here for efficiency
        assertUploadPermissions(buildRun.getName(), buildRun.getNumber(), buildRun.getStarted(),
                buildRun.getBuildRepo());
        BuildPromotionHelper buildPromotionHelper = new BuildPromotionHelper();
        return buildPromotionHelper.promoteBuild(buildRun, promotion);
    }

    @Override
    public void renameBuilds(String from, String to, String buildRepo) {
        Set<BuildRun> buildsToRename = searchBuildsByNameInternally(from, buildRepo);
        if (buildsToRename.isEmpty()) {
            log.error("Could not find builds by the name '{}'. No builds were renamed.", from);
            return;
        }
        //Although the caller ran a permission check (to properly report back before async) we do it again since
        // between async dispatch and execution stuff might change
        Set<BuildRun> allowedBuildsToRename = buildsToRename.stream()
                .filter(build -> authService
                        .canReadBuild(build.getName(), build.getNumber(), build.getStarted(), build.getBuildRepo()))
                .filter(build -> authService
                        .canDeleteBuild(build.getName(), build.getNumber(), build.getStarted(), build.getBuildRepo()))
                .filter(build -> authService
                        .canUploadBuild(build.getName(), build.getNumber(), build.getStarted(), build.getBuildRepo()))
                .collect(Collectors.toSet());
        if (allowedBuildsToRename.size() != buildsToRename.size() && log.isErrorEnabled()) {
            log.error("The user: '{}' lacks the required permissions to rename build '{}' from build-info repo {}",
                    authService.currentUsername(), from, buildRepo);
        }
        for (BuildRun buildToRename : buildsToRename) {
            try {
                getTransactionalMe().renameBuild(buildToRename, to);
                log.info(
                        "Renamed build number '{}' #{} at build-info repository '{}' from '{}' to '{}'.",
                        buildToRename.getNumber(), buildToRename.getStarted(), buildRepo, buildToRename.getName(), to);
            } catch (Exception e) {
                log.error("Failed to rename build: '{}' #{} from '{}' that started at {}: {}", buildToRename.getName(),
                        buildToRename.getNumber(), buildRepo, buildToRename.getStarted(), e.getMessage());
                log.error("", e);
            }
        }
    }

    @Override
    public void renameBuild(BuildRun buildRun, String to) {
        BuildWrapper buildWrapper = getBuildInternally(buildRun);
        if (buildWrapper == null || !buildWrapper.isBuildExist()) {
            throw new NotFoundException("Cannot rename non existent build " + buildRun);
        }
        //shouldRename also modifies the buildJson object's build name attribute
        if (shouldRename(buildRun, to, buildWrapper.getBuild())) {
            // delete according to the build-info state, renaming of build will do migration
            deleteBuild(buildRun, false, new BasicStatusHolder());
            addBuild(buildWrapper.getBuild(), buildWrapper.getBuildRepo());
        }
    }

    /**
     * This piece of legacy code supposedly deals with a case where the db row representing the build (in builds table)
     * and the json representing the build (either in the repo or in the build_jsosns table) have a mismatch in their
     * name.
     * I dunno how or why this would happen but I guess its here for a reason (lost to oblivion) so i'm keeping it.
     */
    private boolean shouldRename(BuildRun buildRun, String to, Build buildJson) {
        boolean changed = false;
        if (!StringUtils.equals(buildJson.getName(), to)) {
            buildJson.setName(to);
            changed = true;
        }
        if (!StringUtils.equals(buildRun.getName(), to)) {
            changed = true;
        }
        if (!changed) {
            log.info("Build {} already named {} nothing to do!", buildRun, to);
        }
        return changed;
    }

    @Override
    public void updateBuild(@NonNull BuildWrapper buildWrapper) {
        validateFeatureAllowed();
        Build build = buildWrapper.getBuild();
        String name = build.getName();
        String number = build.getNumber();
        String started = build.getStarted();
        String buildRepo = buildWrapper.getBuildRepo();

        // Checking permissions here, as cannot check inside async later
        assertDeletePermissions(name, number, started, buildWrapper.getBuildRepo());
        assertUploadPermissions(name, number, started, buildWrapper.getBuildRepo());
        // Reaching update implies the same build (name, number and started) already exists
        log.info("Updating build {} Number {} that started at {} from {} repo", name, number, started, buildRepo);
        getTransactionalMe().addBuild(build, buildWrapper.getBuildRepo());
        log.info("Update of build {} Number {} that started at {} from {} repo completed successfully", name, number,
                started, buildRepo);
    }

    @Override
    public void addPromotionStatus(BuildWrapper buildWrapper, PromotionStatus promotion) {
        validateFeatureAllowed();
        Build build = buildWrapper.getBuild();
        String name = build.getName();
        String number = build.getNumber();
        String started = build.getStarted();
        build.addStatus(promotion);
        // Assert MUST BE KEPT HERE, as we perform delete before upload.
        // If we don't assert and in case there is no upload permission we will be left with no build at all.
        assertUploadPermissions(name, number, started, buildWrapper.getBuildRepo());
        // In the state before build migration is finished: update the DB with the updated build json (as well as
        // writing the build.json to the build-info repository
        if (!buildInfoReady) {
            buildStoreService
                    .addPromotionStatus(buildWrapper, promotion, authService.currentUsername(), buildInfoReady);
        }
        log.debug("Overwriting the .json content of '{}:{}' with new build model including promotion", name, number);
        deletePreviousBuildAsSystem(build);
        addBuild(build, buildWrapper.getBuildRepo());
    }

    /**
     * Performs delete for the build json by the system user (in order to allow promotion even for users without
     * delete permission)
     *
     * Delete is performed in a separate transaction:
     * We need to manually perform the delete BEFORE the deploy,
     * otherwise it can cause deploy-->delete instead of delete-->deploy.
     * This is because {@link org.artifactory.storage.db.fs.session.SqlStorageSession} currently doesn't hold the
     * order of events.
     *
     * @param build Build to be deleted
     */
    private void deletePreviousBuildAsSystem(Build build) {
        BasicStatusHolder status = new BasicStatusHolder();
        Authentication currentAuth = SecurityContextHolder.getContext().getAuthentication();
        try {
            SecurityContextHolder.getContext().setAuthentication(new SystemAuthenticationToken());
            BuildRun buildRun = getBuildRunInternally(build.getName(), build.getNumber(), build.getStarted());
            if (!buildInfoReady) {
                log.debug("Invoking Build info deletion.");
                deleteBuild(buildRun, false, status);
            } else {
                log.debug("Invoking Build info deletion in new transaction in case.");
                dbService.invokeInTransaction(
                        "deleteBuildInPromotionBySystem" + build.getName() + build.getNumber() + build.getStarted(),
                        () -> {
                            getTransactionalMe().deleteBuild(buildRun, false, status);
                            return null;
                        });
            }
        } finally {
            // Restore previous permissions
            SecurityContextHolder.getContext().setAuthentication(currentAuth);
        }
        if (status.hasErrors() || status.hasWarnings()) {
            log.warn("Couldn't delete build as part of promotion: {}.", status.getLastError().getMessage());
        }
    }

    @Nullable
    @Override
    public List<PublishedModule> getPublishedModules(String buildNumber, String buildStarted, String buildName,
            String buildRepo) {
        assertBasicReadPermissions(buildName, buildNumber, formatBuildTime(buildStarted), buildRepo);
        return buildStoreService.getPublishedModules(buildNumber, buildStarted, buildRepo);
    }

    @Nullable
    @Override
    public List<ModuleArtifact> getModuleArtifact(String buildName, String buildNumber, String moduleId,
            String buildStarted, String buildRepo) {
        assertBasicReadPermissions(buildName, buildNumber, formatBuildTime(buildStarted), buildRepo);
        boolean onlyDefaultBuildRepoExists = isOnlyDefaultBuildRepoExists();
        return buildStoreService.getModuleArtifact(buildName, buildNumber, moduleId, buildStarted, buildRepo,
                onlyDefaultBuildRepoExists);
    }

    @Nullable
    @Override
    public List<ModuleDependency> getModuleDependency(String buildNumber, String moduleId, String buildStarted,
            String buildName, String buildRepo) {
        assertBasicReadPermissions(buildName, buildNumber, formatBuildTime(buildStarted), buildRepo);
        return buildStoreService.getModuleDependency(buildName, buildNumber, moduleId, buildStarted, buildRepo);
    }

    private void exportBuild(ExportSettings settings, BuildRun buildRun,
            long exportedBuildCount, File buildsFolder) throws IOException {
        MutableStatusHolder multiStatusHolder = settings.getStatusHolder();

        String buildName = buildRun.getName();
        String buildNumber = buildRun.getNumber();
        String buildStarted = buildRun.getStarted();
        String buildRepo = buildRun.getBuildRepo();
        multiStatusHolder.debug(
                format("Beginning export of build: %s:%s:%s from %s repo", buildName, buildNumber, buildStarted,
                        buildRepo), log);

        ImportableExportableBuild exportedBuild = getExportableBuild(buildRun);

        File buildFile = new File(buildsFolder, "build" + exportedBuildCount + ".xml");
        exportBuildToFile(exportedBuild, buildFile);

        multiStatusHolder.debug(
                format("Finished export of build: %s:%s:%s from %s repo", buildName, buildNumber, buildStarted,
                        buildRepo), log);
    }

    @Override
    public ImportableExportableBuild getExportableBuild(BuildRun buildRun) {
        assertReadPermissions(buildRun.getName(), buildRun.getNumber(), buildRun.getStarted(), buildRun.getBuildRepo());
        BuildEntity buildEntity = buildStoreService.getBuildEntity(buildRun);
        ImportableExportableBuild exportedBuild = new ImportableExportableBuild();
        exportedBuild.setVersion(EXPORTABLE_BUILD_VERSION);
        exportedBuild.setBuildName(buildEntity.getBuildName());
        exportedBuild.setBuildNumber(buildEntity.getBuildNumber());
        exportedBuild.setBuildStarted(formatDateToString(buildEntity.getBuildDate()));
        exportedBuild.setBuildRepo(buildEntity.getRepo());

        String jsonString = getBuildAsJson(buildRun);
        exportedBuild.setJson(jsonString);

        exportedBuild.setCreated(buildEntity.getCreated());
        exportedBuild.setLastModified(buildEntity.getModified());
        exportedBuild.setCreatedBy(buildEntity.getCreatedBy());
        exportedBuild.setLastModifiedBy(buildEntity.getModifiedBy());
        return exportedBuild;
    }

    private void exportBuildToFile(ImportableExportableBuild exportedBuild, File buildFile) throws IOException {
        try (FileOutputStream buildFileOutputStream = new FileOutputStream(buildFile)) {
            buildJaxbAdapter.toXML(exportedBuild, buildFileOutputStream);
        }
    }

    @Override
    public List<ModuleArtifact> getModuleArtifactsForDiffWithPaging(BuildParams buildParams) {
        assertBasicReadPermissions(buildParams.getBuildName(), buildParams.getCurrBuildNum(),
                formatBuildTime(buildParams.getCurrBuildDate()), buildParams.getBuildRepo());
        return buildStoreService.getModuleArtifactsForDiffWithPaging(buildParams, isOnlyDefaultBuildRepoExists());
    }

    @Override
    public List<ModuleDependency> getModuleDependencyForDiffWithPaging(BuildParams buildParams) {
        assertBasicReadPermissions(buildParams.getBuildName(), buildParams.getCurrBuildNum(),
                formatBuildTime(buildParams.getCurrBuildDate()), buildParams.getBuildRepo());
        return buildStoreService.getModuleDependencyForDiffWithPaging(buildParams);
    }

    @Override
    public List<GeneralBuild> getPrevBuildsList(String buildName, String buildDate, String buildRepo) {
        return buildStoreService.getPrevBuildsList(buildName, buildDate, buildRepo).stream()
                .filter(generalBuild -> authService
                        .isBuildBasicRead(generalBuild.getBuildName(), generalBuild.getBuildNumber(),
                                formatBuildTime(generalBuild.getBuildDate()), generalBuild.getBuildRepo()))
                .collect(Collectors.toList());
    }

    @Override
    public List<BuildProps> getBuildPropsData(BuildParams buildParams) {
        String name = buildParams.getBuildName();
        String number = buildParams.getCurrBuildNum();
        String started = formatBuildTime(buildParams.getCurrBuildDate());
        String buildRepo = buildParams.getBuildRepo();
        assertReadPermissions(name, number, started, buildRepo);
        return buildStoreService.getBuildPropsData(buildParams);
    }

    @Override
    public List<BuildId> getLatestBuildIDsPaging(ContinueBuildFilter buildFilter) {
        return BuildServiceUtils.iterateFetchFromDb(buildStoreService::getLatestBuildIDsPaging,
                build -> authService
                        .isBuildBasicRead(build.getName(), build.getNumber(), build.getStarted(), build.getBuildRepo()),
                buildId -> buildId, buildFilter);
    }

    @Override
    public List<String> getDistinctBuildNamesPaging(ContinueBuildFilter buildFilter) {
        return BuildServiceUtils.iterateFetchFromDb(buildStoreService::getDistinctBuildNamesPaging,
                build -> true,
                buildName -> new BuildIdImpl(-1L, buildName, null, -1, null),
                buildFilter);
    }

    @Override
    public int getDistinctBuildNamesCount(ContinueBuildFilter continueBuildFilter) {
        return buildStoreService.getDistinctBuildNamesCount(continueBuildFilter);
    }

    @Override
    public List<BuildId> findBuildIDsByNameCaseInsensitive(BuildsSearchFilter buildSearchFilter) {
        List<BuildId> localBuilds = buildStoreService
                .findLatestBuildIDsByNameCaseInsensitive(buildSearchFilter);
        long resultsLimit = min(buildSearchFilter.getLimit(), ConstantValues.searchUserQueryLimit.getInt());

        if (authService.hasBasicReadPermissionForAllBuilds()) {
            return localBuilds.stream()
                    .limit(resultsLimit)
                    .collect(Collectors.toList());
        }

        return localBuilds.stream()
                .filter(build -> authService
                        .canReadBuild(build.getName(), build.getNumber(), build.getStarted(), build.getBuildRepo()))
                .limit(resultsLimit)
                .collect(Collectors.toList());
    }

    @Override
    public List<GeneralBuild> getBuildVersions(BuildsSearchFilter filter) {
        List<GeneralBuild> buildsList = Lists.newArrayList();
        if (authService.isAdmin()) {
            filter.setLimit(filter.getLimit());
            buildsList = buildStoreService.getBuildNumbersByBuildName(filter);
        } else {
            while (buildsList.size() < filter.getLimit()) {
                List<GeneralBuild> buildVersionsBatch = buildStoreService.getBuildNumbersByBuildName(filter);
                List<GeneralBuild> filteredBuildVersionsByPermissions = filterVersionByPermissions(filter,
                        buildVersionsBatch);
                buildsList.addAll(filteredBuildVersionsByPermissions);
                if (buildVersionsBatch.size() < filter.getDaoLimit()) {
                    break;
                }
            }
        }
        return buildsList.stream().limit(filter.getLimit()).collect(Collectors.toList());
    }

    private List<GeneralBuild> filterVersionByPermissions(BuildsSearchFilter filter,
            List<GeneralBuild> buildVersionsBatch) {
        return buildVersionsBatch.stream()
                .limit(filter.getLimit())
                .filter(build -> authService.isBuildBasicRead(build.getBuildName(), build.getBuildNumber(),
                        formatBuildTime(build.getBuildDate()), build.getBuildRepo()))
                .collect(Collectors.toList());
    }

    @Override
    public List<GeneralBuild> getBuildForName(String buildName, String buildRepo, ContinueBuildFilter buildFilter) {
        return iterateFetchFromDb(continuePageBuildFilter ->
                        buildStoreService.getBuildForName(buildName, buildRepo, continuePageBuildFilter),
                build -> authService
                        .isBuildBasicRead(build.getBuildName(), build.getBuildNumber(),
                                formatBuildTime(build.getBuildDate()), build.getBuildRepo()),
                generalBuild -> new BuildIdImpl(generalBuild.getBuildId(), generalBuild.getBuildName(),
                        generalBuild.getBuildNumber(), generalBuild.getBuildDate(), generalBuild.getBuildRepo()),
                buildFilter);
    }

    @Override
    public List<FileInfo> collectBuildArtifacts(BuildWrapper buildWrapper, @Nullable List<String> sourceRepos,
            @Nullable List<String> excludedRepos, @Nullable BasicStatusHolder status) {
        status = status == null ? new BasicStatusHolder() : status;
        if (buildWrapper == null || !buildWrapper.isBuildExist()) {
            status.error("Invalid build given.", SC_BAD_REQUEST, log);
            return Lists.newArrayList();
        }
        log.info("Collecting Build artifacts for build {}:{} from {} repo", buildWrapper.getName(),
                buildWrapper.getNumber(), buildWrapper.getBuildRepo());
        Set<ArtifactoryBuildArtifact> infos =
                getBuildArtifactsFileInfos(buildWrapper, false, sourceRepos, excludedRepos);
        BuildServiceUtils
                .verifyAllArtifactInfosExistInSet(buildWrapper.getBuild(), true, status, infos, VerifierLogLevel.WARN);
        return Lists.newArrayList(BuildServiceUtils.toFileInfoList(infos));
    }

    @Override
    public void assertBasicReadPermissions(String buildName, String buildNumber, String buildStarted,
            String buildRepo) {
        if (!authService.isBuildBasicRead(buildName, buildNumber, buildStarted, buildRepo)) {
            throw new ForbiddenException(userNotAuthorizedMsg() +
                    " to access build info. 'Global Basic Read' Flag turned on or read permission is needed.");
        }
    }

    @Override
    public void assertReadPermissions(String buildName, String buildNumber, String buildStarted, String buildRepo) {
        if (!authService.canReadBuild(buildName, buildNumber, buildStarted, buildRepo)) {
            throw new ForbiddenException(userNotAuthorizedMsg() + " to access build info. Read permission is needed.");
        }
    }

    @Override
    public void assertDeletePermissions(String buildName, String buildNumber, String buildStarted, String buildRepo) {
        if (!authService.canDeleteBuild(buildName, buildNumber, buildStarted, buildRepo)) {
            throw new ForbiddenException(
                    userNotAuthorizedMsg() + " to delete build info. Delete permission is needed.");
        }
    }

    @Override
    public void assertUploadPermissions(String buildName, String buildNumber, String buildStarted, String buildRepo) {
        if (!authService.canUploadBuild(buildName, buildNumber, buildStarted, buildRepo)) {
            throw new ForbiddenException(
                    userNotAuthorizedMsg() + " to upload build info. Upload permission is needed.");
        }
    }

    @Override
    @NonNull
    public String normaliseBuildRepoName(@Nullable String buildRepo) {
        if (StringUtils.isEmpty(buildRepo)) {
            return getBuildInfoRepoForDefaultProject();
        }
        return buildRepo;
    }

    @Override
    @NonNull
    public String getPreferredBuildRepo(@Nullable String buildRepo, @Nullable String projectKey) {
        if (isNotEmpty(buildRepo)) {
            return normaliseBuildRepoName(buildRepo);
        }
        return getBuildInfoRepoKeyByProject(projectKey);
    }

    @Override
    public BuildGeneralInfo getBuildGeneralInfo(String buildName, String buildNumber, String buildStarted,
            @NonNull String buildRepo) {
        BuildRun buildRun = getBuildRunNonStrict(buildName, buildNumber, buildStarted, buildRepo);
        BuildWrapper buildWrapper = null;
        if (buildRun != null) {
            buildWrapper = getBuildInternally(buildRun);
        }
        if (buildWrapper != null && buildWrapper.isBuildExist()) {
            assertBasicReadPermissions(buildWrapper.getBuild().getName(), buildWrapper.getBuild().getNumber(),
                    buildWrapper.getBuild().getStarted(), buildWrapper.getBuildRepo());
        } else {
            log.warn("Build {}:{}-{} from {} repo was not found.", buildName, buildNumber, buildStarted, buildRepo);
            return null;
        }
        Build build = buildWrapper.getBuild();
        String buildAgent = (build.getBuildAgent() == null) ? null : build.getBuildAgent().toString();
        String agent = (build.getAgent() == null) ? null : build.getAgent().toString();
        String foundBuildName = build.getName();
        String foundBuildNumber = build.getNumber();
        String foundBuildStarted = build.getStarted();
        return BuildGeneralInfo.builder()
                .buildName(foundBuildName)
                .lastBuildTime(foundBuildStarted)
                .agent(agent)
                .buildAgent(buildAgent)
                .artifactoryPrincipal(build.getArtifactoryPrincipal())
                .principal(build.getPrincipal())
                .duration(DateUtils.getDuration(build.getDurationMillis()))
                .buildNumber(buildNumber)
                .time(DateUtils.parseBuildDate(build.getStarted()))
                .url(build.getUrl())
                .isBuildFullView(
                        authService.canReadBuild(foundBuildName, foundBuildNumber, foundBuildStarted, buildRepo))
                .canManage(authService.canManageBuild(foundBuildName, foundBuildNumber, foundBuildStarted, buildRepo))
                .canDelete(authService.canDeleteBuild(foundBuildName, foundBuildNumber, foundBuildStarted, buildRepo))
                .buildRepo(buildWrapper.getBuildRepo())
                .sha256(getBuildSha256(buildRun, buildRepo))
                .vcs(build.getVcs())
                .build();
    }

    /**
     * Non strict since if {@param buildStarted} is not supplied we just get the latest build (some UI functions use
     * this)
     *
     * THIS METHOD ONLY CHECKS FOR BASIC-READ PERMISSIONS since the data returned by the caller requires just that.
     */
    @Nullable
    private BuildRun getBuildRunNonStrict(String buildName, String buildNumber, @Nullable String buildStarted,
            @NonNull String buildRepo) {
        BuildRun buildRun;
        if (isNotBlank(buildStarted)) {
            buildRun = getBuildRunInternally(buildName, buildNumber, buildStarted);
        } else {
            buildRun = getLatestBuildRunByNameAndNumberInternally(buildName, buildNumber, buildRepo);
        }
        return buildRun;
    }

    @Nullable
    private String getBuildSha256(@Nullable BuildRun buildRun, @NonNull String buildRepo) {
        String sha256 = null;
        if (buildRun != null) {
            RepoPath buildFilePath = BuildServiceUtils
                    .generateBuildJsonRepoPath(buildRun.getName(), buildRun.getNumber(), buildRun.getStarted(),
                            buildRepo);
            FileInfo fileInfo = repoService.getFileInfo(buildFilePath);
            if (fileInfo.getChecksumsInfo() != null) {
                sha256 = fileInfo.getChecksumsInfo().getSha256();
            }
        }
        return sha256;
    }

    private void triggerXrayBuildDelete(String buildName, String buildNumber, String buildRepo) {
        // Getting the addon for delete event. Won't send anything in case no licensed installed (coreImpl)
        XrayAddon xrayAddon = addonsManager.addonByType(XrayAddon.class);
        // Sending a delete event
        xrayAddon.callDeleteBuildInterceptor(buildName, buildNumber, buildRepo);
    }

    private void notifyInterceptors(Build build, String buildRepo, Class<? extends PluginAction> interceptor) {
        addonsManager.addonByType(PluginsAddon.class)
                .execPluginActions(interceptor, builds, new DetailedBuildRunImpl(build, buildRepo));
    }

    private boolean exists(BuildWrapper build) {
        return buildStoreService.exists(build);
    }

    /**
     * Returns an internal instance of the service
     */
    private InternalBuildService getTransactionalMe() {
        return ContextHelper.get().beanForType(InternalBuildService.class);
    }

    private List<String> getVirtualRepoKeys() {
        return repoService.getVirtualRepoDescriptors().stream()
                .map(RepoBaseDescriptor::getKey)
                .collect(Collectors.toList());
    }

    private Collector<BuildRun, ?, TreeSet<BuildRun>> sortedSetReverseStartDateCollector() {
        return Collectors.toCollection(() -> new TreeSet<>(
                Collections.reverseOrder(BuildRunComparators.getBuildStartDateComparator())));
    }

    private String userNotAuthorizedMsg() {
        return "The user: '" + authService.currentUsername() + "' is not authorized";
    }

    //** ----- Build Info migration logic ----- **\\

    @Override
    public void onContextCreated() {
        if (!buildInfoReady) {
            startBuildInfoMigrationJob();
        }
    }

    private void startBuildInfoMigrationJob() {
        Task buildInfoJob = TaskUtils.createManualTask(BuildInfoMigrationJob.class, 0L);
        List<Task> activeTasks = taskService
                .getActiveTasks(task -> task.getType().equals(BuildInfoMigrationJob.class));
        if (CollectionUtils.isNullOrEmpty(activeTasks)) {
            taskService.startTask(buildInfoJob, false);
        }
    }

    @Override
    public void migrateBuildInfo(BuildInfoCalculationWorkItem workItem) {
        BuildInfoMigrationJobDelegate delegate = workItem.getDelegate();
        long buildId = workItem.getBuildId();
        ConflictGuard lock = getConflictsGuard().getLock(workItem.getUniqueKey());
        try {
            boolean lockAcquired = lock.tryToLock(10, TimeUnit.SECONDS);
            if (lockAcquired) {
                delegate.incrementCurrentBatchCount();
                // The actual work
                delegate.migrateBuildJsonToRepo(buildId);
                delegate.incrementTotalDone();
            } else {
                throw new BuildInfoCalculationFatalException("Could not obtain lock for Build " + buildId, null);
            }
        } catch (InterruptedException e) {
            delegate.log().error("Failed to acquire lock for Build id '{}. {}'", buildId, e.getMessage());
            delegate.handleExceptionDuringMigration(buildId, e);
            Thread.currentThread().interrupt();
        } catch (BuildInfoCalculationFatalException e) {
            delegate.handleExceptionDuringMigration(buildId, e);
            delegate.log().error("Failed to migrate Build Info of Build id '{}'", buildId, e);
        } finally {
            delegate.markTaskAsFinished(Long.toString(buildId));
            lock.unlock();
        }
    }

    /**
     * Since this service must listen on any reload (to verify the build-info state) a check for whether the reload
     * was called for backup changes is made here.
     */
    private void convertIfNeeded(List<DataDiff<?>> configDiff) {
        boolean shouldConvert = configDiff.stream()
                .filter(Objects::nonNull)
                .anyMatch(diff -> diff.getFieldName().contains(CentralConfigKey.backups.getKey()));
        if (shouldConvert) {
            this.convert(null, null);
        }
    }

    @Override
    public void convert(CompoundVersionDetails source, CompoundVersionDetails target) {
        if (ArtifactoryHome.get().getCreateBackupExcludedBuildNames().exists()) {
            log.debug("Found marker file, executing conversion");
            ArtifactoryContext context = ContextHelper.get();
            CentralConfigService centralConfigService = context.beanForType(CentralConfigService.class);
            SecurityService securityService = context.beanForType(SecurityService.class);
            new BuildBackupConverter(centralConfigService, repoService, securityService).convert();
        }
    }

    /**
     * Protects the realm of Builds by locking any change to a Build json.
     * Key is BuildId.
     */
    private ConflictsGuard<Object> getConflictsGuard() {
        return addonsManager.addonByType(HaCommonAddon.class).getConflictsGuard("builds");
    }

    /**
     * TO BE USED ONLY BY THE BUILD INFO MIGRATION JOB
     * tests the old build_jsosn table (which is not created since 6.6-m001), if it exists, for the presence of the
     * build_jsons_builds_fk foreign key.
     * If either the table or the key doesn't exist then the state is considered new/ready.
     */
    @Override
    public boolean verifyBuildInfoState() {
        if (buildInfoReady) {
            //Already verified to be ok, no need to do it again
            return true;
        }
        try {
            buildInfoReady = !foreignKeyExists(jdbcHelper, dbService.getDatabaseType(), BUILD_JSONS_TABLE,
                    "build_jsons_builds_fk");
        } catch (Exception e) {
            log.error("Cannot verify existence of state-dependant foreign key 'build_jsons_builds_fk', state will " +
                    "be determined based on existence of 'build_jsons' table", e);
            try {
                // This will trigger the migration again if the table is still there (meaning as in all upgrading
                //instances)
                // but worst case it will run its course and won't do any writes.
                buildInfoReady = !tableExists(jdbcHelper, dbService.getDatabaseType(), BUILD_JSONS_TABLE);
            } catch (Exception wtf) {
                log.error(
                        "Cannot verify existence of table 'build_jsons', setting state to ready. If required, v6.6.0 " +
                                "build info repo migration will not run.",
                        wtf);
                buildInfoReady = true;
            }
        }
        return buildInfoReady;
    }

    @Override
    public Stream<ProducedBy> retrieveProducedBuilds(String sha1, String md5) {
        return buildsAql.getProducedBuilds(sha1, md5)
                .filter(aqlBuildItem -> authService
                        .isBuildBasicRead(
                                aqlBuildItem.getBuildName(),
                                aqlBuildItem.getBuildNumber(),
                                BuildInfoUtils.formatBuildTime(aqlBuildItem.getBuildStarted().getTime()),
                                aqlBuildItem.getBuildRepo()))
                .map(aqlBuildItem -> new ProducedBy(
                        aqlBuildItem.getBuildName(),
                        aqlBuildItem.getBuildNumber(),
                        aqlBuildItem.getBuildStarted().getTime(),
                        aqlBuildItem.getBuildUrl(),
                        aqlBuildItem.getBuildPromotionStatus(),
                        aqlBuildItem.getBuildModuleName()));
    }

    @Override
    public boolean isBuildExistsInDb(BuildWrapper build) {
        return buildStoreService.exists(build);
    }

    @Override
    public Stream<UsedBy> retrieveDependantBuilds(String sha1, String md5) {
        return buildsAql.getDependantBuilds(sha1, md5)
                .filter(aqlBuildItem -> authService
                        .isBuildBasicRead(
                                aqlBuildItem.getBuildName(),
                                aqlBuildItem.getBuildNumber(),
                                BuildInfoUtils.formatBuildTime((aqlBuildItem.getBuildStarted().getTime())),
                                aqlBuildItem.getBuildRepo()))
                .map(aqlBuildItem -> new UsedBy(
                        new BuildRunImpl(
                                aqlBuildItem.getBuildName(),
                                aqlBuildItem.getBuildNumber(),
                                BuildInfoUtils.formatBuildTime((aqlBuildItem.getBuildStarted().getTime())),
                                aqlBuildItem.getBuildUrl(),
                                aqlBuildItem.getBuildPromotionStatus(),
                                aqlBuildItem.getBuildRepo()
                        ),
                        aqlBuildItem.getBuildModuleName(),
                        aqlBuildItem.getBuildDependencyScope())
                );
    }

    private void validateFeatureAllowed() {
        if (addonsManager.addonByType(ArtifactRetentionAddon.class).isColdInstance()) {
            throw new BadRequestException("Build Feature is not allowed in Cold instance");
        }
    }

    private static class BuildDependenciesCloseException extends RuntimeException {
        BuildDependenciesCloseException(String message, Throwable cause) {
            super(message, cause);
        }
    }

    private static class BuildRunsProcessingException extends Exception {
        BuildRunsProcessingException(String message, Throwable cause) {
            super(message, cause);
        }
    }

    @Override
    public Map<String, RepoPath> getFileNodesBySha1(List<String> sha1List) {
        List<Node> nodes;
        try {
            nodes = nodesDao.getFileNodesBySha1(sha1List);
        } catch (SQLException e) {
            throw new StorageException("Couldn't retrieve nodes from DB", e);
        }

        return StreamSupportUtils.stream(nodes).collect(Collectors.toMap(
                Node::getSha1Actual, node -> new RepoPathImpl(node.getRepo(), node.getPath()),
                (repoPath1, repoPath2) -> repoPath1));
    }
}
