package org.artifactory.metadata.service.indexer;

import com.google.common.annotations.VisibleForTesting;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.common.ConstantValues;
import org.artifactory.concurrent.ArtifactoryExecutorFactory;
import org.artifactory.concurrent.ThreadPoolConfig;
import org.artifactory.core.api.ph.service.PackageMetadataServiceIndexHandler;
import org.artifactory.fs.ItemInfo;
import org.artifactory.md.Properties;
import org.artifactory.metadata.service.MetadataEntityFacade;
import org.artifactory.metadata.service.MetadataEventService;
import org.artifactory.model.RepoConfig;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.RepoPathFactory;
import org.jfrog.metadata.client.model.MetadataPackage;
import org.springframework.beans.factory.DisposableBean;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import javax.annotation.Nonnull;
import javax.annotation.PostConstruct;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * @author BarakH
 */
@Component
@Slf4j
public class MetadataServiceIndexHandlerImpl
        implements PackageMetadataServiceIndexHandler, InternalMetadataServiceIndexHandler, DisposableBean {

    private MetadataEntityFacade metadataEntityFacade;
    private MetadataEventService metadataEventService;
    private RepositoryService repositoryService;
    private ExecutorService executorService;
    private boolean isRealTimeIndexingEnabled;
    private final BlockingQueue<Runnable> tasksBlockingQueue = new LinkedBlockingQueue<>();
    private final AtomicInteger notifyCounter = new AtomicInteger(0);

    @PostConstruct
    void init() {
        isRealTimeIndexingEnabled = ConstantValues.metadataRealTimeIndexingEnable.getBoolean();
        if (isRealTimeIndexingEnabled) {
            this.executorService = initExecutor();
        }
    }

    @Override
    public void destroy() throws Exception {
        if (executorService != null) {
            executorService.shutdownNow();
        }
    }

    @Override
    public boolean isPackageTypeSupported(String repoPath) {
        if (!isRealTimeIndexingEnabled || StringUtils.isEmpty(repoPath)) {
            return false;
        }
        String repoKey = RepoPathFactory.create(repoPath).getRepoKey();
        RepoConfig repoConfig = repositoryService.repoConfigByKey(repoKey);
        if (repoConfig == null) {
            // Repo might have been deleted since the event created, so we return false.
            return false;
        }
        return REALTIME_INDEXING_PACKAGES.contains(repoConfig.getPackageType());
    }

    @VisibleForTesting
    boolean shouldSkipIndexing(RepoPath repoPath) {
        return !isPackageTypeSupported(repoPath.toPath()) || !metadataEntityFacade.isLeadArtifact(repoPath);
    }

    @Override
    public void deletePackageAsync(RepoPath repoPath) {
        if (shouldSkipIndexing(repoPath)) {
            return;
        }
        log.debug("Submitting async delete package task to Metadata service at {}", repoPath);
        executorService.submit(new MetadataDeletePackageTask(metadataEventService, repoPath));
        warnIfQueueIsFull();
    }

    @Override
    public void deletePackageSync(RepoPath repoPath) {
        if (shouldSkipIndexing(repoPath)) {
            return;
        }
        new MetadataDeletePackageTask(metadataEventService, repoPath).run();
    }

    @Override
    public void indexPackageAsync(ItemInfo itemInfo, Properties properties) {
        RepoPath repoPath = itemInfo.getRepoPath();
        if (shouldSkipIndexing(repoPath)) {
            return;
        }
        MetadataPackage metadataPackage = metadataEntityFacade.createMdsPackage(itemInfo, properties);
        if (metadataPackage == null) {
            log.error("Failed to create metadata package for an item at {}", repoPath);
            return;
        }
        log.debug("Submitting async create/update package task to Metadata service at {}", repoPath);
        executorService.submit(new MetadataCreatePackageTask(metadataEventService, metadataPackage, repoPath));
        warnIfQueueIsFull();
    }

    @Override
    public void indexPackageSync(ItemInfo itemInfo, Properties properties) {
        RepoPath repoPath = itemInfo.getRepoPath();
        if (shouldSkipIndexing(repoPath)) {
            return;
        }
        MetadataPackage metadataPackage = metadataEntityFacade.createMdsPackage(itemInfo, properties);
        if (metadataPackage == null) {
            log.error("Failed to create metadata package for an item at {}", repoPath);
            return;
        }
        new MetadataCreatePackageTask(metadataEventService, metadataPackage, repoPath).run();
    }

    @Nonnull
    private ExecutorService initExecutor() {
        int workersCount = ConstantValues.metadataRealTimeIndexingWorkers.getInt();
        return ArtifactoryExecutorFactory.newClassicThreadPool(ThreadPoolConfig.builder()
                .corePoolSize(workersCount)
                .maxThreads(workersCount)
                .keepAliveTimeout(5L)
                .keepAliveTimeoutTimeUnit(TimeUnit.MINUTES)
                .daemon(true)
                .queue(tasksBlockingQueue)
                .threadNameFormat("mds-index-exec-%s")
                .build(), ContextHelper.get(), true, false);
    }

    private void warnIfQueueIsFull() {
        if (notifyCounter.incrementAndGet() % 100 != 0) {
            // Check it only once in 100 tasks
            return;
        }
        notifyCounter.set(0);
        if (tasksBlockingQueue.size() > 100) {
            log.warn("Metadata service indexing Queue is full - consider adding more workers.");
        }
    }

    @Autowired
    public void setMetadataEntityFacade(MetadataEntityFacade metadataEntityFacade) {
        this.metadataEntityFacade = metadataEntityFacade;
    }

    @Autowired
    public void setRepositoryService(RepositoryService repositoryService) {
        this.repositoryService = repositoryService;
    }

    @Override
    public void setMetadataEventService(MetadataEventService metadataEventService) {
        this.metadataEventService = metadataEventService;
    }
}
