/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2020 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.artifactory.metrics.framework;

import com.jfrog.sysconf.SysConfig;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.SystemUtils;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.metrics.framework.providers.DatabaseConnectionMetricProvider;
import org.artifactory.metrics.framework.providers.GarbageCollectionConfigMetricProvider;
import org.artifactory.metrics.framework.providers.GarbageCollectionMetricProvider;
import org.artifactory.metrics.framework.providers.HttpConnectionsMetricProvider;
import org.artifactory.metrics.framework.providers.ProjectsAmountMetricProvider;
import org.artifactory.metrics.framework.providers.RuntimeMemoryMetricsProvider;
import org.artifactory.metrics.framework.providers.SlowQueriesMetricsProvider;
import org.artifactory.metrics.framework.providers.StorageMetricProvider;
import org.artifactory.projects.ProjectsService;
import org.artifactory.repo.http.HttpConnectionPoolMonitorService;
import org.artifactory.schedule.TaskService;
import org.artifactory.security.access.AccessService;
import org.artifactory.spring.ContextCreationListener;
import org.artifactory.spring.Reloadable;
import org.artifactory.storage.binstore.service.GarbageCollectionDatastoreService;
import org.artifactory.storage.db.util.JdbcHelper;
import org.artifactory.storage.service.StorageMetricCollector;
import org.jetbrains.annotations.Nullable;
import org.jfrog.client.util.PathUtils;
import org.jfrog.metrics.MetricsCollector;
import org.jfrog.metrics.MetricsHelper;
import org.jfrog.metrics.model.MetricProvider;
import org.jfrog.metrics.model.Providers;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ScheduledExecutorService;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * @author leela
 */
@Service
@Reloadable(beanClass = MetricsService.class, initAfter = TaskService.class)
public class MetricsServiceImpl implements ContextCreationListener, MetricsService {
    private static final Logger log = LoggerFactory.getLogger(MetricsServiceImpl.class);
    private static final Logger metricsLogger = LoggerFactory.getLogger("METRICS");
    private static final Logger metricEventsLogger = LoggerFactory.getLogger("METRICS_EVENTS");

    private static final String ARTIFACTORY_METRICS_ENABLED = "artifactory.metrics.enabled";
    private static final String ARTIFACTORY_METRICS_EXCLUDE = "artifactory.metrics.exclude";
    private static final String ARTIFACTORY_METRICS_INTERVAL = "artifactory.metrics.intervalSecs";
    private static final String METRICS_ELASTICSEARCH_URL = "shared.metrics.filebeat.elasticsearch.url";
    private static final String METRICS_ELASTICSEARCH_USERNAME =
            "shared.metrics.filebeat.elasticsearch.username";
    private static final String METRICS_ELASTICSEARCH_CREDENTIAL =
            "shared.metrics.filebeat.elasticsearch.password";
    private static final String METRICS_FILEBEAT_SSL_VERIFICATION_MODE =
            "shared.metrics.filebeat.elasticsearch.ssl.verification_mode";
    private static final String METRICS_FILEBEAT_BINARY = "shared.metrics.filebeat.binary";
    private static final String METRICS_FILEBEAT_CONFIGURATION =
            "shared.metrics.filebeat.configuration";
    private static final String METRICS_FILEBEAT_LOG_TO_FILES = "shared.metrics.filebeat.log.enabled";
    private static final String METRICS_FILEBEAT_LOG_LEVEL = "shared.metrics.filebeat.log.level";
    private static final String METRICS_FILEBEAT_LOG_PATH = "shared.metrics.filebeat.log.path";
    private static final String METRICS_FILEBEAT_LOG_KEEPFILES =
            "shared.metrics.filebeat.log.keepfiles";
    private static final String METRICS_FILEBEAT_ENABLED = "shared.metrics.filebeat.enabled";
    private static final String ES_USERNAME_DEFAULT = "admin";
    private static final String ES_CREDENTIAL_DEFAULT = "password";
    private static final String APP_THIRD_PARTY_FILEBEAT = "app/third-party/filebeat/";
    private static final String FILEBEAT_YAML = "var/etc/";
    private final AccessService accessService;
    private final GarbageCollectionDatastoreService gcService;
    private final HttpConnectionPoolMonitorService httpPool;
    private final JdbcHelper jdbcHelper;
    private final ProjectsService projectsService;
    private ScheduledExecutorService scheduledExecutorService;
    private boolean metricServiceRunning;
    private Process filebeatProcess;
    private MetricsCollector metricsCollector;
    private static final String EXCLUDED_METRICS = "sys_memory,sys_cpu";
    private final StorageMetricCollector storageMetricCollector;
    private final CentralConfigService configService;
    private final MetricsSharedExecutorServiceHolder executorHolder;


    @Autowired
    public MetricsServiceImpl(AccessService accessService,
            GarbageCollectionDatastoreService garbageCollectionDatastoreService,
            JdbcHelper jdbcHelper, HttpConnectionPoolMonitorService httpPoolMonitorService,
            ProjectsService projectsService, StorageMetricCollector storageMetricCollector,
            CentralConfigService centralConfigService, MetricsSharedExecutorServiceHolder executorHolder) {
        this.accessService = accessService;
        this.jdbcHelper = jdbcHelper;
        this.gcService = garbageCollectionDatastoreService;
        this.httpPool = httpPoolMonitorService;
        this.projectsService = projectsService;
        this.storageMetricCollector = storageMetricCollector;
        this.configService = centralConfigService;
        this.executorHolder = executorHolder;
    }

    @Override
    public void onContextCreated() {
        ArtifactoryHome artifactoryHome = ArtifactoryHome.get();
        List<MetricProvider> metricProviders =
                Providers.getPredefinedProviders(artifactoryHome.getVarDir().getAbsolutePath());
        addMetrics(metricProviders);
        SysConfig sysConfig = artifactoryHome.getSysConfig();

        boolean metricsEnabled = isMetricsEnabled(sysConfig);
        log.info("Metric Framework Service is enabled: {}", metricsEnabled);
        if (!metricsEnabled) {
            return;
        }
        scheduledExecutorService = executorHolder.getExecutorService();
        String excludedProvidersStr = sysConfig.get(ARTIFACTORY_METRICS_EXCLUDE, "");
        excludedProvidersStr += EXCLUDED_METRICS;
        List<String> excludedProvidersList = Stream.of(excludedProvidersStr.split(","))
                .collect(Collectors.toList());
        log.debug("Excluded metrics in Artifactory: {}", excludedProvidersList);
        long interval = 60;
        try {
            interval = sysConfig.getLong(ARTIFACTORY_METRICS_INTERVAL, 60);
        } catch (NumberFormatException e) {
            log.warn("Metrics interval configuration is not a number, hence using defaults 60 seconds as the interval");
            log.debug("Exception occurred while reading metrics interval configuration, " +
                    "hence using defaults 60 seconds as the interval", e);
        }
        MetricsHelper mHelper = new MetricsHelper(metricProviders, excludedProvidersList);
        log.debug("Final metric providers list in Artifactory, {}", mHelper.getProviders());
        metricsCollector = mHelper.start(metricsLogger, metricEventsLogger, interval, scheduledExecutorService);
        metricServiceRunning = true;
        log.info("Metric Framework service is started and is running in background...");

        boolean pushMetricsToFilebeat = sysConfig.getBoolean(METRICS_FILEBEAT_ENABLED, false);
        if (pushMetricsToFilebeat) {
            log.debug("Pushing metrics to Filebeat is enabled");
            filebeatProcess = startFilebeat(artifactoryHome, sysConfig);
        } else {
            log.debug("Filebeat is not enabled, hence ignoring to start Filebeat.");
        }
    }

    private void addMetrics(List<MetricProvider> metricProviders) {
        metricProviders.add(new GarbageCollectionMetricProvider(gcService));
        metricProviders.add(new DatabaseConnectionMetricProvider(jdbcHelper));
        metricProviders.add(new SlowQueriesMetricsProvider(jdbcHelper));
        metricProviders.add(new HttpConnectionsMetricProvider(httpPool));
        metricProviders.add(new RuntimeMemoryMetricsProvider());
        metricProviders.add(new StorageMetricProvider(storageMetricCollector));
        metricProviders.add(new GarbageCollectionConfigMetricProvider(configService));
        metricProviders.add(new ProjectsAmountMetricProvider(projectsService));
    }

    private boolean isMetricsEnabled(SysConfig sysConfig) {
        return sysConfig.getBoolean(ARTIFACTORY_METRICS_ENABLED, false) ||
                ConstantValues.metricsEnabled.getBoolean();
    }

    private Map<String, String> getEnvVarsForFilebeat(ArtifactoryHome artifactoryHome, SysConfig sysConfig) {
        String elasticsearchURL = getESUrlFromSysConfig(sysConfig);
        if (StringUtils.isBlank(elasticsearchURL)) {
            log.warn("Elasticsearch Url is not available in system.yaml, hence not starting Filebeat.");
            return null;
        }
        Map<String, String> envVars = new HashMap<>();
        envVars.put("METRICS_LOG_ENABLED", "true");
        envVars.put("NODE_ID", ContextHelper.get().getServerId());
        envVars.put("ARTIFACTORY_SERVICE_ID", accessService.getArtifactoryServiceId().getFormattedName());
        envVars.put("ELASTICSEARCH_URL", elasticsearchURL);
        envVars.put("ELASTICSEARCH_USERNAME", sysConfig.get(METRICS_ELASTICSEARCH_USERNAME, ES_USERNAME_DEFAULT));
        envVars.put("ELASTICSEARCH_PASSWORD", artifactoryHome.getMasterEncryptionWrapper()
                .decryptIfNeeded(sysConfig.get(METRICS_ELASTICSEARCH_CREDENTIAL, ES_CREDENTIAL_DEFAULT))
                .getDecryptedData());
        envVars.put("FILEBEAT_LOG_LEVEL", sysConfig.get(METRICS_FILEBEAT_LOG_LEVEL, "info"));
        envVars.put("FILEBEAT_LOG_PATH", sysConfig.get(METRICS_FILEBEAT_LOG_PATH, ""));
        envVars.put("FILEBEAT_LOG_TOFILES",
                Boolean.toString(sysConfig.getBoolean(METRICS_FILEBEAT_LOG_TO_FILES, false)));
        envVars.put("FILEBEAT_LOG_KEEPFILES", sysConfig.get(METRICS_FILEBEAT_LOG_KEEPFILES, "2"));
        envVars.put("ARTIFACTORY_HOME", artifactoryHome.getHomeDir().getAbsolutePath());
        String verificationModeInConfig = sysConfig.get(METRICS_FILEBEAT_SSL_VERIFICATION_MODE, "");
        log.debug("VerificationMode as set in config : {}", verificationModeInConfig);
        String verificationMode = verificationModeInConfig == "" ? "none" : verificationModeInConfig;
        log.debug("VerificationMode resolved as : {}", verificationMode);
        envVars.put("SSL_VERIFICATION_MODE", verificationMode);
        return envVars;
    }

    @Nullable
    String getESUrlFromSysConfig(SysConfig sysConfig) {
        return sysConfig.get(METRICS_ELASTICSEARCH_URL).map(PathUtils::addTrailingSlash).orElse(null);
    }

    /**
     * Starts Filebeat process with the given environment variables.
     */
    Process startFilebeat(ArtifactoryHome artifactoryHome, SysConfig sysConfig) {
        Process process = null;
        String startCommand = getFilebeatStartCommand(artifactoryHome, sysConfig);
        Map<String, String> env = getEnvVarsForFilebeat(artifactoryHome, sysConfig);
        if (null != env && null != startCommand) {
            try {
                ProcessBuilder builder = new ProcessBuilder("bash", "-c", startCommand, "&");
                builder.environment().putAll(env);
                log.debug("Starting Filebeat.. {}", startCommand);
                process = builder.start();
                log.debug("Filebeat started successfully.");
            } catch (IOException e) {
                log.warn("Unable to start Filebeat, hence metrics may not be available in JFrog Mission Control.", e);
            }
        }
        return process;
    }

    @Nullable
    private String getFilebeatStartCommand(ArtifactoryHome artifactoryHome, SysConfig sysConfig) {
        String fileBeatBinary = SystemUtils.IS_OS_WINDOWS ? "filebeat.exe" : "filebeat";
        String filebeatYamlPath =
                new File(artifactoryHome.getHomeDir(), FILEBEAT_YAML + "filebeat.yaml").getAbsolutePath();
        String filebeatBinaryPath =
                new File(artifactoryHome.getHomeDir(), APP_THIRD_PARTY_FILEBEAT + fileBeatBinary).getAbsolutePath();
        String elasticsearchURL = sysConfig.get(METRICS_ELASTICSEARCH_URL, "");
        if (elasticsearchURL.isEmpty()) {
            log.warn("Elasticsearch Url is empty, hence not starting Filebeat.");
            return null;
        }
        String binary = sysConfig.get(METRICS_FILEBEAT_BINARY, filebeatBinaryPath);
        String configurationPath = sysConfig.get(METRICS_FILEBEAT_CONFIGURATION, filebeatYamlPath);
        if (binary.isEmpty() || configurationPath.isEmpty()) {
            log.warn(
                    "Binary and configuration for Filebeat are empty, hence not starting Filebeat. Please set in " +
                            "system.yaml");
            return null;
        }
        return binary + " -c " + configurationPath;
    }

    @Override
    public void destroy() {
        if (null != filebeatProcess) { filebeatProcess.destroy(); }
        if (null != scheduledExecutorService) { scheduledExecutorService.shutdown(); }
        metricServiceRunning = false;
        log.info("Stopped metric service.");
    }

    @Override
    public boolean isMetricServiceRunning() {
        return metricServiceRunning;
    }

    public String getOpenMetrics() {
        boolean metricsEnabled = isMetricsEnabled(ArtifactoryHome.get().getSysConfig());
        if (!metricsEnabled) {
            return "";
        }
        return metricsCollector.openMetricsCollect();
    }
}
