package org.artifactory.metrics.framework.providers.util;

import lombok.experimental.UtilityClass;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.jfrog.metrics.model.Metric;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;

@Slf4j
@UtilityClass
public class ProviderUtil {

    private static final Pattern POOL_NAME_PATTERN = Pattern.compile("[a-zA-Z_:][a-zA-Z0-9_:]*");
    private static final Pattern NEGATIVE_POOL_NAME_PATTERN = Pattern.compile("[^a-zA-Z_:][^a-zA-Z0-9_:]*");

    private static Metric buildMetric(Map<String, Object> labels, String metricName, Metric.MetricType metricType,
            String description, Number metricValue) {
        Object pool = labels.get("pool");
        String poolName = normalizePoolName(pool);
        String name = StringUtils.isBlank(poolName) ? metricName : String.join("_", metricName, poolName);
        return Metric.builder()
                .name(name)
                .type(metricType)
                .labels(labels)
                .description(description)
                .value(metricValue)
                .build();
    }

    /**
     * The metric name specifies the general feature of a system that is measured.
     * It may contain ASCII letters and digits, as well as underscores and colons.
     * It must match the regex [a-zA-Z_:][a-zA-Z0-9_:]*.
     */
    private String normalizePoolName(Object pool) {
        if (pool == null) {
            return "";
        }
        String poolName = String.valueOf(pool);
        if (POOL_NAME_PATTERN.matcher(poolName).matches()) {
            return poolName;
        }
        log.debug("Normalizing pool Name: {}", poolName);
        poolName = NEGATIVE_POOL_NAME_PATTERN.matcher(poolName).replaceAll("_");
        log.debug("Pool normalized name: {}", poolName);
        return poolName;
    }

    public static void addMetricWithLabelsToList(List<Metric> metrics, Map<String, Object> labels, String metricName,
            Metric.MetricType metricType, String description, Number metricValue) {
        metrics.add(buildMetric(labels, metricName, metricType, description, metricValue));
    }

    public static void addMetricToList(List<Metric> metrics, String metricName, Metric.MetricType metricType,
            String description, Number metricValue) {
        Map<String, Object> labels = new HashMap<>();
        metrics.add(buildMetric(labels, metricName, metricType, description, metricValue));
    }
}
