package org.artifactory.ph;

import com.google.common.collect.Lists;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang.StringUtils;
import org.apache.http.HttpStatus;
import org.artifactory.api.security.UserInfoBuilder;
import org.artifactory.api.security.access.UserTokenSpec;
import org.artifactory.common.crypto.CryptoHelper;
import org.artifactory.core.api.ph.exception.PackageException;
import org.artifactory.core.api.ph.model.SimpleUser;
import org.artifactory.core.api.ph.service.PackageSecurityService;
import org.artifactory.repo.RepoPathFactory;
import org.artifactory.security.AuthenticationHelper;
import org.artifactory.security.InternalSecurityService;
import org.artifactory.security.MutableUserInfo;
import org.artifactory.security.UserInfo;
import org.artifactory.security.access.AccessService;
import org.artifactory.security.access.AccessTokenAuthentication;
import org.artifactory.storage.security.service.UserGroupStoreService;
import org.springframework.security.authentication.AuthenticationManager;
import org.springframework.security.authentication.BadCredentialsException;
import org.springframework.security.authentication.UsernamePasswordAuthenticationToken;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.security.core.userdetails.UserDetails;

import java.util.concurrent.Callable;

@Slf4j
public class PackageSecurityServiceImpl implements PackageSecurityService {

    public static final String FAILED_TO_AUTHENTICATE_LOCALLY = "Failed to authenticate locally.";
    public static final String LOCAL_TOKEN_CREATION_FAILED = "Local token creation failed";
    private final InternalSecurityService securityService;
    private final AccessService accessService;
    private final AuthenticationManager authenticationManager;
    private final UserGroupStoreService userGroupStoreService;

    public PackageSecurityServiceImpl(InternalSecurityService securityService, AccessService accessService,
            AuthenticationManager authenticationManager,
            UserGroupStoreService userGroupStoreService) {
        this.securityService = securityService;
        this.accessService = accessService;
        this.authenticationManager = authenticationManager;
        this.userGroupStoreService = userGroupStoreService;
    }

    @Override
    public boolean canRead(String repoKey, String path) {
        return securityService.canRead(RepoPathFactory.create(repoKey, path));
    }

    @Override
    public boolean canWrite(String repoKey, String path) {
        return securityService.canDeploy(RepoPathFactory.create(repoKey, path));
    }

    @Override
    public boolean canAnnotate(String repoKey, String path) {
        return securityService.canAnnotate(RepoPathFactory.create(repoKey, path));
    }

    @Override
    public boolean canDelete(String repoKey, String path) {
        return securityService.canDelete(RepoPathFactory.create(repoKey, path));
    }

    @Override
    public boolean canManage(String repoKey, String path) {
        return securityService.canManage(RepoPathFactory.create(repoKey, path));
    }

    @Override
    public <V> V callAsSystem(Callable<V> callable) throws PackageException {
        try {
            return securityService.callAsSystem(callable);
        } catch (PackageException e) {
            throw e;
        } catch (Exception e) {
            throw new PackageException(HttpStatus.SC_INTERNAL_SERVER_ERROR, e.getMessage(), e.getMessage(), e);
        }
    }

    @Override
    public boolean isAnonymous() {
        return securityService.isAnonymous();
    }

    @Override
    public boolean isAnonAccessEnabled() {
        return securityService.isAnonAccessEnabled();
    }

    @Override
    @NonNull
    public String getCurrentUserName() {
        return securityService.currentUsername();
    }

    @Override
    public boolean isAuthenticated() {
        return securityService.isAuthenticated();
    }

    @Override
    @NonNull
    public String loginAndGenerateAccessTokenWithBasicAuth(String userName, String password)
            throws PackageException {
        Authentication originalAuth = SecurityContextHolder.getContext().getAuthentication();
        UsernamePasswordAuthenticationToken authenticationToken =
                new UsernamePasswordAuthenticationToken(userName, password);
        try {
            Authentication authentication = authenticationManager.authenticate(authenticationToken);
            if (authentication instanceof AccessTokenAuthentication accessTokenAuthentication) {
                // user came with access token - we should use the same token
                return accessTokenAuthentication.getAccessTokenValue();
            }
            // Adding the authentication so Access will get it for token
            SecurityContextHolder.getContext().setAuthentication(authentication);
            return generateAccessTokenForUser(userName);
        } catch (BadCredentialsException e) {
            log.warn(FAILED_TO_AUTHENTICATE_LOCALLY + ":{}", e.getMessage());
            log.debug(FAILED_TO_AUTHENTICATE_LOCALLY + " for user: '{}'", userName, e);
            throw new PackageException(HttpStatus.SC_UNAUTHORIZED, FAILED_TO_AUTHENTICATE_LOCALLY,
                    FAILED_TO_AUTHENTICATE_LOCALLY);
        } catch (Exception e) {
            log.error(FAILED_TO_AUTHENTICATE_LOCALLY + ":{}", e.getMessage());
            log.debug(FAILED_TO_AUTHENTICATE_LOCALLY + " for user: '{}'", userName, e);
            throw new PackageException(HttpStatus.SC_BAD_REQUEST, FAILED_TO_AUTHENTICATE_LOCALLY,
                    FAILED_TO_AUTHENTICATE_LOCALLY);
        } finally {
            SecurityContextHolder.getContext().setAuthentication(originalAuth);
        }
    }

    @Override
    public void revokeToken(String token) throws PackageException {
        try {
            accessService.revokeToken(token);
        } catch (Exception e) {
            String msg = "Logout locally failed";
            log.warn(msg + ": error on token revoke.", e.getMessage());
            log.debug(msg + ": error on token revoke of: {}", token, e);
            throw new PackageException(HttpStatus.SC_BAD_REQUEST, msg, msg);
        }
    }

    @Override
    public SimpleUser getCurrentUserDetails() throws PackageException {
        Authentication authentication = AuthenticationHelper.getAuthentication();
        if ((authentication == null) || !authentication.isAuthenticated()) {
            String msg = "User is unauthenticated";
            throw new PackageException(HttpStatus.SC_BAD_REQUEST, msg, msg);
        }
        String authUsername = ((UserDetails) authentication.getPrincipal()).getUsername();
        String password = (String) authentication.getCredentials();
        if (StringUtils.isNotBlank(password)) {
            // Check if the user is an Access Token user
            if (authentication instanceof AccessTokenAuthentication accessTokenAuthentication) {
                String username = accessService
                        .extractSubjectUsername(accessTokenAuthentication.getAccessToken());
                MutableUserInfo transientUser = new UserInfoBuilder(username).transientUser().build();
                return new SimpleUser(transientUser.getUsername(), password, transientUser.getEmail());
            }
            UserInfo user = userGroupStoreService.findUser(authUsername);
            if (user != null) {
                if (!CryptoHelper.isApiKey(password)) {
                    password = securityService.createEncryptedPasswordIfNeeded(user, password);
                }
                return new SimpleUser(authUsername, password, user.getEmail());
            }
        }
        String msg = "The user: '" + authentication.getPrincipal() + "' couldn't be found";
        throw new PackageException(HttpStatus.SC_BAD_REQUEST, msg, msg);
    }

    /**
     * Generating Access token for the user with following specs: no Expiry, non-Refreshable, no groups' limitation.
     * We set the scope to be "member-of-groups:*" since it will allow the user to obtain groups access according
     * to his group memberships AT THE TIME OF EACH LOGIN.
     */
    private String generateAccessTokenForUser(String username) throws PackageException {
        // Creating the token to be valid forever + with no specific groups limitation
        try {
            UserTokenSpec tokenSpec = UserTokenSpec.create(username)
                    .noExpiry()
                    .refreshable(false)
                    .scope(Lists.newArrayList("member-of-groups:*"));
            return accessService.createToken(tokenSpec, true, false).getTokenValue();
        } catch (Exception e) {
            log.warn(LOCAL_TOKEN_CREATION_FAILED + ": {}", e.getMessage());
            log.debug("Local token creation failed for user: '{}'", username, e);
            throw new PackageException(HttpStatus.SC_BAD_REQUEST, LOCAL_TOKEN_CREATION_FAILED,
                    LOCAL_TOKEN_CREATION_FAILED);
        }
    }
}
