package org.artifactory.queue.register;

import com.google.common.annotations.VisibleForTesting;
import com.google.common.collect.ArrayListMultimap;
import com.google.common.collect.Lists;
import com.google.common.collect.Multimap;
import com.google.common.collect.Multimaps;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.collections.SetUtils;
import org.artifactory.api.config.RepositoryConfigService;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.common.repo.PackageType;
import org.artifactory.common.repo.RepoType;
import org.artifactory.config.CentralConfigKey;
import org.artifactory.event.queue.QueueConfiguration;
import org.artifactory.event.queue.QueueConfigurationRegistrar;
import org.artifactory.event.queue.QueueCursorType;
import org.artifactory.event.queue.QueueOperations;
import org.artifactory.model.LocalCacheRepoConfig;
import org.artifactory.model.LocalConfig;
import org.artifactory.model.RemoteRepoConfig;
import org.artifactory.model.RepoConfig;
import org.artifactory.model.VirtualRepoConfig;
import org.artifactory.queue.services.PersistentQueue;
import org.artifactory.queue.services.PersistentQueueImpl;
import org.artifactory.repo.RepoConfigChange;
import org.artifactory.repo.RepoConfigReload;
import org.artifactory.repo.RepoDetailsType;
import org.artifactory.repo.service.InternalRepositoryService;
import org.artifactory.spring.Reloadable;
import org.springframework.stereotype.Component;

import javax.annotation.Nullable;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static org.artifactory.repo.system.SystemRepos.isSupportBundle;
import static org.artifactory.util.RepoPathUtils.isCache;
import static org.artifactory.util.RepoPathUtils.isTrash;

/**
 * The registration service is responsible to create and remove all PersistentQueue's instances.
 * an instance of PersistentQueue is being created for each repoKey+cursorType+variant<br>
 * e.g:<br>
 * repoNames = my-local-maven1 ,my-local-maven2 ,my-virtual-maven1
 * cursorType = Incremental-index, replication(var1,var2)
 * will create this 9 PersistentQueues instances:
 * -  my-local-maven1-incremental-index-queue
 * -  my-local-maven2-incremental-index-queue
 * -  my-virtual-maven1-incremental-index-queue
 * -  my-local-maven1-replication-var1-queue
 * -  my-local-maven1-replication-var2-queue
 * -  my-local-maven2-replication-var1-queue
 * -  my-local-maven2-replication-var2-queue
 * -  my-virtual-maven1-replication-var1--queue
 * -  my-virtual-maven1-replication-var2--queue
 *
 * The creation process is being done on start up (init method) and whenever a new repository is added
 * (registerRepository method) - by collecting all QueueConfigurationRegistrar implementations.
 * and by application code using the QueueService.register method which accept a single
 * QueueConfiguration without the need to create a Registrar implementation.
 *
 * the removal of a queue is being done whenever a repository is deleted (deRegisterRepository method)
 *
 * @author Avishay Halpren
 */
@Slf4j
@Component
@Reloadable(beanClass = PersistentQueueRegistrationService.class, initAfter = {InternalRepositoryService.class},
        listenOn = {CentralConfigKey.none})
public class PersistentQueueRegistrationServiceImpl implements PersistentQueueRegistrationService {

    /**
     * Stores all the queues created by this service.
     * the key is repoKey + cursorType for multiple (variant based) queues
     */
    private final Multimap<RepoAndCursor, PersistentQueue> queues =
            Multimaps.synchronizedListMultimap(ArrayListMultimap.create());
    /**
     * this lock defends modifications of queues multimap which is not thread safe.
     */
    private final ReentrantReadWriteLock readWriteLock = new ReentrantReadWriteLock();

    private final List<QueueConfigurationRegistrar> registrars;
    private final RepositoryService repositoryService;
    private final RepositoryConfigService repositoryConfigService;

    public PersistentQueueRegistrationServiceImpl(List<QueueConfigurationRegistrar> registrars,
            RepositoryService repositoryService,
            RepositoryConfigService repositoryConfigService) {
        this.registrars = registrars;
        this.repositoryService = repositoryService;
        this.repositoryConfigService = repositoryConfigService;
    }

    @Override
    public void init() {
        //register all existing repositories on startup
        repositoryService.getAllRepoKeys().forEach(repoKey -> registerRepository(repoKey, false));
    }

    @Override
    public void onRepoConfigReload(Map<RepoDetailsType, List<RepoConfigReload>> repoConfigReload) {
        //This method does not support update.
        // new repos
        repoConfigReload.values().stream()
                .flatMap(Collection::stream)
                .filter(change -> change.getConfigAction().equals(RepoConfigChange.ConfigAction.REPO_CREATE))
                .forEach(change -> registerRepository(change.getRepoKey(), true));

        // deleted repos
        repoConfigReload.values().stream()
                .flatMap(Collection::stream)
                .filter(change -> change.getConfigAction().equals(RepoConfigChange.ConfigAction.REPO_DELETE))
                .forEach(change -> deRegisterRepository(change.getRepoKey()));
    }

    @Override
    public void register(String repoKey, QueueConfiguration configuration) {
        createQueue(repoKey, false, configuration);
    }

    @Override
    public void registerRepository(String repoKey) {
        registerRepository(repoKey, true);
    }

    /**
     * This method is called in the init phase for all active repositories and when creating a new repository.
     *
     * @param newRepository determine if registering an existing repository or a new one
     */
    private void registerRepository(String repoKey, boolean newRepository) {
        if (registrars.isEmpty()) {
            //nothing to do if no registrars
            return;
        }
        RepoConfig repoConfig;
        if (isTrash(repoKey) || isSupportBundle(repoKey) || isCache(repoKey)) {
            repoConfig = repositoryService.repoConfigByKey(repoKey);
        } else {
            repoConfig = repositoryConfigService.getRepoConfigByKey(repoKey);
        }
        if (repoConfig == null) {
            return;
        }

        Set<RepoType> repoTypes = getRepoTypes(repoConfig);
        if (!Objects.isNull(repoTypes)) {
            PackageType packageType = getPackageType(repoConfig);
            //collect all configurations for a repoKey
            List<QueueConfiguration> configurations = registrars.stream()
                    .filter(registrar -> shouldCreateQueue(packageType, repoTypes, registrar))
                    .map(registrar -> registrar.getConfigurations(repoKey))
                    .flatMap(Collection::stream)
                    .collect(Collectors.toList());
            for (QueueConfiguration queueConfiguration : configurations) {
                try {
                    if (!isRegistered(repoKey, queueConfiguration.getCursorType(), queueConfiguration.getVariant())) {
                        boolean resetMarker = newRepository && queueConfiguration.isResetEventMarkerOnCreation();
                        createQueue(repoKey, resetMarker, queueConfiguration);
                    }
                } catch (Exception e) {
                    log.error("An error occurred while trying to register repo '{}' to queue: {}", repoKey,
                            e.getMessage());
                    log.debug("", e);
                }
            }
        }
    }

    @VisibleForTesting
    public void createQueue(String repoKey, boolean resetMarker, QueueConfiguration queueConfiguration) {
        QueueCursorType cursorType = queueConfiguration.getCursorType();
        //create the new queue and register it
        PersistentQueue persistentQueue = new PersistentQueueImpl(queueConfiguration, repoKey);
        try {
            readWriteLock.writeLock().lock();
            RepoAndCursor key = new RepoAndCursor(repoKey, cursorType);
            //before inserting to the queue - check we dont have the same name already:
            //can't have multiple queues with same name
            //e.g: only 1 ("my-local-maven-repo",IncrementalIndex) queue
            //if we want more than 1 we must use different variants
            //e.g: ("my-local-maven-repo",Replication,var1) queue
            //e.g: ("my-local-maven-repo",Replication,var2) queue
            //this is because a repoKey a variant and a cursorType defines the name of the cursor in the queue.
            //we cannot have 2 different queues working on the same cursor!
            Collection<PersistentQueue> persistentQueues = queues.get(key);
            if (persistentQueues.stream().anyMatch(q -> q.getName().equals(persistentQueue.getName()))) {
                throw new IllegalStateException(
                        String.format("found queue with same name %s", persistentQueue.getName()));
            }
            // Create the marker
            persistentQueue.createMarkerIfNotExists();
            if (resetMarker) {
                //whenever a queue for a new repository is created we want to make sure to reset any old marker if
                //exist. The reset method takes care of the cursor and set it to the last event of the queue.
                //This can prevent problems if the original reset(true) call for this queue failed during triggered
                //if we don't do this we could start a new queue with leftover events of a previous queue with same
                //name.
                persistentQueue.reset();
            }
            // Store the qq
            queues.put(key, persistentQueue);
        } finally {
            readWriteLock.writeLock().unlock();
        }
        log.debug("Queue created, name: {}, repo: {}", persistentQueue.getName(), repoKey);
    }

    /**
     * Check if the given repository is matching to the configuration requirements for queue creation.
     */
    private boolean shouldCreateQueue(PackageType packageType, Set<RepoType> repoTypes,
            QueueConfigurationRegistrar configuration) {

        boolean packageTypeMatch =
                Objects.isNull(configuration.getPackageType()) ||
                        Objects.equals(configuration.getPackageType(), packageType);

        boolean repoTypeMatch = CollectionUtils.isEmpty(configuration.getRepoTypes())
                || SetUtils.isEqualSet(configuration.getRepoTypes(), repoTypes);

        return packageTypeMatch && repoTypeMatch;
    }

    private PackageType getPackageType(RepoConfig repoConfig) {
        try {
            return PackageType.valueOf(repoConfig.getPackageType().getType().toUpperCase());
        } catch (Exception e) {
            return null;
        }
    }

    @Nullable
    private Set<RepoType> getRepoTypes(RepoConfig repoConfig) {
        if (repoConfig instanceof LocalCacheRepoConfig) {
            return Set.of(RepoType.REMOTE, RepoType.LOCAL);
        } else if (repoConfig instanceof LocalConfig) {
            return Set.of(RepoType.LOCAL);
        } else if (repoConfig instanceof RemoteRepoConfig) {
            return Set.of(RepoType.REMOTE);
        } else if (repoConfig instanceof VirtualRepoConfig) {
            return Set.of(RepoType.VIRTUAL);
        }
        return null;
    }

    @Override
    public void deRegister(String repoKey, QueueCursorType type, String variant) {
        findQueue(repoKey, type, variant).ifPresentOrElse(queue -> {
            try {
                readWriteLock.writeLock().lock();
                queues.remove(new RepoAndCursor(repoKey, type), queue);
                log.info("Queue removed, name: {}, repo: {}", queue.getName(), repoKey);
            } finally {
                readWriteLock.writeLock().unlock();
            }
            queue.shutdown();
            queue.reset(repoKey, true);
        }, () -> log.debug("Cannot found a queue for {} {} {}", repoKey, type, variant));
    }

    @Override
    public void deRegisterRepository(String repoKey) {
        //this method will be called when deleting a repository
        log.debug("Removing all queues for repo {}", repoKey);
        List<PersistentQueue> removedQueues;
        try {
            readWriteLock.writeLock().lock();
            //for each relevant log types try to remove from the map
            //collect the removed queues
            removedQueues = Stream.of(QueueCursorType.values())
                    .map(type -> queues.removeAll(new RepoAndCursor(repoKey, type)))
                    .flatMap(Collection::stream)
                    .collect(Collectors.toList());
        } finally {
            readWriteLock.writeLock().unlock();
        }
        if (!removedQueues.isEmpty()) {
            log.info("Unregistering {} queues for repo {}", removedQueues.size(), repoKey);
            //call shutdown on all removed queues - signals all jobs to stop
            removedQueues.forEach(PersistentQueue::shutdown);
            //call the reset on all removed queues - make sure marker is deleted
            //deleting the marker will allow the events cleanup job to do his work.
            removedQueues.forEach(queue -> queue.reset(repoKey, true));
        } else {
            log.debug("No queues found for {} ", repoKey);
        }
    }

    @Override
    public Collection<PersistentQueue> getQueues(String repoKey, QueueCursorType cursorType) {
        try {
            readWriteLock.readLock().lock();
            //Wrap the result in a new List to avoid concurrent modification exception
            return Lists.newArrayList(queues.get(new RepoAndCursor(repoKey, cursorType)));
        } finally {
            readWriteLock.readLock().unlock();
        }
    }

    @Override
    public Collection<PersistentQueue> getQueues(Set<String> reposKey) {
        try {
            readWriteLock.readLock().lock();
            return queues.entries().stream()
                    .filter(entry -> reposKey.contains(entry.getKey().getRepoKey()))
                    .map(Map.Entry::getValue)
                    .collect(Collectors.toList());
        } finally {
            readWriteLock.readLock().unlock();
        }
    }

    @Override
    public Collection<PersistentQueue> getAllQueues() {
        try {
            readWriteLock.readLock().lock();
            return queues.values();
        } finally {
            readWriteLock.readLock().unlock();
        }
    }

    @Override
    public boolean isRegistered(String repoKey, QueueCursorType type, String variant) {
        return findQueue(repoKey, type, variant).isPresent();
    }

    @Override
    public void resetQueueIfExist(String repoKey, QueueCursorType type, String variant) {
        try {
            if (isRegistered(repoKey, type, variant)) {
                getQueueOperations(repoKey, type, variant).reset();
            }
        } catch (Exception e) {
            log.error("Failed to reset queue for repo {} with an error: {}", repoKey, e.getMessage());
            log.debug("", e);
        }
    }

    @Override
    public QueueOperations getQueueOperations(String repoKey, QueueCursorType type, String variant) {
        return findQueue(repoKey, type, variant).orElseThrow(() -> new RuntimeException(
                String.format("Couldn't find Queue for %s %s %s", type, repoKey, variant)));
    }

    private Optional<PersistentQueue> findQueue(String repoKey, QueueCursorType type, String variant) {
        try {
            readWriteLock.readLock().lock();
            Collection<PersistentQueue> persistentQueues = queues.get(new RepoAndCursor(repoKey, type));
            return persistentQueues.stream()
                    .filter(q -> q.getQueueConfiguration().getVariant().equals(variant))
                    .findFirst();
        } finally {
            readWriteLock.readLock().unlock();
        }
    }

    @Override
    public void destroy() {
        readWriteLock.writeLock().lock();
        try {
            queues.values().forEach(PersistentQueue::shutdown);
            queues.clear();
        } finally {
            readWriteLock.writeLock().unlock();
        }
    }
}
