/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.repo.interceptor.config;

import com.google.common.collect.Lists;
import org.artifactory.addon.AddonsManager;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.config.RepositoryConfigService;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.entitlement.EntitlementType;
import org.artifactory.model.LocalRepoConfig;
import org.artifactory.repo.RepoConfigChange;
import org.artifactory.repo.RepoDetailsType;
import org.artifactory.repo.interceptor.RepoConfigChangeInterceptor;
import org.artifactory.repo.system.suppliers.EdgeRepo;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.List;
import java.util.Map;

import static org.artifactory.repo.RepoConfigChange.ConfigAction.REPO_CREATE;
import static org.artifactory.repo.RepoDetailsType.LOCAL;
import static org.artifactory.repo.system.SystemRepos.EDGE_UPLOADS_REPO_KEY;

/**
 * @author Fred Simon
 */
@Component
public class EdgeUploadReposConfigInterceptor implements RepoConfigChangeInterceptor {
    private AddonsManager addonsManager;
    private RepositoryConfigService repositoryConfigService;
    private CentralConfigService centralConfigService;

    @Autowired
    public EdgeUploadReposConfigInterceptor(AddonsManager addonsManager,
            CentralConfigService centralConfigService) {
        this.addonsManager = addonsManager;
        this.centralConfigService = centralConfigService;
    }

    @Autowired
    public void setRepositoryConfigService(RepositoryConfigService repositoryConfigService) {
        this.repositoryConfigService = repositoryConfigService;
    }

    @Override
    public void onBeforeSave(Map<RepoDetailsType, List<RepoConfigChange>> mutableRepoConfigChanges) {
        if (addonsManager.isEntitled(EntitlementType.EDGE) &&
                !repositoryConfigService.exists(EDGE_UPLOADS_REPO_KEY)) {
            CentralConfigDescriptor centralConfigDescriptor = centralConfigService.getDescriptor();
            List<RepoConfigChange> localRepoConfigChanges =
                    mutableRepoConfigChanges.getOrDefault(LOCAL, Lists.newArrayList());
            boolean aboutToBeCreated = localRepoConfigChanges.stream()
                    .filter(repoConfigChange -> REPO_CREATE.equals(repoConfigChange.getConfigAction()))
                    .anyMatch(repoConfigChange -> EDGE_UPLOADS_REPO_KEY.equals(repoConfigChange.getRepoKey()));
            if (aboutToBeCreated) {
                return;
            }
            LocalRepoConfig edgeUploadsRepoConfig =
                    new EdgeRepo().createSystemRepoConfig(centralConfigDescriptor, EDGE_UPLOADS_REPO_KEY);

            localRepoConfigChanges
                    .add(new RepoConfigChange(EDGE_UPLOADS_REPO_KEY, REPO_CREATE, edgeUploadsRepoConfig, null));
            mutableRepoConfigChanges.put(LOCAL, localRepoConfigChanges);
        }
    }

    @Override
    public void onAfterSave(Map<RepoDetailsType, List<RepoConfigChange>> repoConfigChanges) {
        if (repoConfigChanges.getOrDefault(LOCAL, List.of()).stream()
                .filter(repoConfigChange -> REPO_CREATE.equals(repoConfigChange.getConfigAction()))
                .anyMatch(repoConfigChange -> EDGE_UPLOADS_REPO_KEY.equals(repoConfigChange.getRepoKey()))) {
            LocalRepoDescriptor edgeUploadsRepoDescriptor = new LocalRepoDescriptor();
            edgeUploadsRepoDescriptor.setKey(EDGE_UPLOADS_REPO_KEY);
            centralConfigService.getMutableDescriptor().conditionallyAddToBackups(edgeUploadsRepoDescriptor.getKey());
        }
    }
}
