package org.artifactory.repo.migration.mappers;

import lombok.experimental.UtilityClass;
import lombok.extern.slf4j.Slf4j;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.descriptor.repo.RepositoryConfigSource;
import org.artifactory.model.BaseConfig;
import org.artifactory.model.SecurityConfig;
import org.artifactory.model.XrayRepoConfig;
import org.jetbrains.annotations.NotNull;

import java.util.List;
import java.util.Map;

import static org.artifactory.descriptor.repo.RepoType.fromType;

@Slf4j
@UtilityClass
public class RepoConfigMigrationPatchUtils {
    public static void buildBaseConfig(@NotNull BaseConfig baseConfig, Map<String, Object> newValues,
            CentralConfigDescriptor descriptor) {
        for (Map.Entry<String, Object> newValue : newValues.entrySet()) {
            try {
                if ("description".equals(newValue.getKey())) {
                    baseConfig.setDescription((String) newValue.getValue());
                } else if ("notes".equals(newValue.getKey())) {
                    baseConfig.setNotes((String) newValue.getValue());
                } else if ("repoLayout".equals(newValue.getKey())) {
                    String repoLayoutRef = (String) newValue.getValue();
                    if (descriptor.getRepoLayout(repoLayoutRef) == null) {
                        log.debug("Repo layout '{}' does not exist in the config descriptor", repoLayoutRef);
                    } else {
                        baseConfig.setRepoLayoutRef(repoLayoutRef);
                    }
                } else if ("includesPattern".equals(newValue.getKey())) {
                    baseConfig.setIncludesPattern((String) newValue.getValue());
                } else if ("excludesPattern".equals(newValue.getKey())) {
                    baseConfig.setExcludesPattern((String) newValue.getValue());
                }
            } catch (ClassCastException e) {
                throw new IllegalArgumentException("No values under " + newValue.getKey());
            }
        }
    }

    @NotNull
    public static SecurityConfig buildSecurityConfig(SecurityConfig securityConfig, Map<String, Object> newValues,
            CentralConfigDescriptor descriptor) {
        for (Map.Entry<String, Object> newValue : newValues.entrySet()) {
            try {
                if ("xray".equals(newValue.getKey())) {
                    XrayRepoConfig xrayRepoConfig = new XrayRepoConfig();
                    if (securityConfig.getXrayConfig() != null) {
                        xrayRepoConfig = securityConfig.getXrayConfig();
                    }
                    Map<String, Object> newXrayRepoConfig = (Map<String, Object>) newValue.getValue();
                    if (newXrayRepoConfig.get("enabled") != null) {
                        xrayRepoConfig.setEnabled((Boolean) newXrayRepoConfig.get("enabled"));
                    }
                    if (newXrayRepoConfig.get("dataTtl") != null) {
                        xrayRepoConfig.setDataTtl((Integer) newXrayRepoConfig.get("dataTtl"));
                    }
                    securityConfig.setXrayConfig(xrayRepoConfig);
                } else if ("primaryKeyPairDescriptor".equals(newValue.getKey())) {
                    String keyPairRef = (String) newValue.getValue();
                    if (descriptor.getKeyPair(keyPairRef) == null) {
                        log.debug("Primary key pair '{}' does not exist in the config descriptor", keyPairRef);
                    } else {
                        securityConfig.setPrimaryKeyPairRef((String) newValue.getValue());
                    }
                } else if ("secondaryKeyPairDescriptor".equals(newValue.getKey())) {
                    String keyPairRef = (String) newValue.getValue();
                    if (descriptor.getKeyPair(keyPairRef) == null) {
                        log.debug("Secondary key pair '{}' does not exist in the config descriptor", keyPairRef);
                    } else {
                        securityConfig.setSecondaryKeyPairRef((String) newValue.getValue());
                    }
                } else if ("signedUrlTtl".equals(newValue.getKey())) {
                    securityConfig.setSignedUrlTtl((long) newValue.getValue());
                } else if ("keyPair".equals(newValue.getKey())) {
                    securityConfig.setKeyPair(String.valueOf(newValue.getValue()));
                }
            } catch (ClassCastException e) {
                throw new IllegalArgumentException("No values under " + newValue.getKey());
            }
        }
        return securityConfig;
    }

    public static void externalDependenciesToConfig(@NotNull Map<String, String> packageTypeConfig,
            Map<String, Object> newValues) {
        if (newValues.get("externalDependencies") != null) {
            Map<String, Object> externalDependenciesNewValues =
                    (Map<String, Object>) newValues.get("externalDependencies");
            for (Map.Entry<String, Object> externalDependenciesNewValue :
                    externalDependenciesNewValues.entrySet()) {
                if ("enabled".equals(externalDependenciesNewValue.getKey())) {
                    packageTypeConfig.put("externalDependenciesEnabled",
                            String.valueOf(externalDependenciesNewValue.getValue()));
                } else if ("patterns".equals(externalDependenciesNewValue.getKey())) {
                    Map<String, Object> properties =
                            (Map<String, Object>) externalDependenciesNewValue.getValue();
                    List<String> patternsList = properties.entrySet().stream()
                            .sorted(Map.Entry.comparingByKey())
                            .map(Map.Entry::getValue)
                            .map(String.class::cast)
                            .toList();
                    RepositoryConfigSource.entityToFlatMapKeys(patternsList, "externalDependenciesPatterns",
                            packageTypeConfig);
                }
            }
        }
    }

    public static void buildLocalPackageTypeConfig(Map<String, String> packageTypeConfig,
            RepoType repoType,
            Map<String, Object> newValues) {
        for (Map.Entry<String, Object> newValue : newValues.entrySet()) {
            try {
                if (repoType.isMavenGroup()) {
                    addValueToPackageTypeConfigIfNeeded("suppressPomConsistencyChecks", packageTypeConfig, newValue);
                    addValueToPackageTypeConfigIfNeeded("forceMavenAuthentication", packageTypeConfig, newValue);
                } else if (RepoType.Cargo.equals(repoType)) {
                    addValueToPackageTypeConfigIfNeeded("cargoAnonymousAccess", packageTypeConfig, newValue);
                } else if (RepoType.Composer.equals(repoType)) {
                    addValueToPackageTypeConfigIfNeeded("enableComposerV1Indexing", packageTypeConfig, newValue);
                } else if (RepoType.Debian.equals(repoType)) {
                    addValueToPackageTypeConfigIfNeeded("debianTrivialLayout", packageTypeConfig, newValue);
                    addValueToPackageTypeConfigIfNeeded("ddebSupported", packageTypeConfig, newValue);
                } else if (RepoType.Docker.equals(repoType)) {
                    addValueToPackageTypeConfigIfNeeded("dockerTagRetention", packageTypeConfig, newValue);
                    addValueToPackageTypeConfigIfNeeded("dockerApiVersion", packageTypeConfig, newValue);
                    addValueToPackageTypeConfigIfNeeded("blockPushingSchema1", packageTypeConfig, newValue);
                    addValueToPackageTypeConfigIfNeeded("maxUniqueTags", packageTypeConfig, newValue);
                } else if (RepoType.NuGet.equals(repoType)) {
                    addValueToPackageTypeConfigIfNeeded("forceNugetAuthentication", packageTypeConfig, newValue);
                } else if (RepoType.YUM.equals(fromType(repoType.getType()))) {
                    addValueToPackageTypeConfigIfNeeded("calculateYumMetadata", packageTypeConfig, newValue);
                    addValueToPackageTypeConfigIfNeeded("yumRootDepth", packageTypeConfig, newValue);
                    addValueToPackageTypeConfigIfNeeded("enableFileListsIndexing", packageTypeConfig, newValue);
                    addValueToPackageTypeConfigIfNeeded("yumGroupFileNames", packageTypeConfig, newValue);
                } else if (RepoType.Terraform.equals(repoType)) {
                    addValueToPackageTypeConfigIfNeeded("terraformType", packageTypeConfig, newValue);
                } else if (RepoType.Conan.equals(repoType)) {
                    addValueToPackageTypeConfigIfNeeded("forceConanAuthentication", packageTypeConfig, newValue);
                }
            } catch (ClassCastException e) {
                throw new IllegalArgumentException("No values under " + newValue.getKey());
            }
        }
    }

    private static void addValueToPackageTypeConfigIfNeeded(String key, Map<String, String> packageTypeConfig,
            Map.Entry<String, Object> entry) {
        if (key.equals(entry.getKey())) {
            packageTypeConfig.put(key, String.valueOf(entry.getValue()));
        }
    }
}
