package org.artifactory.repo.virtual;

import com.google.common.collect.Lists;
import org.artifactory.addon.AddonsManager;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RealRepoDescriptor;
import org.artifactory.descriptor.repo.RemoteRepoDescriptor;
import org.artifactory.descriptor.repo.VirtualRepoResolver;
import org.artifactory.entitlement.EntitlementType;
import org.artifactory.repo.LocalCacheRepo;
import org.artifactory.repo.LocalRepo;
import org.artifactory.repo.RealRepo;
import org.artifactory.repo.RemoteRepo;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.service.InternalRepositoryService;
import org.artifactory.request.RepoRequests;

import java.util.List;

public class VirtualRepositoriesResolver {

    private final VirtualRepoResolver resolver;
    private final InternalRepositoryService repositoryService;
    private final VirtualRepo virtualRepo;
    private final AddonsManager addonsManager;
    private final CentralConfigService centralConfigService;
    private final PathTranslator pathTranslator;
    VirtualRepositoriesResolver(VirtualRepo virtualRepo) {
        this.virtualRepo = virtualRepo;
        ArtifactoryContext artifactoryContext = ContextHelper.get();
        repositoryService = artifactoryContext.beanForType(InternalRepositoryService.class);
        centralConfigService = artifactoryContext.beanForType(CentralConfigService.class);
        addonsManager = artifactoryContext.beanForType(AddonsManager.class);
        resolver = new VirtualRepoResolver(virtualRepo.getDescriptor());
        pathTranslator = artifactoryContext.beanForType(PathTranslator.class);
    }

    public List<LocalRepo<?, ?>> getResolvedLocalAndFederatedRepos() {
        //Add all local repositories
        List<LocalRepo<?, ?>> repositories = Lists.newArrayList();
        for (LocalRepoDescriptor localRepoDescriptor : resolver.getLocalAndFederatedRepos()) {
            String key = localRepoDescriptor.getKey();
            LocalRepo<?, ?> localRepo = repositoryService.localOrFederatedRepositoryByKey(key);
            if (localRepo != null) {
                repositories.add(localRepo);
            }
        }
        return repositories;

    }

    public List<LocalCacheRepo> getResolvedLocalCacheRepos() {
        //Add all local Cache repositories
        List<LocalCacheRepo> repositories = Lists.newArrayList();
        for (RemoteRepoDescriptor remoteRepoDescriptor : resolver.getRemoteRepos()) {
            String key = remoteRepoDescriptor.getKey();
            RemoteRepo<?, ?> remoteRepo = repositoryService.remoteRepositoryByKey(key);
            if (remoteRepo != null) {
                LocalCacheRepo localCacheRepo = remoteRepo.getLocalCacheRepo();
                if (localCacheRepo != null) {
                    repositories.add(localCacheRepo);
                }
            }
        }
        return repositories;

    }

    public List<RemoteRepo<?, ?>> getResolvedRemoteRepos() {
        //Add all Remote repositories
        List<RemoteRepo<?, ?>> repositories = Lists.newArrayList();
        RepoRequests.logToContext("Appending collective remote repositories");
        for (RemoteRepoDescriptor remoteRepoDescriptor : resolver.getRemoteRepos()) {
            String key = remoteRepoDescriptor.getKey();
            RemoteRepo<?, ?> remoteRepo = repositoryService.remoteRepositoryByKey(key);
            if (remoteRepo != null) {
                repositories.add(remoteRepo);
            }
        }
        return repositories;
    }


    /**
     * @return A list of local and remote repositories to search the resource in, ordered first by type (local non-cache
     * first) and secondly by order of appearance. We don't simply add all the lists since some of the real
     * repos are resolved transitively and we might have to remove them in case the path is excluded in the
     * virtual repo they belong to.
     */
    public List<RealRepo<?, ?>> assembleSearchRepositoriesList(RepoPath repoPath, boolean fromAnotherArtifactory) {
        RepoRequests.logToContext("Preparing list of aggregated repositories to search in");
        VirtualResolverRequestFilter filter = new VirtualResolverRequestFilter(virtualRepo, repoPath,
                repositoryService, pathTranslator, centralConfigService);
        VirtualRepoResolver resolver = new VirtualRepoResolver(virtualRepo.getDescriptor(), filter);

        List<RealRepo<?, ?>> prioritisedRepositories = Lists.newArrayList();
        List<RealRepo<?, ?>> nonPrioritisedRepositories = Lists.newArrayList();

        List<RealRepoDescriptor> localRepoDescriptors = Lists.newArrayList();
        List<RealRepoDescriptor> remoteRepoDescriptors = Lists.newArrayList();
        localRepoDescriptors.addAll(resolver.getLocalAndFederatedRepos());
        remoteRepoDescriptors.addAll(resolver.getRemoteRepos());

        RepoRequests.logToContext("Appending collective local repositories");
        localRepoDescriptors.forEach(descriptor -> {
            String key = descriptor.getKey();
            if (descriptor.isPriorityResolution()) {
                addLocalRepo(prioritisedRepositories, repositoryService.localOrFederatedRepositoryByKey(key));
            } else {
                addLocalRepo(nonPrioritisedRepositories, repositoryService.localOrFederatedRepositoryByKey(key));
            }
        });

        RepoRequests.logToContext("Appending collective local cache repositories");
        remoteRepoDescriptors.forEach(descriptor -> {
            String key = descriptor.getKey();
            if (descriptor.isPriorityResolution()) {
                addLocalCacheRepo(prioritisedRepositories,
                        repositoryService.localFederatedOrCachedRepositoryByKey(key));
            } else {
                addLocalCacheRepo(nonPrioritisedRepositories,
                        repositoryService.localFederatedOrCachedRepositoryByKey(key));
            }
        });

        remoteRepoDescriptors.forEach(descriptor -> {
                    if (descriptor.isPriorityResolution()) {
                        addRemoteRepo(fromAnotherArtifactory, prioritisedRepositories, descriptor);
                    } else {
                        addRemoteRepo(fromAnotherArtifactory, nonPrioritisedRepositories, descriptor);
                    }
                }
        );
        prioritisedRepositories.addAll(nonPrioritisedRepositories);
        return prioritisedRepositories;
    }

    private void addRemoteRepo(boolean fromAnotherArtifactory,
            List<RealRepo<?, ?>> repositories, RealRepoDescriptor descriptor) {

        if (fromAnotherArtifactory && !virtualRepo.isArtifactoryRequestsCanRetrieveRemoteArtifacts()) {
            //If the request comes from another artifactory don't bother checking any remote repos
            RepoRequests.logToContext("Collective remote repositories aren't appended - Download request was " +
                    "received from another Artifactory instance but is forbidden to search in remote repositories.");
        } else {
            RepoRequests.logToContext("Appending collective remote repositories");
            RemoteRepo<?, ?> remoteRepo = repositoryService.remoteRepositoryByKey(descriptor.getKey());
            if (remoteRepo != null) {
                repositories.add(remoteRepo);
            }
        }
    }

    private void addLocalRepo(List<RealRepo<?, ?>> repositories, LocalRepo<?, ?> localRepo) {
        if (localRepo != null) {
            repositories.add(localRepo);
        }
    }

    private void addLocalCacheRepo(List<RealRepo<?, ?>> repositories, LocalRepo<?, ?> localRepo) {
        if (!addonsManager.isEntitled(EntitlementType.EDGE_MIXED)) {
            addLocalRepo(repositories, localRepo);
        }
    }
}
