package org.artifactory.servicetrust.distribution;

import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.artifactory.api.security.access.GenericTokenSpec;
import org.artifactory.api.security.access.TokenSpec;
import org.artifactory.servicetrust.ServiceTrust;
import org.artifactory.servicetrust.ServiceTrustUseCase;
import org.artifactory.servicetrust.TokenExtension;
import org.artifactory.servicetrust.TokenSpecInboundService;
import org.jfrog.access.common.PermissionScope;
import org.jfrog.access.common.PermissionScopeAction;
import org.jfrog.access.common.PermissionScopeResourceType;
import org.jfrog.access.common.ScopeContainer;
import org.jfrog.access.common.ServiceId;
import org.jfrog.common.JsonUtils;

import java.util.Collections;
import java.util.List;

import static org.artifactory.servicetrust.ServiceTrustService.APPLIED_PERMISSIONS_SERVICE_PATTERN;
import static org.artifactory.servicetrust.ServiceTrustService.DEFAULT_EXCHANGE_TOKEN_EXPIRATION_TIME_IN_SECS;
import static org.artifactory.servicetrust.ServiceTrustService.DEFAULT_MASTER_TOKEN_EXPIRATION_TIME_IN_SECS;
import static org.artifactory.servicetrust.ServiceTrustService.DEFAULT_PAIRING_TOKEN_EXPIRATION_TIME_IN_SECS;

@Slf4j
public class DistributionTokenSpecService implements TokenSpecInboundService {

    public static final DistributionTokenSpecService DISTRIBUTION_TOKEN_SPEC_SERVICE =
            new DistributionTokenSpecService();

    public static final String INTRANSIT_SCOPE_TOKEN = PermissionScope.create(
            PermissionScopeResourceType.INTERNAL, "_intransit", "admin", PermissionScopeAction.EXECUTE).toString();

    @Override
    @NonNull
    public TokenSpec<GenericTokenSpec> createPairingTokenSpec(
            @NonNull ServiceId artifactoryServiceId,
            @NonNull ServiceTrust trust,
            @NonNull String subject,
            @NonNull String scope,
            @NonNull String baseUrl,
            @NonNull String pairingUrl,
            @NonNull String exchangeUrl) {

        if (trust.equals(ServiceTrust.TWO_WAY)) {
            throw new IllegalArgumentException("Distribution does not support two-way pairing");
        }

        TokenSpec<GenericTokenSpec> spec = GenericTokenSpec.create(subject);
        spec.audience(Collections.singletonList(artifactoryServiceId.getFormattedName()));
        spec.expiresIn(DEFAULT_PAIRING_TOKEN_EXPIRATION_TIME_IN_SECS);
        spec.scope(Collections.singletonList(scope));
        spec.refreshable(false);

        TokenExtension tokenExtension = new TokenExtension(baseUrl, exchangeUrl, pairingUrl, true);
        String extension = JsonUtils.getInstance().valueToString(tokenExtension);
        spec.extension(extension);

        log.debug("Creating pairing token spec: {}", spec);

        return spec;
    }

    @Override
    @NonNull
    public TokenSpec<GenericTokenSpec> createMasterTokenSpec(
            @NonNull ServiceId artifactoryServiceId,
            @NonNull String useCasePath,
            @NonNull String pairingScope,
            @NonNull String baseUrl,
            @NonNull String pairingUrl,
            @NonNull String exchangeUrl) {

        String subject = artifactoryServiceId.getFormattedName();
        TokenSpec<GenericTokenSpec> spec = GenericTokenSpec.create(subject);

        spec.audience(Collections.singletonList(artifactoryServiceId.getFormattedName()));
        spec.expiresIn(DEFAULT_MASTER_TOKEN_EXPIRATION_TIME_IN_SECS);

        String appliedPermissionsScope = String.format(APPLIED_PERMISSIONS_SERVICE_PATTERN,
                ServiceTrustUseCase.DISTRIBUTION.getName());

        spec.scope(List.of(appliedPermissionsScope, pairingScope));
        spec.refreshable(true);

        TokenExtension newExtension = new TokenExtension(baseUrl, exchangeUrl, pairingUrl, true);
        String extension = JsonUtils.getInstance().valueToString(newExtension);
        spec.extension(extension);

        log.debug("Creating master token spec: {}", spec);
        return spec;
    }

    @Override
    @NonNull
    public TokenSpec<GenericTokenSpec> createExchangeTokenSpec(
            @NonNull ServiceId artifactoryServiceId,
            @NonNull String useCasePath) {

        String subject = artifactoryServiceId.getFormattedName();
        TokenSpec<GenericTokenSpec> spec = GenericTokenSpec.create(subject);
        spec.scope(List.of(ScopeContainer.ADMIN_PERMISSION, INTRANSIT_SCOPE_TOKEN));
        spec.audience(List.of(artifactoryServiceId.getFormattedName()));
        spec.expiresIn(DEFAULT_EXCHANGE_TOKEN_EXPIRATION_TIME_IN_SECS);
        spec.refreshable(false);

        log.debug("Creating exchange token spec: {}", spec);
        return spec;
    }
}
