package org.artifactory.config;

import com.jfrog.sysconf.SysConfig;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.descriptor.config.CentralConfigDescriptorImpl;
import org.artifactory.descriptor.security.PasswordExpirationPolicy;
import org.artifactory.descriptor.security.PasswordSettings;
import org.artifactory.descriptor.security.SecurityDescriptor;
import org.artifactory.descriptor.security.ldap.LdapSetting;
import org.artifactory.descriptor.security.ldap.group.LdapGroupSetting;
import org.artifactory.descriptor.security.oauth.OAuthSettings;
import org.artifactory.descriptor.security.signingkeys.SigningKeysSettings;
import org.artifactory.descriptor.security.sshserver.SshServerSettings;
import org.artifactory.descriptor.security.sso.CrowdSettings;
import org.artifactory.descriptor.security.sso.HttpSsoSettings;
import org.artifactory.descriptor.security.sso.SamlSettings;
import org.artifactory.spring.ArtifactoryApplicationContext;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.util.Collections;
import java.util.List;

import static org.artifactory.config.CentralConfigDescriptorSanitizer.merge;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertFalse;
import static org.testng.AssertJUnit.assertEquals;
import static org.testng.AssertJUnit.assertTrue;

/**
 * @author igoru
 */
public class SecurityDescriptorSanitiseTest {

    private CentralConfigDescriptorImpl source;
    private SecurityDescriptor sourceSecurity;
    private CentralConfigDescriptorImpl target;
    private SecurityDescriptor targetSecurity;

    @Mock
    private ArtifactoryApplicationContext applicationContext;
    @Mock
    private RepositoryService repositoryService;

    @BeforeMethod
    public void setUp() {
        MockitoAnnotations.openMocks(this);
        when(applicationContext.beanForType(RepositoryService.class)).thenReturn(repositoryService);
        when(repositoryService.getLocalRepoConfigsExcludingInfoRepos()).thenReturn(List.of());

        source = new CentralConfigDescriptorImpl();
        sourceSecurity = source.getSecurity();
        target = new CentralConfigDescriptorImpl();
        targetSecurity = target.getSecurity();

        ArtifactoryContextThreadBinder.bind(applicationContext);

        ArtifactoryHome artifactoryHome = mock(ArtifactoryHome.class);
        SysConfig sysConfig = mock(SysConfig.class);
        when(artifactoryHome.getSysConfig()).thenReturn(sysConfig);
        ArtifactoryHome.bind(artifactoryHome);
    }

    @AfterMethod
    public void tearDown() {
        ArtifactoryContextThreadBinder.unbind();
    }

    @Test
    public void takesHideUnauthorizedResourcesFromSource() {
        sourceSecurity.setHideUnauthorizedResources(true);
        targetSecurity.setHideUnauthorizedResources(false);

        SecurityDescriptor result = merge(source, target).getSecurity();

        assertTrue(result.isHideUnauthorizedResources());
    }

    @Test
    public void takesPasswordSettingsFromTarget() {
        PasswordSettings sourceSettings = new PasswordSettings();
        PasswordExpirationPolicy sourcePolicy = new PasswordExpirationPolicy();
        sourcePolicy.setPasswordMaxAge(42);
        sourceSettings.setExpirationPolicy(sourcePolicy);
        sourceSecurity.setPasswordSettings(sourceSettings);

        PasswordSettings targetSettings = new PasswordSettings();
        PasswordExpirationPolicy targetPolicy = new PasswordExpirationPolicy();
        targetPolicy.setPasswordMaxAge(43);
        targetSettings.setExpirationPolicy(targetPolicy);
        targetSecurity.setPasswordSettings(targetSettings);

        SecurityDescriptor result = merge(source, target).getSecurity();

        assertEquals(result.getPasswordSettings().getExpirationPolicy().getPasswordMaxAge().longValue(), 43);
    }

    @Test
    public void takesBuildGlobalBasicReadAllowedFromSource() {
        sourceSecurity.setBuildGlobalBasicReadAllowed(true);
        targetSecurity.setBuildGlobalBasicReadAllowed(false);

        SecurityDescriptor result = merge(source, target).getSecurity();

        assertTrue(result.isBuildGlobalBasicReadAllowed());
    }

    @Test
    public void takesBuildGlobalBasicReadForAnonymousFromSource() {
        sourceSecurity.setBuildGlobalBasicReadForAnonymous(true);
        targetSecurity.setBuildGlobalBasicReadForAnonymous(false);

        SecurityDescriptor result = merge(source, target).getSecurity();

        assertTrue(result.isBuildGlobalBasicReadForAnonymous());
    }

    @Test
    public void takesSshServerSettingsFromTarget() {
        SshServerSettings sourceSettings = new SshServerSettings();
        sourceSettings.setSshServerPort(42);
        sourceSecurity.setSshServerSettings(sourceSettings);

        SshServerSettings targetSettings = new SshServerSettings();
        targetSettings.setSshServerPort(43);
        targetSecurity.setSshServerSettings(targetSettings);

        SecurityDescriptor result = merge(source, target).getSecurity();

        assertEquals(result.getSshServerSettings().getSshServerPort(), 43);
    }

    @Test
    public void takesOauthSettingsFromSource() {
        OAuthSettings sourceSettings = new OAuthSettings();
        sourceSettings.setEnableIntegration(true);
        sourceSettings.setDefaultNpm("source");
        sourceSecurity.setOauthSettings(sourceSettings);

        OAuthSettings targetSettings = new OAuthSettings();
        targetSettings.setEnableIntegration(true);
        targetSettings.setDefaultNpm("target");
        targetSecurity.setOauthSettings(targetSettings);

        SecurityDescriptor result = merge(source, target).getSecurity();

        assertFalse(result.getOauthSettings().isEnableIntegration());
        assertEquals(result.getOauthSettings().getDefaultNpm(), "source");
    }

    @Test
    public void takesSigningKeysSettingsFromSource() {
        SigningKeysSettings sourceSettings = new SigningKeysSettings();
        sourceSettings.setPassphrase("source");
        sourceSecurity.setSigningKeysSettings(sourceSettings);

        SigningKeysSettings targetSettings = new SigningKeysSettings();
        targetSettings.setPassphrase("target");
        targetSecurity.setSigningKeysSettings(targetSettings);

        SecurityDescriptor result = merge(source, target).getSecurity();

        assertEquals(result.getSigningKeysSettings().getPassphrase(), "source");
    }

    @Test
    public void takesLdapGroupSettingsFromSource() {
        LdapGroupSetting sourceSetting = new LdapGroupSetting();
        sourceSetting.setName("source");
        sourceSecurity.setLdapGroupSettings(Collections.singletonList(sourceSetting));

        LdapGroupSetting targetSetting = new LdapGroupSetting();
        targetSetting.setName("target");
        targetSecurity.setLdapGroupSettings(Collections.singletonList(targetSetting));

        SecurityDescriptor result = merge(source, target).getSecurity();

        assertEquals(result.getLdapGroupSettings().size(), 1);
        assertEquals(result.getLdapGroupSettings().get(0).getName(), "source");
    }

    @Test
    public void takesLdapSettingsFromSource() {
        LdapSetting sourceSetting = new LdapSetting();
        sourceSetting.setEnabled(true);
        sourceSetting.setKey("source");
        sourceSecurity.setLdapSettings(Collections.singletonList(sourceSetting));

        LdapSetting targetSetting = new LdapSetting();
        targetSetting.setEnabled(true);
        targetSetting.setKey("target");
        targetSecurity.setLdapSettings(Collections.singletonList(targetSetting));

        SecurityDescriptor result = merge(source, target).getSecurity();

        assertEquals(result.getLdapSettings().size(), 1);
        assertEquals(result.getLdapSettings().get(0).getKey(), "source");
        assertFalse(result.getLdapSettings().get(0).isEnabled());
    }

    @Test
    public void ignoreHttpSsoSettingsFromSource() {
        HttpSsoSettings sourceSettings = new HttpSsoSettings();
        sourceSettings.setHttpSsoProxied(true);
        sourceSettings.setRemoteUserRequestVariable("source");
        sourceSecurity.setHttpSsoSettings(sourceSettings);

        HttpSsoSettings targetSettings = new HttpSsoSettings();
        targetSettings.setHttpSsoProxied(true);
        targetSettings.setRemoteUserRequestVariable("target");
        targetSecurity.setHttpSsoSettings(targetSettings);

        SecurityDescriptor result = merge(source, target).getSecurity();

        assertEquals(result.getHttpSsoSettings().getRemoteUserRequestVariable(), "target");
        assertTrue(result.getHttpSsoSettings().isHttpSsoProxied());
    }

    @Test
    public void takesCrowdSettingsFromSource() {
        CrowdSettings sourceSettings = new CrowdSettings();
        sourceSettings.setEnableIntegration(true);
        sourceSettings.setApplicationName("source");
        sourceSecurity.setCrowdSettings(sourceSettings);

        CrowdSettings targetSettings = new CrowdSettings();
        targetSettings.setEnableIntegration(true);
        targetSettings.setApplicationName("target");
        targetSecurity.setCrowdSettings(targetSettings);

        SecurityDescriptor result = merge(source, target).getSecurity();

        assertEquals(result.getCrowdSettings().getApplicationName(), "source");
        assertFalse(result.getCrowdSettings().isEnableIntegration());
    }

    @Test
    public void takesSamlSettingsFromSource() {
        SamlSettings sourceSettings = new SamlSettings();
        sourceSettings.setEnableIntegration(true);
        sourceSettings.setLoginUrl("source");
        sourceSecurity.setSamlSettings(sourceSettings);

        SamlSettings targetSettings = new SamlSettings();
        targetSettings.setEnableIntegration(true);
        targetSettings.setLoginUrl("target");
        targetSecurity.setSamlSettings(targetSettings);

        SecurityDescriptor result = merge(source, target).getSecurity();

        assertEquals(result.getSamlSettings().getLoginUrl(), "source");
        assertFalse(result.getSamlSettings().isEnableIntegration());
    }
}
