package org.artifactory.ph;

import org.artifactory.core.api.ph.exception.PackageException;
import org.artifactory.core.api.ph.listener.PackageConfigChangedEventListener;
import org.artifactory.core.api.ph.listener.PackageConfigChangedEventType;
import org.artifactory.core.api.ph.service.PackageRepoService;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.repo.RepoConfigChange;
import org.artifactory.repo.RepoDetailsType;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoAnnotations;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Stream;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.doAnswer;
import static org.mockito.Mockito.never;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.verifyNoMoreInteractions;
import static org.mockito.Mockito.when;

class PackageConfigurationChangesInterceptorImplTest {

    @Mock
    PackageConfigChangedEventListener packageConfigChangedEventListener;
    @Mock
    PackageRepoService packageRepoService;
    @Mock
    CentralConfigDescriptor oldDescriptor;
    @Mock
    CentralConfigDescriptor newDescriptor;
    PackageConfigurationChangesInterceptorImpl packageConfigurationChangesInterceptor;

    @BeforeEach
    void setup() {
        MockitoAnnotations.openMocks(this);
        packageConfigurationChangesInterceptor =
                spy(new PackageConfigurationChangesInterceptorImpl(packageConfigChangedEventListener,
                        packageRepoService));
    }

    @AfterEach
    void cleanup() {
        verifyNoMoreInteractions(packageConfigChangedEventListener);
    }


    @Test
    void testAfterGlobalSaveWithoutChanges() {
        packageConfigurationChangesInterceptor.onAfterSave(newDescriptor, null);
        verify(this.packageConfigChangedEventListener, never()).onGlobalConfigChange(any());
    }

    @Test
    void testAfterGlobalConfigSave() {
        when(oldDescriptor.getUrlBase()).thenReturn("http://localhost:8082");
        PackageConfigChangedEventListener packageConfigChangedEventListener =
                Mockito.mock(PackageConfigChangedEventListener.class);
        doAnswer(invocationOnMock -> {
            Assertions.assertEquals(PackageConfigChangedEventType.BASE_URL, invocationOnMock.getArgument(0));
            return null;
        }).when(packageConfigChangedEventListener).onGlobalConfigChange(any());

        packageConfigurationChangesInterceptor.onAfterSave(newDescriptor, oldDescriptor);
        verify(this.packageConfigChangedEventListener, times(1)).onGlobalConfigChange(any());
        verifyNoMoreInteractions(packageConfigChangedEventListener);
    }

    @Test
    void testAfterRepoConfigSave() throws PackageException {
        PackageConfigChangedEventListener packageConfigChangedEventListener =
                Mockito.mock(PackageConfigChangedEventListener.class);
        doAnswer(invocationOnMock -> {
            Assertions.assertEquals(PackageConfigChangedEventType.BASE_URL, invocationOnMock.getArgument(0));
            return null;
        }).when(packageConfigChangedEventListener).onRepoConfigChange(any());
        Map<RepoDetailsType, List<RepoConfigChange>> repoConfigChanges = new HashMap<>();
        RepoConfigChange repoConfigChange = new RepoConfigChange("test",
                RepoConfigChange.ConfigAction.REPO_CREATE, null, null);
        repoConfigChanges.put(RepoDetailsType.LOCAL, List.of(repoConfigChange));
        packageConfigurationChangesInterceptor.onAfterSave(repoConfigChanges);
        verify(this.packageConfigChangedEventListener, times(1)).onRepoConfigChange(any());
        verifyNoMoreInteractions(packageConfigChangedEventListener);
    }

    @ParameterizedTest
    @MethodSource("testAfterRepoSaveWithoutChangesProvider")
    void testAfterRepoSaveWithoutChanges(Map<RepoDetailsType, List<RepoConfigChange>> repoConfigChanges)
            throws PackageException {
        packageConfigurationChangesInterceptor.onAfterSave(repoConfigChanges);
        verify(this.packageConfigChangedEventListener, never()).onRepoConfigChange(any());
    }

    static Stream<Arguments> testAfterRepoSaveWithoutChangesProvider() {
        Map<RepoDetailsType, List<RepoConfigChange>> emptyConfig = new HashMap<>();
        Map<RepoDetailsType, List<RepoConfigChange>> emptyRepoConfigChange = new HashMap<>();
        RepoConfigChange repoConfigChange = new RepoConfigChange("test", null,
                null, null);
        emptyRepoConfigChange.put(RepoDetailsType.LOCAL, List.of(repoConfigChange));

        return Stream.of(Arguments.of(emptyConfig),
                Arguments.of(emptyRepoConfigChange));
    }
}