package org.artifactory.repo.cache;

import org.apache.commons.collections.CollectionUtils;
import org.artifactory.model.LocalRepoConfig;
import org.artifactory.model.RepoConfig;
import org.artifactory.model.RepoConfigPropagationModel;
import org.artifactory.repo.RepoConfigChange;
import org.artifactory.repo.RepoConfigReload;
import org.junit.jupiter.api.RepeatedTest;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.BrokenBarrierException;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CyclicBarrier;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.function.BiFunction;
import java.util.function.Function;
import java.util.function.Supplier;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

import static org.artifactory.descriptor.repo.RepoType.Generic;
import static org.artifactory.repo.RepoConfigChange.ConfigAction.REPO_UPDATE;
import static org.artifactory.repo.RepoDetailsType.LOCAL;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertNull;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.atLeastOnce;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

/**
 * @author vovam
 */
@SuppressWarnings("unchecked")
class RepoConfigCacheTest {

    @Test
    void repoCacheInitialization() {
        Supplier<Map<String, RepoConfig>> getAllReposFunction = mock(Supplier.class);
        RepoConfig repoConfig1 = LocalRepoConfig.create("repo1", Generic);
        RepoConfig repoConfig2 = LocalRepoConfig.create("repo2", Generic);
        when(getAllReposFunction.get())
                .thenReturn(Map.of("repo1", repoConfig1))
                .thenReturn(Map.of("repo2", repoConfig2));

        RepoConfigCache repoConfigCache = new RepoConfigCache(Long.MAX_VALUE, Long.MAX_VALUE, Integer.MAX_VALUE,
                Double.MAX_VALUE, getAllReposFunction, null, null, null, null);
        assertTrue(repoConfigCache.existByKey("repo1"));
        assertFalse(repoConfigCache.existByKey("repo2"));
    }

    @Test
    void existByKey() {
        Supplier<Map<String, RepoConfig>> getAllReposFunction = mock(Supplier.class);
        RepoConfig repoConfig = LocalRepoConfig.create("repo1", Generic);
        when(getAllReposFunction.get()).thenReturn(Map.of("repo1", repoConfig));

        RepoConfigCache repoConfigCache = new RepoConfigCache(Long.MAX_VALUE, Long.MAX_VALUE, Integer.MAX_VALUE,
                Double.MAX_VALUE, getAllReposFunction, mock(Function.class), null, null, null);
        assertTrue(repoConfigCache.existByKey("repo1"));
        assertFalse(repoConfigCache.existByKey("repo2"));
    }

    @Test
    void getByKey() {
        Supplier<Map<String, RepoConfig>> getAllReposFunction = mock(Supplier.class);
        RepoConfig repoConfig1 = LocalRepoConfig.create("repo1", Generic);
        RepoConfig repoConfig2 = LocalRepoConfig.create("repo2", Generic);
        when(getAllReposFunction.get()).thenReturn(Map.of("repo1", repoConfig1, "repo2", repoConfig2));

        RepoConfigCache repoConfigCache = new RepoConfigCache(Long.MAX_VALUE, Long.MAX_VALUE, Integer.MAX_VALUE,
                Double.MAX_VALUE, getAllReposFunction, mock(Function.class), null, null, null);

        assertNotNull(repoConfigCache.getByKey("repo1"));
        assertEquals("repo1", repoConfigCache.getByKey("repo1").getKey());
        assertNull(repoConfigCache.getByKey("repo3"));
    }

    @Test
    void getByKeyWhenRepoPresentInKeysSetAndDoesNotExistInDb() throws InterruptedException {
        Supplier<Map<String, RepoConfig>> getAllReposFunction = mock(Supplier.class);
        when(getAllReposFunction.get()).thenReturn(Map.of());
        Supplier<Map<String, Long>> getAllRepoKeys = mock(Supplier.class);
        when(getAllRepoKeys.get()).thenReturn(Map.of("repo1", 0L));

        RepoConfigCache repoConfigCache =
                new RepoConfigCache(Integer.MAX_VALUE, Long.MAX_VALUE, Integer.MAX_VALUE,
                        Double.MAX_VALUE, getAllReposFunction, mock(Function.class), null, null, getAllRepoKeys);

        repoConfigCache.addAndUpdateRepoConfigKeys(Map.of("repo1", 0L));

        // exist because repo keys set contains repo key
        assertTrue(repoConfigCache.existByKey("repo1"));
        // null because DB does not contain it
        assertNull(repoConfigCache.getByKey("repo1"));
    }

    @Test
    void findRepositoryConfigsByKeys() {
        Supplier<Map<String, RepoConfig>> getAllReposFunction = mock(Supplier.class);
        RepoConfig repoConfig1 = LocalRepoConfig.create("repo1", Generic);
        RepoConfig repoConfig2 = LocalRepoConfig.create("repo2", Generic);
        RepoConfig repoConfig3 = LocalRepoConfig.create("repo3", Generic);
        when(getAllReposFunction.get()).thenReturn(
                Map.of("repo1", repoConfig1, "repo2", repoConfig2, "repo3", repoConfig3));

        RepoConfigCache repoConfigCache =
                new RepoConfigCache(Long.MAX_VALUE, Long.MAX_VALUE, Integer.MAX_VALUE,
                        Double.MAX_VALUE, getAllReposFunction, mock(Function.class), null, null, null);
        Collection<RepoConfig> repoConfigs =
                repoConfigCache.getRepositoryConfigsByKeys(Set.of("repo1", "repo2"));
        assertEquals(2, repoConfigs.size());
        assertEquals(repoConfig1, findRepoConfigByKey("repo1", repoConfigs));
        assertEquals(repoConfig2, findRepoConfigByKey("repo2", repoConfigs));
    }

    @Test
    void findReposByKeysWhenSomeReposDoNotExistInDb() {
        Supplier<Map<String, RepoConfig>> getAllReposFunction = mock(Supplier.class);
        when(getAllReposFunction.get()).thenReturn(Map.of());
        Supplier<Map<String, Long>> getAllRepoKeysFunction = mock(Supplier.class);
        when(getAllRepoKeysFunction.get()).thenReturn(Map.of("repo1", 0L));
        Function<Set<String>, Map<String, RepoConfig>> finReposByKeysFunction = mock(Function.class);
        when(finReposByKeysFunction.apply(Set.of("repo1"))).thenReturn(Map.of());

        RepoConfigCache repoConfigCache =
                new RepoConfigCache(Long.MAX_VALUE, Long.MAX_VALUE, Integer.MAX_VALUE,
                        Double.MAX_VALUE, getAllReposFunction, mock(Function.class), finReposByKeysFunction, null,
                        getAllRepoKeysFunction);

        assertTrue(repoConfigCache.getRepositoryConfigsByKeys(Set.of("repo1")).isEmpty());

        // 2 times - first when cache tries to load because item is not cached, second - db query fallback because of
        // the exception
        verify(finReposByKeysFunction, times(2)).apply(Set.of("repo1"));
    }

    @Test
    void findReposByKeysWhenMissedInKeySet() {
        Supplier<Map<String, RepoConfig>> getAllReposFunction = mock(Supplier.class);
        when(getAllReposFunction.get()).thenReturn(Map.of());
        Supplier<Map<String, Long>> getAllRepoKeysFunction = mock(Supplier.class);
        when(getAllRepoKeysFunction.get()).thenReturn(Map.of("repo1", 0L, "repo2", 0L, "repo3", 0L));
        Function<Set<String>, Map<String, RepoConfig>> finReposByKeysFunction = mock(Function.class);
        when(finReposByKeysFunction.apply(Set.of("repo1", "repo4"))).thenReturn(
                Map.of("repo4", LocalRepoConfig.create("repo4", Generic)));

        RepoConfigCache repoConfigCache =
                new RepoConfigCache(Long.MAX_VALUE, Long.MAX_VALUE, Integer.MAX_VALUE,
                        Double.MAX_VALUE, getAllReposFunction, mock(Function.class), finReposByKeysFunction, null,
                        getAllRepoKeysFunction);

        List<RepoConfig> repoConfigs =
                new ArrayList<>(repoConfigCache.getRepositoryConfigsByKeys(Set.of("repo1", "repo4")));
        assertEquals(repoConfigs.size(), 1);
        assertEquals(repoConfigs.get(0).getKey(), "repo4");

        // 2 times - first when cache tries to load because items are not cached, second - db query fallback because of
        // the exception
        verify(finReposByKeysFunction, times(2)).apply(Set.of("repo1", "repo4"));
    }

    @Test
    void getAll() {
        Supplier<Map<String, RepoConfig>> getAllReposFunction = mock(Supplier.class);
        RepoConfig repoConfig1 = LocalRepoConfig.create("repo1", Generic);
        RepoConfig repoConfig2 = LocalRepoConfig.create("repo2", Generic);
        when(getAllReposFunction.get()).thenReturn(Map.of("repo1", repoConfig1, "repo2", repoConfig2));

        RepoConfigCache repoConfigCache =
                new RepoConfigCache(Long.MAX_VALUE, Long.MAX_VALUE, Integer.MAX_VALUE,
                        Double.MAX_VALUE, getAllReposFunction, null, null, null, null);
        assertEquals(2, repoConfigCache.getAll().size());
        assertEquals(repoConfig1, findRepoConfigByKey("repo1", repoConfigCache.getAll()));
        assertEquals(repoConfig2, findRepoConfigByKey("repo2", repoConfigCache.getAll()));
    }

    @Test
    void getAllWhenRepoKeysDoNotExistInCache() {
        // given - cache (and db) contains repo1
        Supplier<Map<String, RepoConfig>> getAllReposFunction = mock(Supplier.class);
        RepoConfig repoConfig1 = LocalRepoConfig.create("repo1", Generic);
        when(getAllReposFunction.get()).thenReturn(Map.of("repo1", repoConfig1));

        RepoConfigCache repoConfigCache = new RepoConfigCache(Integer.MAX_VALUE, Long.MAX_VALUE, Integer.MAX_VALUE,
                Double.MAX_VALUE, getAllReposFunction, null, mock(Function.class),
                null, null);

        // given - key set cache contains repo1, repo2
        repoConfigCache.addRepoConfigKeys(Map.of("repo1", 0L, "repo2", 2L));

        assertEquals(2, repoConfigCache.countAll());
        assertTrue(CollectionUtils.isEqualCollection(repoConfigCache.getAllRepoKeys(), Set.of("repo1", "repo2")));

        // then - getAll returns only repo1
        assertEquals(1, repoConfigCache.getAll().size());
        RepoConfig retrievedRepoConfig = repoConfigCache.getAll().iterator().next();
        assertEquals(repoConfig1, retrievedRepoConfig);
    }

    @Test
    void countAll() {
        Supplier<Map<String, RepoConfig>> getAllReposFunction = mock(Supplier.class);
        RepoConfig repoConfig1 = LocalRepoConfig.create("repo1", Generic);
        RepoConfig repoConfig2 = LocalRepoConfig.create("repo2", Generic);
        when(getAllReposFunction.get()).thenReturn(Map.of("repo1", repoConfig1, "repo2", repoConfig2));

        RepoConfigCache repoConfigCache =
                new RepoConfigCache(Long.MAX_VALUE, Long.MAX_VALUE, Integer.MAX_VALUE,
                        Double.MAX_VALUE, getAllReposFunction, null, null, null, null);

        assertEquals(2, repoConfigCache.getAll().size());
    }

    @Test
    void addRepoConfigKeys() {
        Supplier<Map<String, RepoConfig>> getAllReposFunction = mock(Supplier.class);
        RepoConfig repoConfig1 = LocalRepoConfig.create("repo1", Generic);
        when(getAllReposFunction.get()).thenReturn(Map.of("repo1", repoConfig1));

        RepoConfigCache repoConfigCache =
                new RepoConfigCache(Long.MAX_VALUE, Long.MAX_VALUE, Integer.MAX_VALUE,
                        Double.MAX_VALUE, getAllReposFunction, mock(Function.class), null,
                        null, null);

        assertTrue(repoConfigCache.existByKey("repo1"));
        assertFalse(repoConfigCache.existByKey("repo2"));

        repoConfigCache.addRepoConfigKeys(Map.of("repo2", 0L));

        assertTrue(repoConfigCache.existByKey("repo1"));
        assertTrue(repoConfigCache.existByKey("repo2"));
    }

    @Test
    void updateRepoConfigKeys() {
        Supplier<Map<String, RepoConfig>> getAllReposFunction = mock(Supplier.class);
        RepoConfig repoConfig1 = LocalRepoConfig.create("repo1", Generic);
        when(getAllReposFunction.get()).thenReturn(Map.of("repo1", repoConfig1));

        Function<String, RepoConfig> findByKeyFunction = mock(Function.class);
        when(findByKeyFunction.apply("repo1")).thenReturn(repoConfig1);

        RepoConfigCache repoConfigCache =
                new RepoConfigCache(Long.MAX_VALUE, Long.MAX_VALUE, Integer.MAX_VALUE,
                        Double.MAX_VALUE, getAllReposFunction, findByKeyFunction, null, null, null);

        assertNotNull(repoConfigCache.getByKey("repo1"));
        verify(findByKeyFunction, times(0)).apply("repo1");

        repoConfigCache.updateRepoConfigKeys(Set.of("repo1"));
        verify(findByKeyFunction, times(0)).apply("repo1");

        assertNotNull(repoConfigCache.getByKey("repo1"));
        verify(findByKeyFunction, times(1)).apply("repo1");
    }

    @Test
    void deleteRepoConfigsByKeys() {
        Supplier<Map<String, RepoConfig>> getAllReposFunction = mock(Supplier.class);
        RepoConfig repoConfig1 = LocalRepoConfig.create("repo1", Generic);
        RepoConfig repoConfig2 = LocalRepoConfig.create("repo2", Generic);
        Map<String, RepoConfig> map = Map.of("repo1", repoConfig1, "repo2", repoConfig2);
        when(getAllReposFunction.get()).thenReturn(map);

        RepoConfigCache repoConfigCache = new RepoConfigCache(Long.MAX_VALUE, Long.MAX_VALUE, Integer.MAX_VALUE,
                Double.MAX_VALUE, getAllReposFunction,
                mock(Function.class), mock(Function.class), null, mock(Supplier.class));
        assertTrue(repoConfigCache.existByKey("repo1"));
        assertTrue(repoConfigCache.existByKey("repo2"));

        repoConfigCache.deleteRepoConfigsByKeys(List.of(repoConfig2.getKey()));

        assertTrue(repoConfigCache.existByKey("repo1"));
        assertFalse(repoConfigCache.existByKey("repo2"));
    }

    @Test
    void concurrentAddRepoConfigsToCache() throws InterruptedException, BrokenBarrierException {
        int threadCount = 5;
        CyclicBarrier cyclicBarrier = new CyclicBarrier(threadCount + 1);
        Supplier<Map<String, RepoConfig>> getAllReposFunction = mock(Supplier.class);
        when(getAllReposFunction.get()).thenReturn(Map.of());

        RepoConfigCache repoConfigCache = new RepoConfigCache(Long.MAX_VALUE, Long.MAX_VALUE, Integer.MAX_VALUE,
                Double.MAX_VALUE, getAllReposFunction,
                mock(Function.class), mock(Function.class), null, mock(Supplier.class));

        ExecutorService executorService = Executors.newFixedThreadPool(threadCount);
        List<Future> futures = IntStream.range(0, threadCount)
                .mapToObj(value -> executorService.submit(() -> {
                    awaitCyclicBarrier(cyclicBarrier);
                    repoConfigCache.addRepoConfigKeys(Map.of("repoKey" + value, 0L));
                }))
                .collect(Collectors.toList());

        cyclicBarrier.await();
        try {
            futures.forEach(future -> getFuture(future));
        } finally {
            executorService.shutdown();
        }
        IntStream.range(0, threadCount)
                .forEach(value -> assertTrue(repoConfigCache.existByKey("repoKey" + value)));
    }

    @Test
    void concurrentDeleteRepoConfigsFromCache() throws InterruptedException, BrokenBarrierException {
        int threadCount = 5;
        CyclicBarrier cyclicBarrier = new CyclicBarrier(threadCount + 1);
        Supplier<Map<String, RepoConfig>> getAllReposFunction = mock(Supplier.class);

        Map<String, RepoConfig> allRepoConfigs = new HashMap<>();
        IntStream.range(0, 5).forEach((value) -> {
            RepoConfig repoConfig = LocalRepoConfig.create("repo", Generic);
            allRepoConfigs.put("repo" + value, repoConfig);
        });
        when(getAllReposFunction.get()).thenReturn(allRepoConfigs);

        RepoConfigCache repoConfigCache = new RepoConfigCache(Long.MAX_VALUE, Long.MAX_VALUE, Integer.MAX_VALUE,
                Double.MAX_VALUE, getAllReposFunction, mock(Function.class), mock(Function.class), null,
                mock(Supplier.class));

        ExecutorService executorService = Executors.newFixedThreadPool(threadCount);
        List<Future> futures = IntStream
                .range(0, threadCount)
                .mapToObj(value -> executorService.submit(() -> {
                    awaitCyclicBarrier(cyclicBarrier);
                    repoConfigCache.deleteRepoConfigsByKeys(Set.of("repoKey" + value));
                }))
                .collect(Collectors.toList());

        cyclicBarrier.await();
        try {
            futures.forEach(future -> getFuture(future));
        } finally {
            executorService.shutdown();
        }
        IntStream.range(0, threadCount)
                .forEach(value -> assertFalse(repoConfigCache.existByKey("repoKey" + value)));
    }

    @Test
    void concurrentUpdateRepoConfigsInCache() throws InterruptedException, BrokenBarrierException {
        int threadCount = 5;
        CyclicBarrier cyclicBarrier = new CyclicBarrier(threadCount + 1);
        Supplier<Map<String, RepoConfig>> getAllReposFunction = mock(Supplier.class);

        Map<String, RepoConfig> allRepoConfigs = new HashMap<>();
        IntStream.range(0, 5).forEach((value) -> {
            RepoConfig repoConfig = LocalRepoConfig.create("repoKey" + value, Generic);
            repoConfig.getBaseConfig().setDescription("original");
            allRepoConfigs.put("repoKey" + value, repoConfig);
        });
        when(getAllReposFunction.get()).thenReturn(allRepoConfigs);

        Function<String, RepoConfig> findByKeyFunction = repoKey -> {
            RepoConfig repoConfig = LocalRepoConfig.create(repoKey, Generic);
            repoConfig.getBaseConfig().setDescription("updated");
            return repoConfig;
        };

        RepoConfigCache repoConfigCache = new RepoConfigCache(Long.MAX_VALUE, Long.MAX_VALUE, Integer.MAX_VALUE,
                Double.MAX_VALUE, getAllReposFunction, findByKeyFunction, mock(Function.class), null,
                mock(Supplier.class));

        ExecutorService executorService = Executors.newFixedThreadPool(threadCount);
        List<Future> futures = IntStream
                .range(0, threadCount)
                .mapToObj(value -> executorService.submit(() -> {
                    awaitCyclicBarrier(cyclicBarrier);
                    repoConfigCache.updateRepoConfigKeys(Set.of("repoKey" + value));
                }))
                .collect(Collectors.toList());

        cyclicBarrier.await();
        try {
            futures.forEach(future -> getFuture(future));
        } finally {
            executorService.shutdown();
        }
        IntStream.range(0, threadCount)
                .forEach(value -> assertEquals(
                        repoConfigCache.getByKey("repoKey" + value).getBaseConfig().getDescription(), "updated"));
    }

    @RepeatedTest(5)
    void concurrentReadWriteSyncRepoKeysOperationsInCache() throws InterruptedException, BrokenBarrierException {
        Map<String, Long> allRepoConfigKeys = new ConcurrentHashMap<>();
        Supplier<Map<String, Long>> getAllRepoKeysFunction = mock(Supplier.class);
        when(getAllRepoKeysFunction.get()).thenReturn(allRepoConfigKeys);

        Function<String, RepoConfig> findByKeyFunction = repoKey -> {
            RepoConfig repoConfig = LocalRepoConfig.create(repoKey, Generic);
            return repoConfig;
        };

        Supplier<Map<String, RepoConfig>> getAllReposFunction = mock(Supplier.class);
        when(getAllReposFunction.get()).thenReturn(Map.of());

        long syncAllKeysPeriodInMillis = 10;
        RepoConfigCache repoConfigCache =
                new RepoConfigCache(syncAllKeysPeriodInMillis, syncAllKeysPeriodInMillis * 3, Integer.MAX_VALUE,
                        Double.MAX_VALUE, getAllReposFunction, findByKeyFunction, mock(Function.class), null,
                        getAllRepoKeysFunction);

        int threadCountPerAction = 5;
        // to sync threads that run add/update/delete operations
        CyclicBarrier cyclicBarrier = new CyclicBarrier(threadCountPerAction * 3 + 1);
        ExecutorService executorService = Executors.newFixedThreadPool(threadCountPerAction * 3);

        // add a repo key to a cache
        List<Future> futures = IntStream
                .range(0, threadCountPerAction)
                .mapToObj(value -> executorService.submit(() -> {
                    // wait until all threads will be ready
                    awaitCyclicBarrier(cyclicBarrier);
                    // add to the set - data source for getAllRepoKeys method
                    allRepoConfigKeys.put("repoKey" + value, 0L);
                    // add repo key to cache
                    repoConfigCache.addRepoConfigKeys(Map.of("repoKey" + value, 0L));
                    // assert that repo exists
                    assertTrue(repoConfigCache.existByKey("repoKey" + value));
                    assertNotNull(repoConfigCache.getByKey("repoKey" + value));
                }))
                .collect(Collectors.toList());

        // update repo key in a cache
        futures.addAll(IntStream
                .range(threadCountPerAction * 3, threadCountPerAction * 4)
                .mapToObj(value -> executorService.submit(() -> {
                    // add a repo to execute an update operation
                    allRepoConfigKeys.put("repoKey" + value, 0L);
                    repoConfigCache.addRepoConfigKeys(Map.of("repoKey" + value, 0L));
                    // wait until all threads will be ready
                    awaitCyclicBarrier(cyclicBarrier);
                    repoConfigCache.updateRepoConfigKeys(Set.of("repoKey" + value));
                    assertNotNull(repoConfigCache.getByKey("repoKey" + value));
                }))
                .collect(Collectors.toList()));

        futures.addAll(IntStream
                .range(threadCountPerAction, threadCountPerAction * 2)
                .mapToObj(value -> executorService.submit(() -> {
                    // add a repo to execute a remove operation
                    allRepoConfigKeys.put("repoKey" + value, 0L);
                    repoConfigCache.addRepoConfigKeys(Map.of("repoKey" + value, 0L));
                    // wait until all threads will be ready
                    awaitCyclicBarrier(cyclicBarrier);
                    allRepoConfigKeys.remove("repoKey" + value);
                    repoConfigCache.deleteRepoConfigsByKeys(Set.of("repoKey" + value));
                    assertFalse(repoConfigCache.existByKey("repoKey" + value));
                }))
                .collect(Collectors.toList()));

        cyclicBarrier.await();

        Thread.sleep(syncAllKeysPeriodInMillis);
        try {
            futures.forEach(future -> getFuture(future));
        } finally {
            executorService.shutdown();
        }

        IntStream
                .range(0, threadCountPerAction * 2)
                .forEach(value -> repoConfigCache.existByKey("repoKey" + value));
    }

    @Test
    void syncRepoCacheByModificationDateWhenModificationIsNotChanged() {
        // given - cache contains repo without modification date
        Supplier<Map<String, RepoConfig>> getAllReposFunction = mock(Supplier.class);
        RepoConfig repoConfigWithoutModificationDate = LocalRepoConfig.create("repo", Generic);
        when(getAllReposFunction.get()).thenReturn(Map.of("repo", repoConfigWithoutModificationDate));

        Function<String, RepoConfig> findByKeyFunction = mock(Function.class);

        // given - db contains repo without modification date
        Supplier<Map<String, Long>> getAllRepoKeys = mock(Supplier.class);
        when(getAllRepoKeys.get()).thenReturn(Map.of("repo", 0L));

        BiFunction<RepoConfigPropagationModel, RepoConfigCache, Void> repoConfigurationsFunction =
                mock(BiFunction.class);

        RepoConfigCache repoConfigCache = new RepoConfigCache(Integer.MAX_VALUE, Long.MAX_VALUE, Integer.MAX_VALUE,
                Double.MAX_VALUE, getAllReposFunction, findByKeyFunction, null, repoConfigurationsFunction,
                getAllRepoKeys);

        RepoConfig repoConfigOriginal = repoConfigCache.getByKey("repo");
        assertNotNull(repoConfigOriginal);
        assertEquals(0L, repoConfigOriginal.getModification());
        verify(findByKeyFunction, times(0)).apply("repo");

        // when
        repoConfigCache.syncRepoConfigKeys();

        // then
        verify(getAllRepoKeys, atLeastOnce()).get();
        verify(repoConfigurationsFunction, times(0)).apply(any(), any());
        verify(findByKeyFunction, times(0)).apply("repo");
    }

    @Test
    void syncRepoCacheWhenModificationIsChanged() {
        // given - cache contains repo without modification date
        Supplier<Map<String, RepoConfig>> getAllReposFunction = mock(Supplier.class);
        RepoConfig repoConfigWithoutModificationDate = LocalRepoConfig.create("repo", Generic);
        when(getAllReposFunction.get()).thenReturn(Map.of("repo", repoConfigWithoutModificationDate));

        // given - db contains repo with modification date
        RepoConfig repoConfigWithModificationDate = LocalRepoConfig.create("repo", Generic);
        repoConfigWithModificationDate.setModification(1L);
        Function<String, RepoConfig> findByKeyFunction = mock(Function.class);
        when(findByKeyFunction.apply("repo")).thenReturn(repoConfigWithModificationDate);

        Supplier<Map<String, Long>> getAllRepoKeys = mock(Supplier.class);
        when(getAllRepoKeys.get()).thenReturn(Map.of("repo", 1L));

        BiFunction<RepoConfigPropagationModel, RepoConfigCache, Void> repoConfigurationsFunction =
                mock(BiFunction.class);
        RepoConfigCache repoConfigCache = new RepoConfigCache(Integer.MAX_VALUE, Long.MAX_VALUE, Integer.MAX_VALUE,
                Double.MAX_VALUE, getAllReposFunction, findByKeyFunction, null, repoConfigurationsFunction,
                getAllRepoKeys);

        RepoConfig repoConfigOriginal = repoConfigCache.getByKey("repo");
        assertNotNull(repoConfigOriginal);
        assertEquals(0L, repoConfigOriginal.getModification());
        verify(findByKeyFunction, times(0)).apply("repo");

        // when
        repoConfigCache.syncRepoConfigKeys();

        // then - call reload with REPO_UPDATE action
        verify(getAllRepoKeys, atLeastOnce()).get();
        ArgumentCaptor<RepoConfigPropagationModel> captor = ArgumentCaptor.forClass(RepoConfigPropagationModel.class);
        verify(repoConfigurationsFunction, atLeastOnce()).apply(captor.capture(), any());
        RepoConfigPropagationModel propagationModel = captor.getValue();
        Set<RepoConfigReload> repoConfigCreateOrUpdate = propagationModel.getRepoConfigCreateOrUpdate();
        assertEquals(1, propagationModel.getRepoConfigCreateOrUpdate().size());
        assertEquals(0, propagationModel.getRepoConfigDelete().size());
        RepoConfigReload configReload = repoConfigCreateOrUpdate.iterator().next();
        assertEquals("repo", configReload.getRepoKey());
        assertEquals(LOCAL, configReload.getRepoDetailsType());
        assertEquals(Generic, configReload.getRepoType());
        assertEquals(REPO_UPDATE, configReload.getConfigAction());
        assertEquals(1L, configReload.getModification());
    }

    @Test
    void syncRepoCacheWhenRepoDoesNotExistInDB() {
        // given - cache contains repo
        RepoConfig repoConfig = LocalRepoConfig.create("repo", Generic);
        Supplier<Map<String, RepoConfig>> getAllReposFunction = mock(Supplier.class);
        when(getAllReposFunction.get()).thenReturn(Map.of("repo", repoConfig));

        Function<String, RepoConfig> findByKeyFunction = mock(Function.class);
        when(findByKeyFunction.apply("repo")).thenReturn(repoConfig);

        // given - db does not contain repo
        Supplier<Map<String, Long>> getAllRepoKeys = mock(Supplier.class);
        when(getAllRepoKeys.get()).thenReturn(Map.of());

        BiFunction<RepoConfigPropagationModel, RepoConfigCache, Void> repoConfigurationsFunction =
                mock(BiFunction.class);

        RepoConfigCache repoConfigCache = new RepoConfigCache(Integer.MAX_VALUE, Long.MAX_VALUE, Integer.MAX_VALUE,
                Double.MAX_VALUE, getAllReposFunction, findByKeyFunction, null, repoConfigurationsFunction,
                getAllRepoKeys);

        RepoConfig repoConfigOriginal = repoConfigCache.getByKey("repo");
        assertNotNull(repoConfigOriginal);
        verify(findByKeyFunction, times(0)).apply("repo");

        // when
        repoConfigCache.syncRepoConfigKeys();

        // then - call reload with REPO_DELETE action
        verify(getAllRepoKeys, atLeastOnce()).get();
        ArgumentCaptor<RepoConfigPropagationModel> captor = ArgumentCaptor.forClass(RepoConfigPropagationModel.class);
        verify(repoConfigurationsFunction, atLeastOnce()).apply(captor.capture(), any());
        RepoConfigPropagationModel propagationModel = captor.getValue();
        assertEquals(0, propagationModel.getRepoConfigCreateOrUpdate().size());
        assertEquals(1, propagationModel.getRepoConfigDelete().size());
        RepoConfigReload repoConfigDeletion = propagationModel.getRepoConfigDelete().iterator().next();
        assertEquals("repo", repoConfigDeletion.getRepoKey());
        assertEquals(LOCAL, repoConfigDeletion.getRepoDetailsType());
        assertEquals(Generic, repoConfigDeletion.getRepoType());
        assertEquals(RepoConfigChange.ConfigAction.REPO_DELETE, repoConfigDeletion.getConfigAction());
    }

    @Test
    void syncRepoCacheWhenRepoExistInDbButNotInCache() {
        // given - cache does not contain repo
        Supplier<Map<String, RepoConfig>> getAllReposFunction = mock(Supplier.class);
        when(getAllReposFunction.get()).thenReturn(Map.of());
        Supplier<Map<String, Long>> getAllRepoKeys = mock(Supplier.class);
        when(getAllRepoKeys.get()).thenReturn(Map.of("repo", 0L));

        // given - db contains repo
        Function<String, RepoConfig> repoConfigByKeyFunction = mock(Function.class);
        RepoConfig repoConfig = LocalRepoConfig.create("repo", Generic);
        when(repoConfigByKeyFunction.apply("repo")).thenReturn(repoConfig);

        BiFunction<RepoConfigPropagationModel, RepoConfigCache, Void> repoConfigurationsFunction =
                mock(BiFunction.class);

        RepoConfigCache repoConfigCache = spy(new RepoConfigCache(Integer.MAX_VALUE, Long.MAX_VALUE, Integer.MAX_VALUE,
                Double.MAX_VALUE, getAllReposFunction, repoConfigByKeyFunction, null, repoConfigurationsFunction,
                getAllRepoKeys));

        // when
        repoConfigCache.syncRepoConfigKeys();

        // then - call reload with REPO_CREATE action
        verify(getAllRepoKeys, atLeastOnce()).get();
        ArgumentCaptor<RepoConfigPropagationModel> captor = ArgumentCaptor.forClass(RepoConfigPropagationModel.class);
        verify(repoConfigurationsFunction, atLeastOnce()).apply(captor.capture(), any());
        RepoConfigPropagationModel propagationModel = captor.getValue();
        assertEquals(1, propagationModel.getRepoConfigCreateOrUpdate().size());
        assertEquals(0, propagationModel.getRepoConfigDelete().size());
        RepoConfigReload configReload = propagationModel.getRepoConfigCreateOrUpdate().iterator().next();
        assertEquals("repo", configReload.getRepoKey());
        assertEquals(LOCAL, configReload.getRepoDetailsType());
        assertEquals(Generic, configReload.getRepoType());
        assertEquals(RepoConfigChange.ConfigAction.REPO_CREATE, configReload.getConfigAction());
    }

    private RepoConfig findRepoConfigByKey(String key, Collection<RepoConfig> repoConfigs) {
        return repoConfigs.stream()
                .filter(config -> config.getKey().equals(key))
                .findFirst()
                .orElse(null);
    }

    private void awaitCyclicBarrier(CyclicBarrier cyclicBarrier) {
        try {
            cyclicBarrier.await();
        } catch (InterruptedException | BrokenBarrierException e) {
            throw new RuntimeException(e);
        }
    }

    private void getFuture(Future future) {
        try {
            future.get();
        } catch (InterruptedException | ExecutionException e) {
            throw new RuntimeException(e);
        }
    }
}