/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.traffic;

import org.apache.commons.lang3.StringUtils;
import org.artifactory.api.security.SecurityService;
import org.artifactory.security.access.AccessTokenAuthentication;
import org.artifactory.traffic.v2.TrafficEntryV2;
import org.jfrog.access.common.ServiceId;
import org.jfrog.access.common.ServiceType;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.security.core.Authentication;

import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;

import static org.apache.commons.lang.StringUtils.startsWith;
import static org.artifactory.addon.xray.XrayAddon.ARTIFACTORY_XRAY_USER;
import static org.artifactory.api.security.access.UserTokenSpec.ARTIFACTORY_FEDERATION_USER_PREFIX;
import static org.artifactory.api.security.access.UserTokenSpec.TOKEN_PREFIX;

/**
 * Utility class for the traffic module
 *
 * @author Noam Tenne
 */
public abstract class TrafficUtils {

    private static final String XRAY_REPETITIVE = "XRAY_REPETITIVE";
    private static final String BA_BILLING_LOGSHIPPING = "BA_BILLING_LOGSHIPPING";
    private static final String XRAY_CREDIT = "XRAY_CREDIT";

    private static final Map<String, TrafficEntryV2.TrafficActor> ACTORS_MAP = new HashMap<>();
    private static final Logger log = LoggerFactory.getLogger(TrafficUtils.class);

    static {
        ACTORS_MAP.put(ServiceType.XRAY, TrafficEntryV2.TrafficActor.XRAY);
        ACTORS_MAP.put(ServiceType.XRAY_INDEXER, TrafficEntryV2.TrafficActor.XRAY);
        ACTORS_MAP.put(ServiceType.XRAY_PERSIST, TrafficEntryV2.TrafficActor.XRAY);
        ACTORS_MAP.put(ServiceType.XRAY_ANALYSIS, TrafficEntryV2.TrafficActor.XRAY);
        ACTORS_MAP.put(ServiceType.PIPELINES, TrafficEntryV2.TrafficActor.PIPELINES);
        ACTORS_MAP.put(ServiceType.REPLICATOR, TrafficEntryV2.TrafficActor.REPLICATOR);
        ACTORS_MAP.put(ServiceType.JF_CONNECT, TrafficEntryV2.TrafficActor.JFCONNECT);
        ACTORS_MAP.put(ServiceType.OBSERVABILITY, TrafficEntryV2.TrafficActor.OBSERVABILITY);
        ACTORS_MAP.put(XRAY_REPETITIVE, TrafficEntryV2.TrafficActor.XRAY_REPETITIVE);
        ACTORS_MAP.put(XRAY_CREDIT, TrafficEntryV2.TrafficActor.XRAY_CREDIT);
        ACTORS_MAP.put(BA_BILLING_LOGSHIPPING, TrafficEntryV2.TrafficActor.BA_BILLING_LOGSHIPPING);
    }

    private TrafficUtils() {
        // utility class
    }

    /**
     * Performs a comparison that checks if the date to compare is equals or before the date we want to compare to
     *
     * @param to Compare Date to compare
     * @param to Compare to
     * @return boolean - If the date to compare is equals or before the date we compare to
     */
    public static boolean dateEqualsBefore(Date toCompare, Date to) {
        return toCompare.compareTo(to) < 1;
    }

    /**
     * Performs a comparison that checks if the date to compare is equals or after the date we want to compare to
     *
     * @param to Compare Date to compare
     * @param to Compare to
     * @return boolean - If the date to compare is equals or after the date we compare to
     */
    public static boolean dateEqualsAfter(Date toCompare, Date to) {
        return toCompare.compareTo(to) > -1;
    }

    public static TrafficEntryV2.TrafficActor getTrafficActor(String userAgent, String username,
            Authentication authentication) {
        String serviceType = getServiceType(authentication);
        TrafficEntryV2.TrafficActor trafficActor =
                ACTORS_MAP.getOrDefault(serviceType, getTrafficActor(userAgent, username));
        log.trace("Traffic actor is {}", trafficActor);
        return trafficActor;
    }

    public static TrafficEntryV2.TrafficActor getTrafficActor(String identifier, boolean xrayAuthenticationUser,
            Authentication authentication) {
        if (StringUtils.isNotBlank(identifier)) {
            if (xrayAuthenticationUser) {
                String serviceType = getServiceType(authentication);
                return ACTORS_MAP.getOrDefault(serviceType, TrafficEntryV2.TrafficActor.XRAY);
            }
            if (startsWith(identifier.toLowerCase(), "jfrog-replicator/")) {
                return TrafficEntryV2.TrafficActor.REPLICATOR;
            }
            if (startsWith(identifier.toLowerCase(), "observability")) {
                return TrafficEntryV2.TrafficActor.OBSERVABILITY;
            }
            if (startsWith(identifier.toLowerCase(), "ba-billing-logshipping")) {
                return TrafficEntryV2.TrafficActor.BA_BILLING_LOGSHIPPING;
            }
        }
        return TrafficEntryV2.TrafficActor.USER;
    }

    private static String getServiceType(Authentication authentication) {
        String result = "";
        if (authentication instanceof AccessTokenAuthentication accessTokenAuthentication) {
            String subject = accessTokenAuthentication.getAccessToken().getSubject();
            if (StringUtils.isNotBlank(subject)) {
                try {
                    ServiceId serviceId = ServiceId.fromFormattedName(subject);
                    String serviceType = serviceId.getServiceType();
                    if (ServiceType.XRAY.equals(serviceType)) {
                        serviceType = handleXraySubject(subject);
                    }
                    result = serviceType;
                    log.trace("Service type is {}", result);
                } catch (Exception e) {
                    log.trace("Error getting service id from subject {}", subject, e);
                }
            }
        } else {
            if (authentication == null) {
                log.trace("Authentication is null");
            } else {
                log.trace("Authentication class {}", authentication.getClass().getName());
            }
        }
        return result;
    }

    private static String handleXraySubject(String subject) {
        return Arrays.stream(XraySpecialSubjects.values())
                .filter(val -> subject.startsWith(val.getSubject()))
                .map(XraySpecialSubjects::getActor)
                .findFirst().orElse(ServiceType.XRAY);
    }

    private static TrafficEntryV2.TrafficActor getTrafficActor(String userAgent, String username) {
        log.trace("Get traffic actor using useragent: {} username: {}", userAgent, username);
        if (ARTIFACTORY_XRAY_USER.equals(username)) {
            return TrafficEntryV2.TrafficActor.XRAY;
        }

        if (StringUtils.isNoneBlank(username) &&
                username.startsWith(TOKEN_PREFIX + ARTIFACTORY_FEDERATION_USER_PREFIX)) {
            return TrafficEntryV2.TrafficActor.FEDERATION;
        }

        String lowercased = userAgent == null ? "" : userAgent.toLowerCase();
        if (lowercased.contains("xray")) {
            return TrafficEntryV2.TrafficActor.XRAY;
        }
        if (lowercased.contains("jfrog-replicator/")) {
            return TrafficEntryV2.TrafficActor.REPLICATOR;
        }
        if (lowercased.contains("observability") && username.equals(SecurityService.USER_SYSTEM)) {
            return TrafficEntryV2.TrafficActor.OBSERVABILITY;
        }
        if (lowercased.contains("jfconnect") && username.equals(SecurityService.USER_SYSTEM)) {
            return TrafficEntryV2.TrafficActor.JFCONNECT;
        }
        if (lowercased.contains("ba-billing-logshipping") && username.equals(SecurityService.USER_SYSTEM)) {
            return TrafficEntryV2.TrafficActor.BA_BILLING_LOGSHIPPING;
        }

        return TrafficEntryV2.TrafficActor.USER;
    }

}
