package org.artifactory.opentracing;

import java.util.concurrent.Callable;

/**
 * @author Uriah Levy
 */
public class TraceableCallableDecorator<V> implements Callable<V> {
    private final ArtifactoryTracer artifactoryTracer;
    private final Callable<V> callable;
    private final ServerTracingSpan currentSpan;

    public TraceableCallableDecorator(Callable<V> callable, ArtifactoryTracer artifactoryTracer, boolean inheritSpan) {
        this.callable = callable;
        this.artifactoryTracer = artifactoryTracer;
        this.currentSpan = inheritSpan ? artifactoryTracer.getCurrentThreadActiveSpan() : null;
    }

    @Override
    public V call() throws Exception {
        String original = Thread.currentThread().getName();
        try {
            if (currentSpan != null) {
                // Set the current span as active for the current thread
                try (ServerTracingScope ignored = artifactoryTracer.activateSpan(currentSpan)) {
                    ThreadNameTraceUtil.enhanceThreadName();
                    return callable.call();
                }
            }
            // Start a new span that ends at the end of this invocation
            try (ServerTracingScope ignored = artifactoryTracer.startSpanOnCurrentThread(
                    "Artifactory Async Invocation")) {
                ThreadNameTraceUtil.enhanceThreadName();
                return callable.call();
            }
        } finally {
            Thread.currentThread().setName(original);
        }
    }
}
