package org.artifactory.request;

import org.artifactory.aol.AolService;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.common.ConstantValues;
import org.artifactory.repo.http.DenyListCacheHolderService;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.mockito.Mockito;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;
import org.xbill.DNS.lookup.ServerFailedException;

import java.net.URISyntaxException;
import java.net.UnknownHostException;

import static org.mockito.Mockito.any;
import static org.mockito.Mockito.doThrow;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertFalse;
import static org.testng.Assert.assertTrue;

/**
 * @author Nadav Yogev
 */
public class UrlVerifierTest extends ArtifactoryHomeBoundTest {

    @BeforeMethod
    public void setup() {
        ArtifactoryContext context = Mockito.mock(ArtifactoryContext.class);
        AolService aolService = Mockito.mock(AolService.class);
        when(aolService.isAol()).thenReturn(true);
        when(context.beanForType(AolService.class)).thenReturn(aolService);
        ArtifactoryContextThreadBinder.bind(context);
    }

    @AfterMethod
    public void tearDown() {
        ArtifactoryContextThreadBinder.unbind();
    }

    @Test
    public void testLoopBack() {
        UrlVerifier urlVerifier = createUrlVerifier();
        boolean loopback = urlVerifier.isRemoteRepoBlocked("http://localhost:8080/artifactory/dangerZone", "test");
        assertTrue(loopback);

        loopback = urlVerifier.isRemoteRepoBlocked("http://127.0.0.1:8080/artifactory/dangerZone", "test");
        assertTrue(loopback);

        loopback = urlVerifier.isRemoteRepoBlocked("http://[::1]:8080/artifactory/dangerZone", "test");
        assertTrue(loopback);

        loopback = urlVerifier.isRemoteRepoBlocked("http://jfrog.com/artifactory/dangerZone", "test");
        assertFalse(loopback);

        loopback = urlVerifier.isRemoteRepoBlocked("http://jfrog.com/artifactory/dangerZone", "test");
        assertFalse(loopback);

        // test whitelist (localhost with /)
        getBound().setProperty(ConstantValues.whitelistRemoteRepoUrls, "https://127.0.0.1,http://localhost/");
        urlVerifier = createUrlVerifier();
        loopback = urlVerifier.isRemoteRepoBlocked("http://localhost/artifactory/dangerZone", "test");
        assertFalse(loopback);
        urlVerifier = createUrlVerifier();
        loopback = urlVerifier.isRemoteRepoBlocked("http://127.0.0.1:8080/artifactory/dangerZone", "test");
        assertTrue(loopback);
        urlVerifier = createUrlVerifier();
        loopback = urlVerifier.isRemoteRepoBlocked("https://127.0.0.1:8080/artifactory/dangerZone", "test");
        assertFalse(loopback);
    }

    @Test
    public void testSiteAndLinkLocal() {
        getBound().setProperty(ConstantValues.remoteRepoBlockUrlStrictPolicy, "true");
        UrlVerifier urlVerifier = createUrlVerifier();
        boolean loopback = urlVerifier.isRemoteRepoBlocked("http://10.0.0.1:8080/internal/dangerZone", "test");
        assertTrue(loopback);

        urlVerifier = createUrlVerifier();
        loopback = urlVerifier.isRemoteRepoBlocked("http://169.254.169.254/latest/meta-data/dangerZone", "test");
        assertTrue(loopback);
    }

    @Test
    public void shouldParseDomainName() throws URISyntaxException {
        assertEquals(UrlVerifier.getDomainName("http://domain.com"), "domain.com");
        assertEquals(UrlVerifier.getDomainName("http://www.domain.com"), "domain.com");
        assertEquals(UrlVerifier.getDomainName("https://www.domain.com"), "domain.com");
        assertEquals(UrlVerifier.getDomainName("https://www.domain.com/one/two.jsp"), "domain.com");
        assertEquals(UrlVerifier.getDomainName("https://192.168.2.3:8080/one/two.jsp"), "192.168.2.3");
        // not parsed because of "_1" in domain name.
        assertEquals(UrlVerifier.getDomainName("https://www.domain_1.com/one/two.jsp"),
                "https://www.domain_1.com/one/two.jsp");
    }

    @Test(expectedExceptions = URISyntaxException.class)
    public void badUriFailed() throws URISyntaxException {
        UrlVerifier.getDomainName("//o\\");
    }


    @Test
    public void shouldExecuteCalculateRepoBlockedOnceForDomain() throws UnknownHostException, URISyntaxException {
        UrlVerifier urlVerifierSpy = Mockito.spy(createUrlVerifier());
        final String url = "http://domain.com/externalDomain";
        final String url1 = "https://domain.com/differentExternalUrl";
        final String domain = UrlVerifier.getDomainName(url);
        final String repoKey = "test";
        assertFalse(urlVerifierSpy.isRemoteRepoBlocked(url, repoKey));
        assertFalse(urlVerifierSpy.isRemoteRepoBlocked(url, repoKey));
        assertFalse(urlVerifierSpy.isRemoteRepoBlocked(url1, repoKey));

        //result calculated only once for domain and retrieved two other times from the cache.
        Mockito.verify(urlVerifierSpy, times(1)).isRepoBlocked(domain, url, repoKey);
    }

    @Test
    public void shouldExecuteCalculateRepoBlockedThreeTimesForDifferentDomains()
            throws URISyntaxException, UnknownHostException {
        UrlVerifier urlVerifierSpy = Mockito.spy(createUrlVerifier());
        final String url1 = "http://domaina.com/externalDomain";
        final String url2 = "https://domainb.com/differentExternalUrl";
        final String url3 = "https://domainc.com/differentExternalUrl";

        final String domain1 = UrlVerifier.getDomainName(url1);
        final String domain2 = UrlVerifier.getDomainName(url2);
        final String domain3 = UrlVerifier.getDomainName(url3);
        final String repoKey = "test";
        assertFalse(urlVerifierSpy.isRemoteRepoBlocked(url1, repoKey));
        assertFalse(urlVerifierSpy.isRemoteRepoBlocked(url2, repoKey));
        assertFalse(urlVerifierSpy.isRemoteRepoBlocked(url3, repoKey));

        //result calculated only once per domain.
        Mockito.verify(urlVerifierSpy, times(1)).isRepoBlocked(domain1, url1, repoKey);
        Mockito.verify(urlVerifierSpy, times(1)).isRepoBlocked(domain2, url2, repoKey);
        Mockito.verify(urlVerifierSpy, times(1)).isRepoBlocked(domain3, url3, repoKey);

        //calculated three times altogether.
        Mockito.verify(urlVerifierSpy, times(3)).isRepoBlocked(any(), any(), any());
    }

    @Test(dataProvider = "isRemoteRepoBlockedShouldCatchRuntimeExceptionProvider")
    public void isRemoteRepoBlockedShouldCatchRuntimeException(Exception exceptionToCatch) throws UnknownHostException {
        UrlVerifier urlVerifierSpy = Mockito.spy(createUrlVerifier());
        final String url = "http://domain.com/externalDomain";
        final String repoKey = "test";
        doThrow(exceptionToCatch).when(urlVerifierSpy).isRepoBlocked(any(), any(), any());
        assertTrue(urlVerifierSpy.isRemoteRepoBlocked(url, repoKey));
        Mockito.verify(urlVerifierSpy, times(1)).isRepoBlocked(any(), any(), any());
    }

    private UrlVerifier createUrlVerifier() {
        return new UrlVerifier(new DenyListCacheHolderService());
    }

    @DataProvider
    private static Object[][] isRemoteRepoBlockedShouldCatchRuntimeExceptionProvider() {
        return new Object[][] {
                {new ServerFailedException()},
                {new UnknownHostException()}
        };
    }
}