package org.artifactory.repo;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Component;

import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import static org.artifactory.repo.PackageConfigConverterEnum.getRelevantConverters;

/**
 * @author Tamar Rosen
 */
@Component
public class PackageTypeConfigConverter implements RepositoriesConfigSubConverter {
    private static final Logger log = LoggerFactory.getLogger(PackageTypeConfigConverter.class);
    public static final String PACKAGE_TYPE_CONFIG_KEY = "package_type_config";
    private static final int VERSION = 1;

    @Override
    public boolean isInterested(String key) {
        return PACKAGE_TYPE_CONFIG_KEY.equals(key);
    }

    @Override
    public Map<String, Object> convert(Map<String, Object> toConvert) {
        if (!toConvert.containsKey("modelVersion")) {
            return toConvert;
        }
        Map<String, String> flatStringMap = toConvert.entrySet().stream()
                .collect(Collectors.toMap(Map.Entry::getKey, entry -> (String) entry.getValue()));
        int modelVersion = Integer.parseInt(flatStringMap.get("modelVersion"));
        List<PackagesConverter> converters = getRelevantConverters(modelVersion);
        if (converters.isEmpty()) {
            log.debug("No Packages Config converters found between version {} and {}", modelVersion, VERSION);
        } else {
            log.info("Starting Packages Config conversion from {} to {}", modelVersion, VERSION);
            for (PackagesConverter converter : converters) {
                if (converter.isInterested(modelVersion)) {
                    converter.convert(flatStringMap);
                } else {
                    log.info("Not running Packages Config converter '{}' (converter criteria not met)",
                            converter.getClass().getSimpleName());
                }
            }
            log.info("Finished Packages Config conversion from {} to {}", modelVersion, VERSION);
        }
        return Collections.unmodifiableMap(flatStringMap);
    }
}
