package org.artifactory.core.api.ph.service;

import lombok.NonNull;
import org.artifactory.common.repo.PackageType;
import org.artifactory.common.repo.RepoInfo;
import org.artifactory.core.api.ph.exception.PackageException;
import org.artifactory.core.api.ph.exception.PackageVerificationException;
import org.artifactory.core.api.ph.model.PackageArtifact;

import java.io.InputStream;

/**
 * The PackageInfoAdapter handles package type verifications over specific artifacts paths.
 * The Implementation is provided by the package handler (in OSS, a default empty bean is loaded by the core)
 *
 * @author BarakH
 */
public interface PackageSpecificationAdapter {

    /**
     * Indicates whether the specified path is a package in the given repository key.
     *
     * @return true if the artifact path represent a package
     */
    boolean isPackageArtifact(@NonNull String repoKey, @NonNull String path) throws PackageException;

    /**
     * Indicates whether the specified path is metadata path
     *
     * @return true if the path is metadata, false otherwise
     */
    boolean isExpirable(@NonNull PackageType packageType, @NonNull String path) throws PackageException;

    /**
     * Indicates whether the file is generated by Artifactory.
     *
     * @return true if the file should not be replicated (either generated or tmp).
     */
    boolean isLocalGenerated(@NonNull PackageType packageType, @NonNull String path) throws PackageException;

    /**
     * Indicates whether the property is locally generated.
     *
     * @return True if the property key can be generated natively when deploying a new package to the repo.
     */
    boolean isLocallyGeneratedProperty(@NonNull PackageType packageType, @NonNull String propertyKey)
            throws PackageException;

    /**
     * The method is used by Artifactory to identify whether the package is supported or not by the new framework
     *
     * @param packageType the package type to check
     * @return true if the package is supported
     */
    boolean isSupportedPackage(PackageType packageType) throws PackageException;


    /**
     * The method is used by Artifactory to identify whether the package type of the repo is supported or not by the
     * new framework
     *
     * @param repoKey the repoKey type to check
     * @return true if the package type of the repo is supported
     */
    boolean isSupportedPackage(String repoKey) throws PackageException;

    /**
     * The method is used by Artifactory to identify whether the package type & repo type are supporting reindex
     */
    boolean hasReindex(RepoInfo repoInfo);

    /**
     * The method is used by Artifactory to identify whether the artifact content can be viewed in the UI.
     */
    boolean isViewable(String repoKey, String path);

    /**
     * The method is being triggered synchronously by The StorageInterceptor, during artifact upload in remote-cache
     * repository (right after it was initially downloaded).
     * It is used by Artifactory to determine if the content of the newly downloaded remote artifact should get
     * verified.
     */
    boolean shouldVerifyRemoteArtifact(PackageArtifact artifact);

    /**
     * The method is being triggered synchronously by The StorageInterceptor, during artifact upload in remote-cache
     * repository (right after it was initially downloaded).
     * It is used by Artifactory to determine if the remote artifact should not get cached after it was downloaded.
     * Note that it might be called within transaction
     *
     * @param artifact Artifact metadata related to the created artifact
     * @throws PackageVerificationException thrown in case of cancel, and it will break the download with cancellation.
     *                                      All others exceptions will seep up
     */
    void verifyRemoteArtifactNotCorrupted(PackageArtifact artifact, InputStream is) throws PackageVerificationException;

    /**
     * return true if the artifact's current properties (before they get overridden) should get saved as additional
     * snapshot for later usage in the afterCreate/afterPropertyChange callback
     *
     * @param artifactPath - path of the artifact
     */
    boolean shouldMarkOriginalPropertiesInBeforeEvents(String repoKey, String artifactPath);
}
