/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.workqueue;

import org.jetbrains.annotations.NotNull;

import java.util.Map;

/**
 * @author gidis
 */
public interface AsyncWorkQueueService {

    /**
     * Get a work queue. If no work queue exists yet, create one.
     */
    WorkQueue<WorkItem> getWorkQueue(String name);

    /**
     * Get all existing work queues.
     *
     * @return all existing work queues.
     */
    Map<String, WorkQueue<WorkItem>> getExistingWorkQueues();

    void closeAllQueues();

    int getEstimatedPendingTasksSize(String workQueueName);

    /**
     * Execute the runnable in a work queue for the specified work item
     *
     * @param workQueueName     the work queue name
     * @param workItem          the work item to execute
     * @param runnable          the runnable for the work item
     * @param wait4Done         wait for done
     * @param delayAfterCommit  delay the execution after commit
     * @param failIfNoSession   fail if delay after commit is required and there is no session(TX)
     */
    void execute(@NotNull String workQueueName, @NotNull WorkItem workItem, @NotNull Runnable runnable,
            boolean wait4Done, boolean delayAfterCommit, boolean failIfNoSession);

    /**
     * Execute the runnable in a work queue for the specified work item
     *
     * @param workQueueName the work queue name
     * @param workItem      the work item to execute
     * @param runnable      the runnable for the work item
     * @param wait4Done     wait for done
     */
    void execute(@NotNull String workQueueName, @NotNull WorkItem workItem, @NotNull Runnable runnable,
            Boolean wait4Done);


    /**
     * Execute the runnable in a work queue for the specified work item
     * with no wait for done and delay after commit options.
     *
     * @param workQueueName    the work queue name
     * @param workItem         the work item to execute
     * @param runnable         the runnable for the work item
     * @param delayAfterCommit delay the execution after commit
     * @param failIfNoSession  fail if delay after commit is required and there is no session(TX)
     */
    void execute(@NotNull String workQueueName, @NotNull WorkItem workItem, @NotNull Runnable runnable,
            boolean delayAfterCommit, boolean failIfNoSession);

    /**
     * Execute the runnable in a work queue for the specified work item and not wait for done
     *
     * @param workQueueName the work queue name
     * @param workItem      the work item to execute
     * @param runnable      the runnable for the work item
     */
    default void execute(@NotNull String workQueueName, @NotNull WorkItem workItem,
            @NotNull Runnable runnable) {
        execute(workQueueName, workItem, runnable, false);
    }

    /**
     * Execute clean job:
     * remove work queues if repo not exists
     * execute stack items
     */
    void clean();

    /**
     * Start work queue job with decorator
     */
    void scheduleWorkQueueCleanupJob(WorkQueueJobDecorator decorator);

}