/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.workqueue;

import ch.qos.logback.classic.Level;
import com.google.common.collect.Lists;
import com.jfrog.test.LogTestUtils;
import org.artifactory.async.AsyncExecutor;
import org.fest.assertions.Assertions;
import org.jfrog.common.concurrency.ConflictsGuard;
import org.jfrog.common.concurrency.JvmConflictsGuard;
import org.jfrog.config.utils.IndependentCondition;
import org.testng.Assert;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.Test;

import java.time.Instant;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.RejectedExecutionException;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyLong;
import static org.mockito.Mockito.doAnswer;
import static org.mockito.Mockito.doThrow;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.never;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.fail;

/**
 * Unit tests for the {@link WorkQueueImpl}.
 *
 * @author Yossi Shaul
 */
@Test
public class WorkQueueImplTest {

    private static final String QUEUE_NAME = "Test WQ";

    @BeforeClass
    public void setup() {
        LogTestUtils.setLoggingLevel(WorkQueueImpl.class, Level.INFO);
    }

    /**
     * Verify that adding workers affect the available permits
     *
     * @throws Exception
     */
    public void addWorkers() throws Exception {

        WorkQueueImpl<IntegerWorkItem> wq =
                new WorkQueueImpl<>(QUEUE_NAME, 5, mock(AsyncExecutor.class),
                        mock(ConflictsGuard.class), 60, 60);

        Assertions.assertThat(wq.availablePermits()).isEqualTo(5);

        wq.updateWorkers(10);

        Assertions.assertThat(wq.availablePermits()).isEqualTo(10);

    }

    /**
     * Verify that reducing workers affects the available permits
     *
     * @throws Exception
     */
    public void freeWorkers() throws Exception {

        WorkQueueImpl<IntegerWorkItem> wq =
                new WorkQueueImpl<>(QUEUE_NAME, 5, mock(AsyncExecutor.class),
                        mock(ConflictsGuard.class), 60, 60);

        Assertions.assertThat(wq.availablePermits()).isEqualTo(5);

        wq.updateWorkers(3);

        Assertions.assertThat(wq.availablePermits()).isEqualTo(3);

    }

    /**
     * Verify that reduce used workers affect the available permits after a used worker is being free
     *
     * @throws Exception
     */
    @Test(expectedExceptions = InterruptedException.class)
    public void freeUsedWorkers() throws Exception {

        WorkQueueImpl<IntegerWorkItem> wq =
                new WorkQueueImpl<>(QUEUE_NAME, 5, mock(AsyncExecutor.class),
                        mock(ConflictsGuard.class), 60, 60);

        Assertions.assertThat(wq.availablePermits()).isEqualTo(5);

        wq.tryAcquire();
        wq.tryAcquire();

        Assertions.assertThat(wq.availablePermits()).isEqualTo(3);

        new IndependentCondition().waitForConditionOnCall(() -> false, TimeUnit.SECONDS, 5L);

        wq.updateWorkers(1);

        fail();

    }

    @Test(expectedExceptions = IllegalArgumentException.class)
    public void executeNullWorkItem() {
        WorkQueueImpl<IntegerWorkItem> wq =
                spy(new WorkQueueImpl<>(QUEUE_NAME, 1, mock(AsyncExecutor.class),
                        mock(ConflictsGuard.class), 60, 60));

        wq.execute(null, () -> {});
    }

    @Test(expectedExceptions = IllegalArgumentException.class)
    public void executeNullRunnable() {
        WorkQueueImpl<IntegerWorkItem> wq =
                spy(new WorkQueueImpl<>(QUEUE_NAME, 1, mock(AsyncExecutor.class),
                        mock(ConflictsGuard.class), 60, 60));

        wq.execute(new IntegerWorkItem(1), null);
    }

    /**
     * verify that execute a work queue acquire permit and executed in the async executor
     *
     * @throws Exception
     */
    public void executeAcquired() throws Exception {

        AsyncExecutor executor = mock(AsyncExecutor.class);

        WorkQueueImpl<IntegerWorkItem> wq =
                spy(new WorkQueueImpl<>(QUEUE_NAME, 1, executor,
                        mock(ConflictsGuard.class), 60, 60));

        wq.execute();

        verify(wq, times(1)).tryAcquire();
        verify(wq, never()).release();

        Assertions.assertThat(wq.availablePermits()).isEqualTo(0);
        verify(executor, times(1)).execute(any(Runnable.class));
    }

    /**
     * Verify that execute a work queue with no permit, is not consuming permit and not executed in the async executor
     */
    public void executeNotAcquired() {
        AsyncExecutor executor = mock(AsyncExecutor.class);

        WorkQueueImpl<IntegerWorkItem> wq =
                spy(new WorkQueueImpl<>(QUEUE_NAME, 0, executor,
                        mock(ConflictsGuard.class), 60, 60));

        wq.execute();

        verify(wq, times(1)).tryAcquire();
        verify(wq, never()).release();

        Assertions.assertThat(wq.availablePermits()).isEqualTo(0);
        verify(executor, never()).execute(any(Runnable.class));
    }

    /**
     * Verify that a rejected work item (by the thread pool queue) the available permit is finally not affected
     * (Acquire and released)
     */
    public void executeRejected() {
        AsyncExecutor executor = mock(AsyncExecutor.class);
        doThrow(new RejectedExecutionException()).when(executor).execute(any(Runnable.class));
        WorkQueueImpl<IntegerWorkItem> wq =
                spy(new WorkQueueImpl<>(QUEUE_NAME, 1, executor,
                        mock(ConflictsGuard.class), 60, 60));

        wq.execute();

        verify(wq, times(1)).tryAcquire();
        verify(wq, times(1)).release();

        Assertions.assertThat(wq.availablePermits()).isEqualTo(1);
        verify(executor, times(1)).execute(any(Runnable.class));
    }


    /**
     * Verify that item which was not promoted due to long run is re-executed
     *
     * @throws Exception
     */
    public void testShouldNotPromote() throws Exception {

        AsyncExecutor executor = mock(AsyncExecutor.class);
        WorkQueueImpl<IntegerWorkItem> wq =
                spy(new WorkQueueImpl<>(QUEUE_NAME, 1, executor,
                        mock(ConflictsGuard.class), 60, 60));

        doAnswer(invocation -> false).when(wq).shouldPromote(any(Instant.class));
        when(wq.getQueueSize()).thenReturn(1);

        wq.doJobs();

        verify(wq, times(1)).execute();
    }

    /**
     * Verify that item which was not promoted due to long run is re-executed
     *
     * submit 2 items, first item with sleep inorder to prevent the direct execution of the next item
     *
     * The next item, should be re-executed through the executor
     *
     * @throws Exception
     */
    public void testShouldNotPromoteExecTimeElapsed() throws Exception {
        try (TestAsyncExecutor executor = spy(new TestAsyncExecutor(1, 5))) {
            WorkQueueImpl<IntegerWorkItem> wq =
                    spy(new WorkQueueImpl<>(QUEUE_NAME, 1, executor,
                            new JvmConflictsGuard<>(1), 1, 60));

            //A countdown to wait until 0
            CountDownLatch count = new CountDownLatch(2);

            //ofer item 1
            wq.offerWork(new IntegerWorkItem(1), () -> {
                try {
                    //Sleep 2 sec in order to not promote the next item
                    Thread.sleep(2000);
                    count.countDown();
                    when(executor.getPendingTasks()).thenReturn(1);
                } catch (InterruptedException e) {
                }
            });

            // ofer item 2
            wq.offerWork(new IntegerWorkItem(2), () -> count.countDown());

            //execute both items
            wq.execute();

            if (!count.await(5000, TimeUnit.SECONDS)) {
                fail(wq.getQueueSize() + " item(s) were not executed");
            }

            verify(wq, times(2)).execute();
            verify(wq, times(2)).doJobs();
            assertEquals(wq.getQueueSize(), 0);
        }
    }

    /**
     * Verify should promote when item exec time is short and there are pending thread pool tasks
     */
    public void testShouldPromoteTimeNotElapsedAndPendingThreadPoolTasks() {
        AsyncExecutor executor = mock(AsyncExecutor.class);
        when(executor.getPendingTasks()).thenReturn(1);

        WorkQueueImpl<IntegerWorkItem> wq =
                new WorkQueueImpl<>(QUEUE_NAME, 1, executor,
                        mock(ConflictsGuard.class), 60, 60);

        Assertions.assertThat(wq.shouldPromote(Instant.now())).isTrue();
    }

    /**
     * Verify should promote when item exec time is short and time there are no pending thread pool tasks
     */
    public void testShouldPromoteTimeElapsedAndNoPendingThreadPoolTasks() {
        AsyncExecutor executor = mock(AsyncExecutor.class);
        when(executor.getPendingTasks()).thenReturn(0);
        WorkQueueImpl<IntegerWorkItem> wq =
                new WorkQueueImpl<>(QUEUE_NAME, 1, executor,
                        mock(ConflictsGuard.class), 60, 60);


        Assertions.assertThat(wq.shouldPromote(Instant.now().minusSeconds(61))).isTrue();
    }

    /**
     * Verify should not promote when item exec time is long and there are thread pool pending tasks
     */
    public void testShouldNotPromoteTimeElapsedAndPendingThreadPoolTasks() {
        AsyncExecutor executor = mock(AsyncExecutor.class);
        when(executor.getPendingTasks()).thenReturn(1);
        WorkQueueImpl<IntegerWorkItem> wq =
                new WorkQueueImpl<>(QUEUE_NAME, 1, executor,
                        mock(ConflictsGuard.class), 60, 60);


        Assertions.assertThat(wq.shouldPromote(Instant.now().minusSeconds(61))).isFalse();
    }


    public void offerTheSameWorkConcurrently() throws Exception {
        List<AtomicInteger> counts = Collections.nCopies(10, new AtomicInteger(0));
        Counter counter = new Counter(counts);
        ConflictsGuard guard = mock(ConflictsGuard.class);
        when(guard.tryToLock(any(Object.class), anyLong(), any(TimeUnit.class))).thenReturn(true);

        try (TestAsyncExecutor executor = spy(new TestAsyncExecutor(20, 1500))) {
            WorkQueueImpl<IntegerWorkItem> wq =
                    new WorkQueueImpl<>("Test Queue", 1, executor, guard, 60, 60);

            for (int i = 0; i < 10; i++) {
                for (int j = 0; j < 10000; j++) {
                    IntegerWorkItem workItem = new IntegerWorkItem(i);
                    wq.execute(workItem, () -> counter.count(workItem));
                }
            }
        }

        // We expect each job to be executed at least once
        counts.forEach(count -> Assertions.assertThat(count.get()).isGreaterThanOrEqualTo(1));
    }


    public void offerWorkWithNWorkers() throws Exception {
        List<AtomicInteger> counts = Collections.nCopies(1000, new AtomicInteger(0));
        Counter counter = new Counter(counts);
        ConflictsGuard guard = mock(ConflictsGuard.class);
        when(guard.tryToLock(any(Object.class), anyLong(), any(TimeUnit.class))).thenReturn(true);
        try (TestAsyncExecutor executor = spy(new TestAsyncExecutor(20, 1500))) {

            WorkQueueImpl<IntegerWorkItem> wq =
                    new WorkQueueImpl<>("Test Queue", 4, executor, guard, 60, 60);

            for (int i = 0; i < 10; i++) {
                for (int j = 0; j < 10; j++) {
                    IntegerWorkItem workItem = new IntegerWorkItem(i);
                    wq.execute(workItem, () -> counter.count(workItem));
                }
            }
        }

        // We expect each job to be executed at least once
        counts.forEach(count -> Assertions.assertThat(count.get()).isGreaterThanOrEqualTo(1));
    }

    // When waiting for item (blockUntilFinished = true), the item we are waiting for can either be pending, or
    //promoted state.
    // This tests a bug that when the item is promoted, we never actually waited for it to complete
    public void waitUntilWhenItemIsPromotedTest() throws Exception {
        SleepingDummyService dummyService = new SleepingDummyService();
        ConflictsGuard guard = mock(ConflictsGuard.class);
        when(guard.tryToLock(any(Object.class), anyLong(), any(TimeUnit.class))).thenReturn(true);
        when(guard.isLocked(any(Object.class))).thenReturn(true);

        try (TestAsyncExecutor executor = spy(new TestAsyncExecutor(10, 1500))) {
            WorkQueueImpl<DummyWorkItem> queue =
                    new WorkQueueImpl<>("Test Queue", 4, executor, guard, 60, 60);
            for (int i = 0; i < 20; i++) {
                DummyWorkItem workItem = new DummyWorkItem("x-" + i);
                queue.offerWork(workItem, () -> dummyService.serve(workItem));
                queue.execute();
                queue.waitForItemDone(workItem);
                Assert.assertTrue(workItem.isProcessed(), workItem.getUniqueKey() + " Failed");
            }
        }
    }

    public void waitUntilTest() throws Exception {
        List<AtomicInteger> counts = Collections.nCopies(10, new AtomicInteger(0));
        Counter counter = new Counter(counts);
        ConflictsGuard guard = mock(ConflictsGuard.class);
        when(guard.tryToLock(any(Object.class), anyLong(), any(TimeUnit.class))).thenReturn(true);
        try (TestAsyncExecutor executor = spy(new TestAsyncExecutor(10, 1500))) {
            WorkQueueImpl<IntegerWorkItem> wq =
                    new WorkQueueImpl<>("Test Queue", 4, executor, guard, 60, 60);

            ExecutorService testExecutor = Executors.newFixedThreadPool(10);
            try {
                ArrayList<Future> futures = Lists.newArrayList();
                for (int i = 0; i < 5; i++) {
                    futures.add(testExecutor.submit(() -> {
                        for (int j = 0; j < 10000; j++) {
                            IntegerWorkItem workItem = new IntegerWorkItem((int) (Math.random() * 10d));
                            wq.execute(workItem, () -> counter.count(workItem));
                            wq.waitForItemDone(workItem);
                        }
                        return null;
                    }));
                }
                for (Future future : futures) {
                    future.get();
                }
            } finally {
                testExecutor.shutdown();
                testExecutor.awaitTermination(1500, TimeUnit.SECONDS);
            }
        }
        // We expect each job to be executed at least once
        counts.forEach(count -> Assertions.assertThat(count.get()).isGreaterThanOrEqualTo(1));
    }

    class TestAsyncExecutor implements AsyncExecutor, AutoCloseable {

        private final ThreadPoolExecutor executor;
        private int awaitSec;

        public TestAsyncExecutor(int threadsCount, int awaitSec) {
            this.executor = (ThreadPoolExecutor) Executors.newFixedThreadPool(threadsCount);
            this.awaitSec = awaitSec;
        }

        @Override
        public void close() throws Exception {
            executor.shutdown();
            executor.awaitTermination(awaitSec, TimeUnit.SECONDS);
        }

        @Override
        public <T> Future<T> execute(Callable<T> task) {
            return executor.submit(task);
        }

        @Override
        public void execute(Runnable runnable) {
            executor.execute(runnable);
        }

        @Override
        public int getPendingTasks() {
            return executor.getQueue().size();
        }

        @Override
        public int getActiveTasks() {
            return executor.getActiveCount();
        }
    }
}
