/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.storage.binstore.service;

import org.artifactory.api.common.BasicStatusHolder;
import org.artifactory.storage.StorageException;
import org.jfrog.storage.binstore.exceptions.BinaryNotFoundException;
import org.jfrog.storage.binstore.ifc.UsageTracking;
import org.jfrog.storage.binstore.ifc.model.BinaryElement;
import org.jfrog.storage.binstore.ifc.model.BinaryElementRequest;
import org.jfrog.storage.binstore.ifc.model.BinaryProvidersResponseMap;
import org.jfrog.storage.binstore.ifc.model.StorageInfo;
import org.jfrog.storage.binstore.manager.BinaryProviderManager;
import org.jfrog.storage.binstore.providers.base.PersistedRemoteTask;
import org.springframework.transaction.annotation.Propagation;
import org.springframework.transaction.annotation.Transactional;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.io.File;
import java.sql.SQLException;
import java.util.Collection;
import java.util.List;
import java.util.Map;

/**
 * Date: 12/11/12
 * Time: 2:13 PM
 *
 * @author freds
 */
public interface InternalBinaryService extends BinaryService, UsageTracking {

    /**
     * Add the binary data line to this data store based on all params after verification
     * from binary provider delegate.
     * Verification steps depends on configuration and delegates answers.
     * Method will returns null if no binary object with specified info is found,
     * or no filestore delegate was configured.
     *
     * @param sha1   The SHA1 key value
     * @param md5    The md5 of this new entry
     * @param length The length of the binary
     * @return The full binary info object created and stored, or null if no object
     * @throws BinaryNotFoundException if verification failed
     */
    @Nullable
    BinaryElement addBinaryRecord(Map<String, String> headers, String sha1, String sha2, String md5, long length)
            throws BinaryNotFoundException;

    /**
     * Create the entry in DB for this binary if it does not exists already.
     * Returns the actual entry created.
     * Open and close a DB TX specifically for this.
     *
     * @return the actual entry in DB matching the inputs
     * @throws StorageException
     */
    @Nonnull
    @Transactional(propagation = Propagation.REQUIRES_NEW)
    BinaryElement insertRecordInDb(String sha1, String sha2, String md5, long length) throws StorageException;


    /**
     * Methos to use garbage collection.
     */
    void startGarbageCollect();

    /**
     * Activate Garbage Collection for this binary store
     *
     * @return The results of the GC process
     */
    GarbageCollectorInfo runFullGarbageCollect();

    /**
     * Run all (trash + binaries) garbage collection strategies
     */
    void runAllGarbageCollectStrategies();

    void deleteBlobInfos(List<String> deletedBinaries);

    /**
     * Delete all left over folders and files in the binaries directory that are not declared
     * at all in this binary store.
     *
     * @param statusHolder A status holder of the process messages
     */
    void prune(BasicStatusHolder statusHolder);

    /**
     * The actual local directory where checksum binary files will be stored.
     *
     * @return the folder for this binary store
     * @deprecated Cannot be used anymore
     */
    @Nullable
    @Deprecated
    File getBinariesDir();

    /**
     * Indicates if the current binary provider is a cloud based and supports redirection
     */
    boolean isCloudProviderSupportingRedirectionConfigured();

    /**
     * Map each file provider to it's filestore folder
     *
     * @return A map of each filestore folder mapped to it's provider
     */
    StorageInfo getStorageInfoSummary();

    BinaryProviderManager getBinaryProviderManager();

    void addGCListener(GarbageCollectorListener garbageCollectorListener);

    Collection<BinaryElement> findAllBinaries();

    /**
     * Deletes unreferenced archive paths. Shared archive paths might not be used after a binary is deleted.
     */
    @Transactional
    int deleteUnusedArchivePaths();

    /**
     * Deletes unreferenced archive names. Shared archive names might not be used after a binary is deleted.
     */
    @Transactional
    int deleteUnusedArchiveNames();

    /**
     * Increments the active users of a certain binary to prevent deletion while still in usage.
     *
     * @param sha1 The sha1 checksum to protect
     */
    @Override
    int incrementNoDeleteLock(String sha1);

    /**
     * Decrements the active users of a certain binary. Indicates that the active usage was ended.
     *
     * @param sha1 The sha1 checksum to remove protection from
     */
    @Override
    void decrementNoDeleteLock(String sha1);

    /**
     * BEWARE - this method is meant for a very specific usage - the sha2 db conversion.
     * NEVER EVER touch the binaries table outside of normal flow!
     */
    @Transactional(propagation = Propagation.REQUIRES_NEW)
    boolean updateSha2ForSha1(String targetSha1, String newSha2) throws SQLException;

    /**
     * Checks whether the requested {@param sha1} is in cloud provider - via cache of sha1 string only
     * (not the actual artifact), and verifies that redirect are allowed in that cloud provider.
     * Going out to real cloud provider only in case cache doesn't hold the sha1 in it.
     */
    boolean canCloudProviderGenerateRedirection(String sha1);

    /**
     * This method is meant for retention purposes only - restoring files from cold storage.
     *
     * Checks if file is restored from cold storage.
     * This will populate return object with a dedicated archive attribute to indicate the binary restore
     * state.
     *
     * @param request The binary request containing the restore candidate sha1
     * @return object with a dedicated archive attribute to indicate the binary restore state.
     */
    BinaryProvidersResponseMap isFileRestoredFromColdStorage(BinaryElementRequest request);

    /**
     * This method is meant for retention purposes only - restoring files from cold storage.
     *
     * Initiates the restore process in the scope of the archive binary provider.
     *
     * @param request The binary request containing the restore candidate sha1
     */
    void initiateRestoreFromColdStorage(BinaryElementRequest request);

    List<PersistedRemoteTask> getNextRemoteTaskFromDataDb(long batchSize, int maxExecutionsCount,
            long retryDelayMs,
            long maximumExecTimeMs) throws SQLException;

    @Transactional
    List<PersistedRemoteTask> getRemoteTasksBatch(long batchSize, int maxExecutionsCount, long retryDelayMs,
            long maximumExecTimeMs) throws SQLException;

    /**
     *
     * @param persistedRemoteTask remote task to queue in the database
     * @throws SQLException
     * the method enclosed in transaction in order to make this task visible to its consumers as soon as possible.
     * We are offering the task as part of the add artificial binary flow and we don't need the nodes table data
     * to complete the task thus the task can be fired immediately
     */
    @Transactional(propagation = Propagation.REQUIRES_NEW)
    void offerRemoteTask(PersistedRemoteTask persistedRemoteTask) throws SQLException;

    List<PersistedRemoteTask> getTasksBySha1(String sha1) throws SQLException;
}
