package org.artifactory.storage.db.repo.service;

import org.artifactory.api.context.ContextHelper;
import org.artifactory.storage.StorageException;
import org.artifactory.storage.db.fs.service.RepositoryPersistingService;
import org.artifactory.storage.db.repo.dao.RepositoryConfigDao;
import org.artifactory.storage.db.repo.entity.DbRepositoryConfig;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.support.TransactionSynchronizationManager;
import org.springframework.util.CollectionUtils;

import java.sql.SQLException;
import java.util.Collection;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.function.Function;
import java.util.stream.Collectors;

/**
 * @author Tamar Rosen
 */
@Service
public class RepositoryConfigStorageServiceImpl implements RepositoryConfigStorageService {
    private static final Logger log = LoggerFactory.getLogger(RepositoryConfigStorageServiceImpl.class);
    private final RepositoryConfigDao repositoryConfigDao;
    private final RepositoryPersistingService repositoryPersistingService;

    @Autowired
    public RepositoryConfigStorageServiceImpl(RepositoryConfigDao repositoryConfigDao,
            RepositoryPersistingService repositoryPersistingService) {
        this.repositoryConfigDao = repositoryConfigDao;
        this.repositoryPersistingService = repositoryPersistingService;
    }

    @Override
    public DbRepositoryConfig findRepositoryByKey(String repoKey) {
        try {
            Optional<DbRepositoryConfig> dbRepositoryConfigOptional = repositoryConfigDao.findRepositoryByKey(repoKey);
            if (dbRepositoryConfigOptional.isPresent()) {
                return dbRepositoryConfigOptional.get();
            }
        } catch (SQLException e) {
            throw new StorageException("Unable to find repository : " + repoKey);
        }
        return null;
    }

    @Override
    public List<DbRepositoryConfig> getAllRepositories() {
        try {
            return repositoryConfigDao.getAll();
        } catch (SQLException e) {
            throw new StorageException("Unable to fetch all repositories", e);
        }
    }

    @Override
    public Set<DbRepositoryConfig> findRepositoryConfigsByKeys(Set<String> repoKeys) {
        try {
            return repositoryConfigDao.findByKeysWithPagination(repoKeys);
        } catch (SQLException e) {
            throw new StorageException("Unable to fetch repositories", e);
        }
    }

    @Override
    public void insertRepositoryConfig(DbRepositoryConfig dbRepositoryConfig) {
        try {
            log.info("About to insert new repository configuration '{}'", dbRepositoryConfig.getKey());
            populateUpdatedDate(dbRepositoryConfig);
            repositoryConfigDao.insert(dbRepositoryConfig);
            repositoryPersistingService.markDirty();
        } catch (SQLException e) {
            if (findRepositoryByKey(dbRepositoryConfig.getKey()) != null) {
                log.warn("Repository with key '{}' already exists.", dbRepositoryConfig.getKey());
                return;
            }
            throw new StorageException(
                    "Unable to insert new configuration for repository : " + dbRepositoryConfig.getKey());
        }
    }

    @Override
    public void bulkInsertRepositoryConfigs(Set<DbRepositoryConfig> dbRepositoryConfigs) {
        try {
            if (!CollectionUtils.isEmpty(dbRepositoryConfigs)) {
                dbRepositoryConfigs.stream().forEach(RepositoryConfigStorageServiceImpl::populateUpdatedDate);
                log.info("About to insert {} new repository configurations", dbRepositoryConfigs.size());
                repositoryConfigDao.bulkInsert(dbRepositoryConfigs.stream()
                        .sorted(Comparator.comparing(DbRepositoryConfig::getKey))
                        .collect(Collectors.toList()));
                repositoryPersistingService.markDirty();
            }
        } catch (SQLException e) {
            throw new StorageException("Failed to insert new repository configurations", e);
        }
    }

    @Override
    public void bulkDeleteRepositoryConfigs(Set<String> repoKeys) {
        try {
            if (!CollectionUtils.isEmpty(repoKeys)) {
                log.debug("About to delete configuration blob for repository: '{}'", repoKeys.size());
                repositoryConfigDao.bulkDelete(repoKeys.stream().sorted(Comparator.comparing(Function.identity()))
                        .collect(Collectors.toList()));
                repositoryPersistingService.markDirty();
            }
        } catch (SQLException e) {
            throw new StorageException("Unable to delete configuration for repository: ", e);
        }
    }

    @Override
    public void updateRepositoryConfig(DbRepositoryConfig dbRepositoryConfig) {
        try {
            populateUpdatedDate(dbRepositoryConfig);
            log.debug("About to update configuration for repository: '{}'", dbRepositoryConfig.getKey());
            repositoryConfigDao.update(dbRepositoryConfig);
            repositoryPersistingService.markDirty();
        } catch (SQLException e) {
            throw new StorageException(
                    "Unable to update configuration for repository: " + dbRepositoryConfig.getKey(), e);
        }
    }

    @Override
    public void bulkUpdateRepositoryConfigs(Set<DbRepositoryConfig> dbRepositoryConfigs) {
        try {
            if (!CollectionUtils.isEmpty(dbRepositoryConfigs)) {
                dbRepositoryConfigs.stream().forEach(RepositoryConfigStorageServiceImpl::populateUpdatedDate);
                log.info("About to update {} repository configurations", dbRepositoryConfigs.size());
                repositoryConfigDao.bulkUpdate(
                        dbRepositoryConfigs.stream().sorted(Comparator.comparing(DbRepositoryConfig::getKey))
                                .collect(Collectors.toList()));
                repositoryPersistingService.markDirty();
            }
        } catch (SQLException e) {
            throw new StorageException("Failed to update repository configurations", e);
        }
    }

    @Override
    public void delete(String repoKey) {
        try {
            if (repositoryConfigDao.delete(repoKey) > 0) {
                log.debug("Repository '{}' was deleted.", repoKey);
                repositoryPersistingService.markDirty();
            }
        } catch (SQLException e) {
            throw new StorageException("Unable to delete repository " + repoKey, e);
        }
    }

    @Override
    public boolean exists(String repoKey) {
        try {
            return repositoryConfigDao.exists(repoKey);
        } catch (SQLException e) {
            throw new StorageException("Unable to validate repository's " + repoKey + "existence", e);
        }
    }

    @Override
    public int countAll() {
        try {
            return repositoryConfigDao.countAll();
        } catch (SQLException e) {
            throw new StorageException("Unable to count all repositories", e);
        }
    }

    @Override
    public int deleteAll() {
        try {
            int numDeleted = repositoryConfigDao.deleteAll();
            repositoryPersistingService.markDirty();
            return numDeleted;
        } catch (SQLException e) {
            throw new StorageException("Unable to delete all repositories", e);
        }
    }

    @Override
    public void flushBulkChanges(Set<DbRepositoryConfig> repositoriesToCreate,
            Set<DbRepositoryConfig> repositoriesToUpdate, Set<String> repositoriesToDelete) {
        if (TransactionSynchronizationManager.isSynchronizationActive()) {
            throw new StorageException("Active transaction is not allowed here");
        }
        RepositoryConfigStorageService repositoryConfigStorageService = getAdvisedMe();
        repositoryConfigStorageService.bulkInsertRepositoryConfigs(repositoriesToCreate);
        repositoryConfigStorageService.bulkUpdateRepositoryConfigs(repositoriesToUpdate);
        repositoryConfigStorageService.bulkDeleteRepositoryConfigs(repositoriesToDelete);
        repositoryPersistingService.markDirty();
    }

    @Override
    public Collection<DbRepositoryConfig> findRepositoriesByDetailsType(String repoDetailsType) {
        try {
            return repositoryConfigDao.findRepositoriesByDetailsType(repoDetailsType);
        } catch (SQLException e) {
            throw new StorageException("Unable to fetch repositories", e);
        }
    }

    private RepositoryConfigStorageService getAdvisedMe() {
        return ContextHelper.get().beanForType(RepositoryConfigStorageService.class);
    }

    @Override
    public Map<String, Long> getAllRepositoryKeysAndModificationDates() {
        try {
            return repositoryConfigDao.getAllRepositoryKeysAndModificationDates();
        } catch (SQLException e) {
            throw new StorageException("Unable to get all repositories keys", e);
        }
    }

    public static void populateUpdatedDate(DbRepositoryConfig dbRepositoryConfig) {
        dbRepositoryConfig.setModified(System.currentTimeMillis());
    }
}
