package org.artifactory.storage.db.version.converter;

import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang.StringUtils;
import org.jfrog.converter.AbstractDbConverter;
import org.jfrog.converter.DbConverterExecutionException;
import org.jfrog.converter.model.ChangeLogContext;
import org.jfrog.storage.DbType;
import org.jfrog.storage.util.DbUtils;
import org.jfrog.storage.util.PostgresqlUtils;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.List;
import java.util.function.BiPredicate;

/**
 * @author Hussein Alshqerat
 */
@Slf4j
public class SafeIndexConcurrentCreator extends AbstractDbConverter {
    private final String indexName;
    private final String tableName;
    private final List<String> columnNames;
    private final String customQuery;

    @SafeVarargs
    public SafeIndexConcurrentCreator(String indexName, String tableName, List<String> columnNames,
            String customQuery, BiPredicate<Connection, DbType>... conditions) {
        super("Creating index: " + indexName + " concurrently.", conditions);
        this.indexName = indexName;
        this.tableName = tableName;
        this.columnNames = columnNames;
        this.customQuery = customQuery;
    }

    @SafeVarargs
    public SafeIndexConcurrentCreator(String indexName, String tableName, List<String> columnsNames,
            BiPredicate<Connection, DbType>... conditions) {
        super("Creating index: " + indexName + " concurrently.", conditions);
        this.indexName = indexName;
        this.tableName = tableName;
        this.columnNames = columnsNames;
        this.customQuery = "";
    }

    boolean indexExists(Connection conn, DbType dbType) throws SQLException {
        for (String columnName : columnNames) {
            if (!DbUtils.indexExists(conn, dbType, tableName, columnName, indexName, false)) {
                return false;
            }
        }
        return true;
    }

    boolean isIndexValid(Connection conn) throws SQLException {
        return PostgresqlUtils.isIndexValid(conn, indexName);
    }

    void createIndexConcurrently(Connection conn) throws SQLException {
        log.info("Index not exists, creating index {} concurrently.", indexName);
        if (StringUtils.isBlank(customQuery)) {
            PostgresqlUtils.createIndexConcurrently(conn, tableName, indexName, columnNames);
        } else {
            PostgresqlUtils.createCustomIndexConcurrently(conn, customQuery);
        }
    }

    boolean checkIndexExistsAndValidBeforeCreation(Connection conn, DbType dbType) throws SQLException {
        if (indexExists(conn, dbType)) {
            log.info("Index {} already exists - Checking if is valid.", indexName);
            if (isIndexValid(conn)) {
                log.info("Index {} is valid - Skipping upgrade.", indexName);
                return true;
            } else {
                log.warn("Index {} invalid - Dropping index.", indexName);
                DbUtils.dropIndex(conn, indexName);
            }
        }
        return false;
    }

    void assertIndexExistsAndValidAfterCreation(Connection conn, DbType dbType)
            throws SQLException, DbConverterExecutionException {
        String msg = "Could not convert DB: CREATE INDEX CONCURRENTLY " + indexName;
        if (indexExists(conn, dbType)) {
            if (!isIndexValid(conn)) {
                log.warn("Index {} invalid after concurrent creation - Dropping index.", indexName);
                DbUtils.dropIndex(conn, indexName);

                log.error("{} due to {}.", msg, "Index invalid after creation.");
                throw new DbConverterExecutionException(msg);
            }
        } else {
            log.error("{} due to {}", msg, "Index not exists after creation.");
            throw new DbConverterExecutionException(msg);
        }
        log.info("Index {} created successfully.", indexName);
    }

    @Override
    protected void executeConverter(ChangeLogContext context) throws DbConverterExecutionException {
        try {
            Connection conn = context.getConnection();
            DbType dbType = context.getDbType();
            if (checkIndexExistsAndValidBeforeCreation(conn, dbType)) {
                return;
            }
            createIndexConcurrently(conn);
            assertIndexExistsAndValidAfterCreation(conn, dbType);
        } catch (SQLException e) {
            String msg = "Could not convert DB: CREATE INDEX CONCURRENTLY " + indexName;
            log.error("{} due to {}", msg, e.getMessage(), e);
            throw new DbConverterExecutionException(msg, e);
        }
    }
}