package org.artifactory.storage.db.binstore.service;

import com.google.common.collect.Lists;
import com.google.common.collect.MapMaker;
import org.apache.commons.lang3.StringUtils;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.ha.HaCommonAddon;
import org.artifactory.aol.AolService;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.concurrent.NonStandardLockCategory;
import org.artifactory.entitlement.EntitlementType;
import org.artifactory.repo.trash.TrashService;
import org.artifactory.storage.StorageException;
import org.artifactory.storage.binstore.service.BinariesGarbageCollectorService;
import org.artifactory.storage.binstore.service.GarbageCollectListenerResultStatus;
import org.artifactory.storage.binstore.service.GarbageCollectionDatastoreService;
import org.artifactory.storage.binstore.service.GarbageCollectorInfo;
import org.artifactory.storage.binstore.service.GarbageCollectorStrategy;
import org.artifactory.storage.db.InternalDbService;
import org.artifactory.storage.db.binstore.exceptions.RemoteTasksStatusException;
import org.artifactory.storage.db.servers.model.ArtifactoryServer;
import org.artifactory.storage.task.BinariesTasksInfo;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.jfrog.common.BinaryStoreConfigService;
import org.jfrog.common.concurrency.ConflictGuard;
import org.jfrog.common.concurrency.JvmConflictsGuard;
import org.jfrog.security.util.Pair;
import org.jfrog.storage.DbType;
import org.jfrog.storage.binstore.BinariesInfo;
import org.jfrog.storage.binstore.common.BinaryElementImpl;
import org.jfrog.storage.binstore.config.model.ChainMetaData;
import org.jfrog.storage.binstore.config.model.Param;
import org.jfrog.storage.binstore.config.model.ProviderMetaData;
import org.jfrog.storage.binstore.db.dao.RemoteBinaryTaskDao;
import org.jfrog.storage.binstore.exceptions.BinaryDbException;
import org.jfrog.storage.binstore.exceptions.BinaryStorageException;
import org.jfrog.storage.binstore.exceptions.NoCloudProviderConfiguredException;
import org.jfrog.storage.binstore.exceptions.SignedUrlException;
import org.jfrog.storage.binstore.ifc.BinaryProviderInfo;
import org.jfrog.storage.binstore.ifc.BinaryStoreFederatedMember;
import org.jfrog.storage.binstore.ifc.model.BinaryElement;
import org.jfrog.storage.binstore.ifc.model.BinaryElementRequest;
import org.jfrog.storage.binstore.ifc.model.BinaryProvidersInfo;
import org.jfrog.storage.binstore.ifc.model.BinaryProvidersResponseMap;
import org.jfrog.storage.binstore.ifc.model.BinaryTreeElement;
import org.jfrog.storage.binstore.ifc.provider.BinaryProvider;
import org.jfrog.storage.binstore.manager.BinaryProviderManager;
import org.jfrog.storage.binstore.providers.base.CloudBinaryProvider;
import org.jfrog.storage.binstore.providers.base.ManualRedirectType;
import org.jfrog.storage.binstore.providers.base.MutableBinaryProvider;
import org.jfrog.storage.binstore.providers.base.PersistedRemoteTask;
import org.jfrog.storage.binstore.providers.base.RedirectType;
import org.jfrog.storage.binstore.providers.base.StorageInfoImpl;
import org.jfrog.storage.binstore.service.BinariesDbService;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.springframework.test.util.ReflectionTestUtils;
import org.testng.Assert;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;
import org.testng.collections.Maps;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

import static org.jfrog.storage.binstore.ifc.model.BinaryElementHeaders.REPOSITORY_KEY;
import static org.mockito.ArgumentMatchers.anyInt;
import static org.mockito.ArgumentMatchers.anyLong;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.Mockito.any;
import static org.mockito.Mockito.doAnswer;
import static org.mockito.Mockito.doNothing;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.eq;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertFalse;
import static org.testng.Assert.assertTrue;

/**
 * @author Shay Bagants
 * Author: Ivan Perepelytsya
 * Date: 6/3/20.
 */
@Test
public class BinaryServiceImplTest extends ArtifactoryHomeBoundTest {

    @InjectMocks
    private BinaryServiceImpl binaryService;

    @Mock
    private BinaryProviderManager binaryProviderManager;
    @Mock
    private CloudBinaryProvider cloudBinaryProvider;
    @Mock
    private AddonsManager addonsManager;
    @Mock
    private HaCommonAddon haCommonAddon;
    @Mock
    private AolService aolService;
    @Mock
    private MutableBinaryProvider mutableBinaryProvider;
    @Mock
    private ProviderMetaData providerMetaData;
    @Mock
    private ArtifactoryContext artifactoryContext;
    @Mock
    private BinariesGarbageCollectorService gcService;
    @Mock
    private TrashService trashService;
    @Mock
    private BinariesDbService binariesDbService;
    @Mock
    private InternalDbService dbService;
    @Mock
    private GarbageCollectionDatastoreService garbageCollectionDatastoreService;
    @Mock
    private DeleteBlobInfosWrapper deleteBlobInfosWrapper;
    @Mock
    private BinaryStoreConfigService binaryStoreConfigService;
    private int originalBatchSize;

    @BeforeMethod
    private void setup() {
        MockitoAnnotations.openMocks(this);
        when(artifactoryContext.beanForType(BinariesGarbageCollectorService.class)).thenReturn(gcService);
        when(artifactoryContext.beanForType(TrashService.class)).thenReturn(trashService);
        ArtifactoryContextThreadBinder.bind(artifactoryContext);
        ReflectionTestUtils.setField(binaryService, "binariesDbService", binariesDbService);
        ReflectionTestUtils
                .setField(binaryService, "garbageCollectionDatastoreService", garbageCollectionDatastoreService);
        binaryService.setDbService(dbService);
        binaryService.setDeleteBlobInfosWrapper(deleteBlobInfosWrapper);
        originalBatchSize = ConstantValues.trashcanMaxSearchResults.getInt();
        ArtifactoryHome.get().getArtifactoryProperties().setProperty(
                ConstantValues.trashcanMaxSearchResults.getPropertyName(), "3");
        when(artifactoryContext.beanForType(AddonsManager.class)).thenReturn(addonsManager);
        when(cloudBinaryProvider.getProviderMetaData()).thenReturn(providerMetaData);
        when(cloudBinaryProvider.getMutableBinaryProvider()).thenReturn(mutableBinaryProvider);
        when(binaryProviderManager.getCloudProviders()).thenReturn(List.of(cloudBinaryProvider));
        when(artifactoryContext.beanForType(AolService.class)).thenReturn(aolService);
        when(addonsManager.addonByType(HaCommonAddon.class)).thenReturn(haCommonAddon);
        when(haCommonAddon.getConflictsGuard(NonStandardLockCategory.FORCE_BINARY_SHARDING_OPTIMIZATION.getName()))
                .thenReturn(new JvmConflictsGuard<>(1));

    }

    @AfterMethod
    private void cleanup() {
        ArtifactoryContextThreadBinder.unbind();
        ArtifactoryHome.get().getArtifactoryProperties().setProperty(
                ConstantValues.trashcanMaxSearchResults.getPropertyName(), String.valueOf(originalBatchSize));
    }

    public void testStartGarbageCollect() throws BinaryDbException {
        String expected = "EXPECTED";
        when(binariesDbService.getCountAndTotalSize()).thenThrow(new BinaryDbException(expected));

        when(trashService.isTrashcanEnabled()).thenReturn(true);

        AtomicInteger fullGcErrorCount = new AtomicInteger(0);
        for (int i = 0; i < 100; i++) {
            try {
                binaryService.startGarbageCollect();
            } catch (StorageException e) {
                if (StringUtils.endsWith(e.getMessage(), expected)) {
                    fullGcErrorCount.incrementAndGet();
                }
            }
        }
        verify(gcService, times(100)).startGCByStrategy(any());
        // ensure full gc was triggered once
        assertEquals(5, fullGcErrorCount.get());
        AtomicInteger gcIteration = (AtomicInteger) ReflectionTestUtils.getField(binaryService, "gcIteration");
        Assert.assertNotNull(gcIteration);
        assertEquals(gcIteration.get(), 0);
        binaryService.startGarbageCollect();
        assertEquals(gcIteration.get(), 1);
    }

    @Test
    public void runFullGarbageCollectCompletedTest() throws InterruptedException {
        BinariesInfo binariesInfo = mock(BinariesInfo.class);
        when(binariesInfo.getBinariesCount()).thenReturn(10L);
        when(binariesInfo.getBinariesSize()).thenReturn(200L);
        try {
            when(binariesDbService.getCountAndTotalSize()).thenReturn(binariesInfo);
        } catch (BinaryDbException e) {
            throw new StorageException(e);
        }
        when(dbService.invokeInTransaction(eq("BinaryCleaner#sha1a1"), any())).thenReturn(true);
        Collection<BinaryElement> results = Lists.newArrayList();
        for (int i = 0; i < 10; i++) {
            BinaryElement binaryElement = new BinaryElementImpl("sha1a" + i, "sha2a" + i, "sha5a" + i, 5 + i);
            results.add(binaryElement);
        }
        doNothing().when(deleteBlobInfosWrapper).deleteBlobInfosAsync(any());
        try {
            when(binariesDbService.findPotentialDeletion()).thenReturn(results);
        } catch (BinaryDbException e) {
            throw new StorageException(e);
        }
        GarbageCollectorInfo garbageCollectorInfo = binaryService.runFullGarbageCollect();
        assertEquals(garbageCollectorInfo.status, GarbageCollectListenerResultStatus.COMPLETED);
        assertEquals(garbageCollectorInfo.failures, 9);
        assertEquals(garbageCollectorInfo.strategy, GarbageCollectorStrategy.FULL);
        when(garbageCollectionDatastoreService.getLastGCRunInfo()).thenReturn(garbageCollectorInfo);
        verify(binaryProviderManager).optimize(false);
        JvmConflictsGuard<Object> conflictsGuard = new JvmConflictsGuard<>(1);
        when(haCommonAddon.getConflictsGuard(NonStandardLockCategory.FORCE_BINARY_SHARDING_OPTIMIZATION.getName()))
                .thenReturn(
                        conflictsGuard);
        ConflictGuard lock = conflictsGuard.getLock("forceOptimizationOnce");
        lock.tryToLock(10, TimeUnit.SECONDS);
        binaryService.runFullGarbageCollect();
        assertEquals(garbageCollectionDatastoreService.getLastGCRunInfo(), garbageCollectorInfo);
        verify(binaryProviderManager).optimize(true);
    }

    @Test
    public void testTryToOptimizeWithLockNoError() {
        JvmConflictsGuard<Object> conflictsGuard = mock(JvmConflictsGuard.class);
        ConflictGuard lock = mock(ConflictGuard.class);
        when(lock.isLocked()).thenReturn(true);
        when(conflictsGuard.getLock(anyString())).thenReturn(lock);
        when(haCommonAddon.getConflictsGuard(NonStandardLockCategory.FORCE_BINARY_SHARDING_OPTIMIZATION.getName()))
                .thenReturn(
                        conflictsGuard);
        when(binaryProviderManager.optimize(lock.isLocked())).thenReturn(true);
        binaryService.tryToOptimize();
        verify(lock, times(1)).forceUnlock();
    }

    @Test
    public void testTryToOptimizeNoLockNoError() {
        JvmConflictsGuard<Object> conflictsGuard = mock(JvmConflictsGuard.class);
        ConflictGuard lock = mock(ConflictGuard.class);
        when(lock.isLocked()).thenReturn(false);
        when(conflictsGuard.getLock(anyString())).thenReturn(lock);
        when(haCommonAddon.getConflictsGuard(NonStandardLockCategory.FORCE_BINARY_SHARDING_OPTIMIZATION.getName()))
                .thenReturn(
                        conflictsGuard);
        when(binaryProviderManager.optimize(lock.isLocked())).thenReturn(true);
        binaryService.tryToOptimize();
        verify(lock, times(0)).forceUnlock();
    }

    @Test
    public void runFullGarbageCollectAbortedTest() {
        BinariesInfo binariesInfo = mock(BinariesInfo.class);
        when(binariesInfo.getBinariesCount()).thenReturn(10L);
        when(binariesInfo.getBinariesSize()).thenReturn(200L);
        try {
            when(binariesDbService.getCountAndTotalSize()).thenReturn(binariesInfo);
        } catch (BinaryDbException e) {
            throw new StorageException(e);
        }
        Collection<BinaryElement> results = Lists.newArrayList();
        when(dbService.invokeInTransaction(eq("BinaryCleaner#sha1a1"), any())).thenReturn(true);
        for (int i = 0; i < 20; i++) {
            BinaryElement binaryElement = new BinaryElementImpl("sha1a" + i, "sha2a" + i, "sha5a" + i, 5 + i);
            results.add(binaryElement);
        }
        doNothing().when(deleteBlobInfosWrapper).deleteBlobInfosAsync(any());
        try {
            when(binariesDbService.findPotentialDeletion()).thenReturn(results);
        } catch (BinaryDbException e) {
            throw new StorageException(e);
        }
        GarbageCollectorInfo garbageCollectorInfo = binaryService.runFullGarbageCollect();
        assertEquals(garbageCollectorInfo.status, GarbageCollectListenerResultStatus.UNCOMPLETED);
        assertEquals(garbageCollectorInfo.failures, 10);
        assertEquals(garbageCollectorInfo.strategy, GarbageCollectorStrategy.FULL);
    }

    @Test
    public void testCdnIsAllowedS3Storage() {
        when(aolService.isAol()).thenReturn(true);
        when(providerMetaData.getType()).thenReturn("s3-storage-v3");
        when(cloudBinaryProvider.supportsRedirection()).thenReturn(true);
        when(addonsManager.isEntitled(EntitlementType.CDN)).thenReturn(true);
        final boolean cdnAllowed = binaryService.isCdnAllowed();
        assertTrue(cdnAllowed);
    }

    @Test
    public void testCdnIsAllowedS3Sharding() {
        when(aolService.isAol()).thenReturn(true);
        when(providerMetaData.getType()).thenReturn("state-aware-s3");
        when(cloudBinaryProvider.supportsRedirection()).thenReturn(true);
        when(addonsManager.isEntitled(EntitlementType.CDN)).thenReturn(true);
        final boolean cdnAllowed = binaryService.isCdnAllowed();
        assertTrue(cdnAllowed);
    }

    @Test
    public void testCdnIsNotAllowedNotAol() {
        when(providerMetaData.getType()).thenReturn("s3-storage-v3");
        when(cloudBinaryProvider.supportsRedirection()).thenReturn(true);
        when(addonsManager.isEntitled(EntitlementType.CDN)).thenReturn(false);
        final boolean cdnAllowed = binaryService.isCdnAllowed();
        assertFalse(cdnAllowed);
    }

    @Test
    public void testCdnIsNotAllowedS3ShardingNotAol() {
        when(providerMetaData.getType()).thenReturn("state-aware-s3");
        when(cloudBinaryProvider.supportsRedirection()).thenReturn(true);
        when(addonsManager.isEntitled(EntitlementType.CDN)).thenReturn(false);
        final boolean cdnAllowed = binaryService.isCdnAllowed();
        assertFalse(cdnAllowed);
    }

    @Test
    public void testCdnIsNotAllowedNotS3() {
        when(providerMetaData.getType()).thenReturn("GCP");
        when(cloudBinaryProvider.supportsRedirection()).thenReturn(true);
        when(addonsManager.isEntitled(EntitlementType.CDN)).thenReturn(true);
        final boolean cdnAllowed = binaryService.isCdnAllowed();
        assertFalse(cdnAllowed);
    }

    @Test
    public void testCdnIsNotAllowedNoRedirects() {
        when(providerMetaData.getType()).thenReturn("s3-storage-v3");
        when(cloudBinaryProvider.supportsRedirection()).thenReturn(false);
        when(addonsManager.isEntitled(EntitlementType.CDN)).thenReturn(true);
        final boolean cdnAllowed = binaryService.isCdnAllowed();
        assertFalse(cdnAllowed);
    }

    @Test
    public void testCdnIsNotAllowedS3ShardingNoRedirects() {
        when(providerMetaData.getType()).thenReturn("state-aware-s3");
        when(cloudBinaryProvider.supportsRedirection()).thenReturn(false);
        when(addonsManager.isEntitled(EntitlementType.CDN)).thenReturn(true);
        final boolean cdnAllowed = binaryService.isCdnAllowed();
        assertFalse(cdnAllowed);
    }

    @Test
    public void testCdnIsNotAllowedMixedProviders() {
        when(binaryProviderManager.getCloudProviders()).thenReturn(List.of(cloudBinaryProvider, cloudBinaryProvider));
        when(providerMetaData.getType()).thenReturn("state-aware-s3").thenReturn("s3-storage-v3");
        when(cloudBinaryProvider.supportsRedirection()).thenReturn(true).thenReturn(true);
        when(addonsManager.isEntitled(EntitlementType.CDN)).thenReturn(true);
        final boolean cdnAllowed = binaryService.isCdnAllowed();
        assertFalse(cdnAllowed);
    }

    @Test
    public void testCdnIsAllowedS3ShardingProviders() {
        when(aolService.isAol()).thenReturn(true);
        when(binaryProviderManager.getCloudProviders()).thenReturn(List.of(cloudBinaryProvider, cloudBinaryProvider));
        when(providerMetaData.getType()).thenReturn("state-aware-s3").thenReturn("state-aware-s3");
        when(cloudBinaryProvider.supportsRedirection()).thenReturn(true).thenReturn(true);
        when(addonsManager.isEntitled(EntitlementType.CDN)).thenReturn(true);
        final boolean cdnAllowed = binaryService.isCdnAllowed();
        assertTrue(cdnAllowed);
    }

    @Test
    public void testCdnIsNotAllowedWrongLicense() {
        when(providerMetaData.getType()).thenReturn("s3-storage-v3");
        when(cloudBinaryProvider.supportsRedirection()).thenReturn(true);
        final boolean cdnAllowed = binaryService.isCdnAllowed();
        assertFalse(cdnAllowed);
    }

    @Test
    public void testCdnIsNotAllowedS3ShardingWrongLicense() {
        when(providerMetaData.getType()).thenReturn("state-aware-s3");
        when(cloudBinaryProvider.supportsRedirection()).thenReturn(true);
        final boolean cdnAllowed = binaryService.isCdnAllowed();
        assertFalse(cdnAllowed);
    }

    @Test
    public void testGetSignedUrl() {
        // setup
        BinaryElement binaryElement = new BinaryElementImpl("sha1", "sha2", "md5", 100);
        CloudBinaryProvider mockedProvider = mock(CloudBinaryProvider.class);
        when(binaryProviderManager.getCloudProviders()).thenReturn(List.of(mockedProvider));
        when(mockedProvider.getSignedUrl(binaryElement)).thenReturn("http://signed.url");

        // test
        String signedUrl = binaryService.getSignedUrl(binaryElement);
        assertTrue(binaryElement.getHeaders().containsKey("skipRemoteCheck"));

        // verify
        assertEquals(signedUrl, "http://signed.url");
    }

    @Test
    public void testGetSignedUrlMultipleProvidersFirstProvider() {
        // setup
        BinaryElement binaryElement = new BinaryElementImpl("sha1", "sha2", "md5", 100);
        CloudBinaryProvider mockedProvider = mock(CloudBinaryProvider.class);
        when(binaryProviderManager.getCloudProviders()).thenReturn(List.of(mockedProvider, mockedProvider));
        when(mockedProvider.getSignedUrl(binaryElement)).thenReturn("http://signed.url");

        // test
        String signedUrl = binaryService.getSignedUrl(binaryElement);
        assertFalse(binaryElement.getHeaders().containsKey("skipRemoteCheck"));
        verify(mockedProvider, times(1)).getSignedUrl(binaryElement);

        // verify
        assertEquals(signedUrl, "http://signed.url");
    }

    @Test
    public void testGetSignedUrlMultipleProvidersNotFirstProvider() {
        // setup
        BinaryElement binaryElement = new BinaryElementImpl("sha1", "sha2", "md5", 100);
        CloudBinaryProvider mockedProvider = mock(CloudBinaryProvider.class);
        when(binaryProviderManager.getCloudProviders()).thenReturn(List.of(mockedProvider, mockedProvider));
        when(mockedProvider.getSignedUrl(binaryElement)).thenReturn(null).thenReturn("http://signed.url");

        // test
        String signedUrl = binaryService.getSignedUrl(binaryElement);
        assertFalse(binaryElement.getHeaders().containsKey("skipRemoteCheck"));
        verify(mockedProvider, times(2)).getSignedUrl(binaryElement);

        // verify
        assertEquals(signedUrl, "http://signed.url");
    }

    @Test(expectedExceptions = SignedUrlException.class)
    public void testGetSignedUrl_NoCloudProvider() {
        // setup
        BinaryElement binaryElement = new BinaryElementImpl("sha1", "sha2", "md5", 100);
        when(binaryProviderManager.getCloudProviders()).thenReturn(List.of());
        // test
        binaryService.getSignedUrl(binaryElement);
    }

    @Test(expectedExceptions = SignedUrlException.class)
    public void testGetSignedUrl_EmptyUrl() {
        // setup
        BinaryElement binaryElement = new BinaryElementImpl("sha1", "sha2", "md5", 100);
        CloudBinaryProvider mockedProvider = mock(CloudBinaryProvider.class);
        when(binaryProviderManager.getCloudProviders()).thenReturn(List.of(mockedProvider));
        when(mockedProvider.getSignedUrl(binaryElement)).thenReturn(null);

        // test
        binaryService.getSignedUrl(binaryElement);
    }

    @Test
    public void testUpdateFederatedRepoBinaryProvider() {
        List<BinaryStoreFederatedMember> addMembers = new ArrayList<>();
        addMembers.add(new BinaryStoreFederatedMember("localKey", "remoteKey", "remoteUrl", null, "audience"));
        List<BinaryStoreFederatedMember> removedMembers = new ArrayList<>();
        removedMembers.add(
                new BinaryStoreFederatedMember("removedKey", "remoteRemovedKey", "remoteUrl", null, "audience"));

        binaryService.updateFederateRepoBinaryProvider(addMembers, removedMembers, null);
        verify(binaryProviderManager, times(1)).signalMirrorChanged(addMembers, removedMembers);
    }

    @Test
    public void testGetBinaryForChecksumDeploy() {
        Map<String, String> headers = Maps.newHashMap();
        headers.put(REPOSITORY_KEY.getHeaderName(), "repo-key");
        BinaryElement binaryElement = new BinaryElementImpl("sha1", "sha2", "md5", 1, headers);
        BinaryProvider binaryProvider = mock(BinaryProvider.class);
        BinaryProvidersResponseMap response = mock(BinaryProvidersResponseMap.class);
        when(response.exists()).thenReturn(true);
        doAnswer(invocation -> {
            final BinaryElementRequest argument = invocation.getArgument(0, BinaryElementRequest.class);
            assertEquals(argument.getHeaders(), headers);
            return response;
        }).when(binaryProvider).exists(any());

        when(binaryProviderManager.getFirstBinaryProvider()).thenReturn(binaryProvider);
        ConcurrentMap<String, Pair<AtomicInteger, Long>> deleteProtectedBinaries = new MapMaker().makeMap();
        ReflectionTestUtils.setField(binaryService, "deleteProtectedBinaries", deleteProtectedBinaries);

        binaryService.getBinaryForChecksumDeploy(binaryElement);
        verify(binaryProvider, times(1)).exists(any());
    }

    @Test
    public void testGetBinariesTasksSummaryByRepo() throws SQLException {
        long currentTime = System.currentTimeMillis();
        ChainMetaData chainMetaData = mock(ChainMetaData.class);
        when(binaryProviderManager.getChain()).thenReturn(chainMetaData);
        when(binaryProviderManager.getChain().getProviderMetaData()).thenReturn(providerMetaData);
        when(providerMetaData.getParam("maxRetry")).thenReturn(new Param("maxRetry", "10"));
        RemoteBinaryTaskDao remoteBinaryTaskDao = mock(RemoteBinaryTaskDao.class);
        when(remoteBinaryTaskDao.getInProgressTasksCountByRepo("repo1", currentTime, 10)).thenReturn(2L);
        when(remoteBinaryTaskDao.getFailingTasksCountByRepo("repo1", currentTime, 10)).thenReturn(5L);
        binaryService.setRemoteBinaryTaskDao(remoteBinaryTaskDao);
        BinariesTasksInfo binariesTasksInfo = binaryService.getBinariesTasksSummary("repo1", currentTime);
        assertEquals(binariesTasksInfo.getInProgressTasks().longValue(), 2L);
        assertEquals(binariesTasksInfo.getFailingTasks().longValue(), 5L);
    }

    @Test(expectedExceptions = RemoteTasksStatusException.class)
    public void testGetTasksSummaryFailsOnSqlException() throws SQLException {
        long currentTime = System.currentTimeMillis();
        ChainMetaData chainMetaData = mock(ChainMetaData.class);
        when(binaryProviderManager.getChain()).thenReturn(chainMetaData);
        when(binaryProviderManager.getChain().getProviderMetaData()).thenReturn(providerMetaData);
        when(providerMetaData.getParam("maxRetry")).thenReturn(new Param("maxRetry", "10"));
        RemoteBinaryTaskDao remoteBinaryTaskDao = mock(RemoteBinaryTaskDao.class);
        binaryService.setRemoteBinaryTaskDao(remoteBinaryTaskDao);
        when(remoteBinaryTaskDao.getInProgressTasksCountByRepo("repo2", currentTime, 10)).thenThrow(new SQLException());
        BinariesTasksInfo binariesTasksInfo = binaryService.getBinariesTasksSummary("repo2", currentTime);
    }

    @Test(expectedExceptions = NoCloudProviderConfiguredException.class)
    public void testCalcRedirectNoCloudProviders() {
        when(binaryProviderManager.getCloudProviders()).thenReturn(List.of());
        binaryService.calcRedirectType(RedirectType.S3, true, ManualRedirectType.NONE);
    }

    @Test(expectedExceptions = BinaryStorageException.class)
    public void testCalcRedirectDifferentProviders() {
        when(binaryProviderManager.getCloudProviders()).thenReturn(List.of(cloudBinaryProvider, cloudBinaryProvider));
        when(providerMetaData.getType()).thenReturn("state-aware-s3").thenReturn("s3-storage-v3");
        binaryService.calcRedirectType(RedirectType.S3, true, ManualRedirectType.NONE);
    }

    @Test(expectedExceptions = BinaryStorageException.class)
    public void testCanCloudProviderGenerateRedirectionDifferentProviders() {
        when(binaryProviderManager.getCloudProviders()).thenReturn(List.of(cloudBinaryProvider, cloudBinaryProvider));
        when(providerMetaData.getType()).thenReturn("state-aware-s3").thenReturn("s3-storage-v3");
        binaryService.canCloudProviderGenerateRedirection("sha1");
    }

    @Test
    public void testCanCloudProviderGenerateRedirectionWithoutSha1() {
        when(binaryProviderManager.getCloudProviders()).thenReturn(List.of(cloudBinaryProvider, cloudBinaryProvider));
        when(providerMetaData.getType()).thenReturn("state-aware-s3").thenReturn("s3-storage-v3");
        assertFalse(binaryService.canCloudProviderGenerateRedirection(null));
    }

    @Test
    public void testStorageSummaryInfoIgnoresSeenSubProviders() {
        when(binaryProviderManager.getBinaryProvidersInfo()).thenReturn(createInfoForProviders());
        assertEquals(binaryService.getStorageInfoSummary(),
                new StorageInfoImpl(2, 0, 2, 0, 30));
    }

    @Test(dataProvider = "remoteTaskConfiguration")
    public void testGetNextRemoteTaskFromDataDb(boolean isHa, boolean topologyChanged, int expectedResult)
            throws SQLException {
        when(binaryStoreConfigService.isHaConfigured()).thenReturn(isHa);
        when(dbService.getDatabaseType()).thenReturn(DbType.DERBY);
        PersistedRemoteTask persistedRemoteTask = mock(PersistedRemoteTask.class);
        if (topologyChanged) {
            binaryService.clusterTopologyChanged(List.of(mock(ArtifactoryServer.class)));
        }
        BinaryServiceImpl binaryServiceSpy = spy(binaryService);
        doReturn(List.of(persistedRemoteTask)).when(binaryServiceSpy)
                .fetchTasksWithLock(anyLong(), anyInt(), anyLong(), anyLong());
        assertEquals(binaryServiceSpy.getNextRemoteTaskFromDataDb(10L, 10, 10L, 10L).size(), expectedResult);
    }

    @DataProvider
    public Object[][] remoteTaskConfiguration() {
        return new Object[][] {
                {false, false, 1},
                {true, false, 0},
                {true, true, 1}
        };
    }

    private BinaryProvidersInfo<BinaryProviderInfo> createInfoForProviders() {
        BinaryTreeElement<BinaryProviderInfo> elem = new BinaryTreeElement<>();
        BinaryTreeElement<BinaryProviderInfo> elem2 = new BinaryTreeElement<>();
        BinaryProviderInfo info = new BinaryProviderInfo();
        BinaryProviderInfo info2 = new BinaryProviderInfo();
        BinaryProvidersInfo<BinaryProviderInfo> finalInfo = new BinaryProvidersInfo<>("", elem);
        info.setStorageInfo(new StorageInfoImpl(2, 0, 2, 0, 30));
        info2.setStorageInfo(new StorageInfoImpl(1, 0, 1, 0, 20));
        elem2.setData(info2);
        elem.setData(info);
        elem.setSubBinaryTreeElements(List.of(elem2, elem2));

        return finalInfo;
    }
}