package org.artifactory.storage.db.binstore.service.garbage;

import com.google.common.util.concurrent.ThreadFactoryBuilder;
import org.artifactory.api.security.SecurityService;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.storage.DeletedGCCandidate;
import org.artifactory.storage.GCCandidate;
import org.artifactory.storage.binstore.service.GCProvider;
import org.artifactory.storage.binstore.service.GCProviderWrapper;
import org.artifactory.storage.binstore.service.GarbageCollectListenerResultStatus;
import org.artifactory.storage.binstore.service.GarbageCollectionDatastoreService;
import org.artifactory.storage.binstore.service.GarbageCollectorInfo;
import org.artifactory.storage.binstore.service.GarbageCollectorStrategy;
import org.artifactory.storage.binstore.service.InternalBinaryService;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.jfrog.storage.DbType;
import org.mockito.ArgumentCaptor;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.mockito.invocation.InvocationOnMock;
import org.mockito.stubbing.Answer;
import org.mockito.stubbing.Stubber;
import org.testng.Assert;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Set;
import java.util.concurrent.Future;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static org.mockito.Mockito.any;
import static org.mockito.Mockito.doAnswer;
import static org.mockito.Mockito.doThrow;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

/**
 * @author dudim
 */
@Test
public class GarbageCollectorTest extends ArtifactoryHomeBoundTest {

    private GarbageCollector garbageCollector;

    private int originalBatchSize;
    @Mock
    private GarbageCollectionDatastoreService garbageCollectionDatastoreService;
    @Mock
    private SecurityService securityService;
    @Mock
    private InternalBinaryService binaryService;
    @Mock
    private GCProvider gcProvider;

    @BeforeMethod
    private void setup() {
        MockitoAnnotations.initMocks(this);
        when(gcProvider.getName()).thenReturn("myname");
        originalBatchSize = ConstantValues.trashcanMaxSearchResults.getInt();
        ArtifactoryHome.get().getArtifactoryProperties()
                .setProperty(ConstantValues.trashcanMaxSearchResults.getPropertyName(), "3");
        garbageCollector = new GarbageCollector(gcProvider, securityService, binaryService, DbType.MYSQL,
                GarbageCollectorStrategy.TRASH_AND_BINARIES, garbageCollectionDatastoreService);
    }

    @AfterMethod
    private void cleanup() {
        ArtifactoryHome.get().getArtifactoryProperties().setProperty(
                ConstantValues.trashcanMaxSearchResults.getPropertyName(), String.valueOf(originalBatchSize));
    }

    public void testRun() {
        testGarbageCollection();
    }

    public void testRunWithError() {
        doThrow(new RuntimeException("Error occurred")).when(gcProvider).getBatch();
        garbageCollector.run();

        verify(gcProvider, times(1)).getBatch();
        verify(binaryService, times(0)).deleteBlobInfos(any());
        verify(binaryService, times(0)).deleteUnusedArchiveNames();
        verify(binaryService, times(0)).deleteUnusedArchivePaths();
        verify(garbageCollectionDatastoreService, times(0)).saveLastGCRunInfo(any());
    }

    public void testGarbageCollection() {
        List<GCCandidate> firstBatch = Arrays.asList(
                new GCCandidate(null, null, "x", null, 0),
                new GCCandidate(null, null, "y", null, 0),
                new GCCandidate(null, null, "`", null, 0)
        );
        List<GCCandidate> secondBatch = Arrays.asList(
                new GCCandidate(null, null, "1", null, 0),
                new GCCandidate(null, null, "2", null, 0),
                new GCCandidate(null, null, "3", null, 0)
        );

        mockTwoBatches(firstBatch, secondBatch).when(gcProvider).getBatch();

        when(gcProvider.getAction()).thenReturn((candidate, result) -> true);
        garbageCollector.run();

        verify(gcProvider, times(3)).getBatch();
        List<GCCandidate> allBatches = Stream.concat(firstBatch.stream(), secondBatch.stream())
                .collect(Collectors.toList());
        Set<String> sha2s = allBatches.stream().map(GCCandidate::getSha2).collect(Collectors.toSet());
        ArgumentCaptor<List<String>> argument = ArgumentCaptor.forClass(List.class);
        verify(binaryService, times(2)).deleteBlobInfos(argument.capture());
        List<List<String>> capturedBlobInfoDeletionRequests = argument.getAllValues();
        Set<String> capturedShas = capturedBlobInfoDeletionRequests.stream().flatMap(List::stream)
                .collect(Collectors.toSet());
        Assert.assertEquals(sha2s, capturedShas);

        // should be called at the end of the GC
        verify(binaryService, times(1)).deleteUnusedArchiveNames();
        verify(binaryService, times(1)).deleteUnusedArchivePaths();
        verify(garbageCollectionDatastoreService, times(2)).saveLastGCRunInfo(any());
    }

    public void testGarbageCollectionSmallBatch() {
        List<GCCandidate> firstBatch = Arrays.asList(
                new GCCandidate(null, null, "file-1", null, 0),
                new GCCandidate(null, null, "file-2", null, 0));
        List<GCCandidate> secondBatch = Collections.singletonList(
                new GCCandidate(null, null, "x", null, 0));
        mockTwoBatches(firstBatch, secondBatch).when(gcProvider).getBatch();
        when(gcProvider.getAction()).thenReturn((candidate, result) -> true);

        garbageCollector.run();
        // ensure second batch was never called as the first one was not full
        verify(gcProvider, times(1)).getBatch();
    }

    public void testGarbageCollectionWithSingleError() {
        List<GCCandidate> firstBatch = Arrays.asList(
                new GCCandidate(null, null, "x", null, 0),
                new GCCandidate(null, null, "y", null, 0),
                new GCCandidate(null, null, "z", null, 0)
        );
        List<GCCandidate> secondBatch = Arrays.asList(
                new GCCandidate(null, null, "1", null, 0),
                new GCCandidate(null, null, "2", null, 0),
                new GCCandidate(null, null, "3", null, 0)

        );
        mockTwoBatches(firstBatch, secondBatch).when(gcProvider).getBatch();

        AtomicInteger counter = new AtomicInteger(0);
        when(gcProvider.getAction()).thenReturn((candidate, result) -> {
            counter.getAndIncrement();
            return !candidate.getSha2().equals("x");
        });
        garbageCollector.run();

        verify(gcProvider, times(3)).getBatch();
        List<GCCandidate> allBatches = Stream.concat(firstBatch.stream(), secondBatch.stream())
                .collect(Collectors.toList());
        Set<String> sha2s = allBatches.stream().map(GCCandidate::getSha2)
                // filter out the first event which failed by the action
                .filter(sha -> !sha.equals(firstBatch.get(0).getSha2()))
                .collect(Collectors.toSet());
        ArgumentCaptor<List<String>> argument = ArgumentCaptor.forClass(List.class);
        verify(binaryService, times(2)).deleteBlobInfos(argument.capture());
        List<List<String>> capturedBlobInfoDeletionRequests = argument.getAllValues();
        Set<String> capturedShas = capturedBlobInfoDeletionRequests.stream().flatMap(List::stream)
                .collect(Collectors.toSet());
        Assert.assertEquals(sha2s, capturedShas);

        // should be called at the end of the GC
        verify(binaryService, times(1)).deleteUnusedArchiveNames();
        verify(binaryService, times(1)).deleteUnusedArchivePaths();
    }

    public void testExecuteTask() {
        GCCandidate candidate = new GCCandidate(null, null, "x", null, 0);
        GarbageCollectorInfo gcInfo = new GarbageCollectorInfo(GarbageCollectorStrategy.TRASH_AND_BINARIES);
        when(gcProvider.getAction()).thenReturn((gcCandidate, rs) -> true);

        DeletedGCCandidate deletedGCCandidate = garbageCollector.executeTask(gcProvider, candidate, gcInfo);
        Assert.assertEquals(deletedGCCandidate.getDeletedCandidate().getSha2(), candidate.getSha2());

        verify(securityService, times(1)).authenticateAsSystem();
        verify(gcProvider, times(1)).getAction();
    }

    public void testExecuteTaskWithError() {
        GCCandidate candidate = new GCCandidate(null, null, "x", null, 0);
        GarbageCollectorInfo gcInfo = new GarbageCollectorInfo(GarbageCollectorStrategy.TRASH_AND_BINARIES);

        doThrow(new RuntimeException("message")).when(gcProvider).getAction();
        DeletedGCCandidate deletedGCCandidate = garbageCollector.executeTask(gcProvider, candidate, gcInfo);
        Assert.assertNull(deletedGCCandidate);
        verify(securityService, times(1)).authenticateAsSystem();
        verify(gcProvider, times(1)).getAction();
    }


    private Stubber mockTwoBatches(List<GCCandidate> firstBatchCandidates, List<GCCandidate> secondBatchCandidates) {
        return doAnswer(new Answer() {
            private int count = 0;

            public Object answer(InvocationOnMock invocation) {
                count++;
                if (count == 1) {
                    return firstBatchCandidates;
                }
                if (count == 2) {
                    return secondBatchCandidates;
                }
                return Collections.emptyList();
            }
        });
    }

    public void testDeleteBlobInfos() {
        List<DeletedGCCandidate> deletedBinaries = Arrays.asList(
                new DeletedGCCandidate(new GCCandidate(null, "1", "x", null, 0)),
                new DeletedGCCandidate(new GCCandidate(null, "2", "", null, 0)),
                new DeletedGCCandidate(new GCCandidate(null, "3", null, null, 0))
        );
        garbageCollector.deleteBlobInfos(deletedBinaries);
        ArgumentCaptor<List<String>> argument = ArgumentCaptor.forClass(List.class);
        verify(binaryService, times(1)).deleteBlobInfos(argument.capture());
        List<String> capturedSha2s = argument.getValue();
        Assert.assertEquals(capturedSha2s.size(), 1);
        Assert.assertEquals(capturedSha2s.get(0), deletedBinaries.get(0).getDeletedCandidate().getSha2());
    }

    public void testWaitForWorkers() throws InterruptedException {
        List<Future<DeletedGCCandidate>> futures = new ArrayList<>();
        GCProvider gcProviderMockFirst = mock(GCProvider.class);
        when(gcProviderMockFirst.getAction()).thenReturn((candidate, result) -> true);
        GCCandidate gcCandidate = new GCCandidate(null, null, "t", null, 0);
        GarbageCollectorInfo info = new GarbageCollectorInfo(GarbageCollectorStrategy.TRASH_AND_BINARIES);
        ThreadPoolExecutor executor = getExecutorService(garbageCollector);
        try {
            futures.add(executor.submit(() -> garbageCollector.executeTask(gcProviderMockFirst, gcCandidate, info)));

            GCProvider gcProviderMockSecond = mock(GCProvider.class);
            when(gcProviderMockSecond.getAction()).thenReturn((candidate, result) -> false);
            GCCandidate gcCandidateTwo = new GCCandidate(null, null, "y", null, 0);
            futures.add(
                    executor.submit(() -> garbageCollector.executeTask(gcProviderMockSecond, gcCandidateTwo, info)));

            GCProvider gcProviderMockThird = mock(GCProvider.class);

            GCProviderWrapper gcProviderWrapper = new GCProviderWrapper(gcProviderMockThird);
            futures.add(executor.submit(() -> garbageCollector.executeTask(gcProviderWrapper, gcCandidateTwo, info)));
            info.candidatesForDeletion = 3;
            List<DeletedGCCandidate> resultsDeletedGcCandidates = garbageCollector
                    .waitForWorkers(gcProvider, futures,
                            info);
            Assert.assertEquals(resultsDeletedGcCandidates.size(), 1);
            Assert.assertEquals(info.failures, 1);
            Assert.assertEquals(info.status, GarbageCollectListenerResultStatus.COMPLETED);
            Assert.assertEquals(resultsDeletedGcCandidates.get(0).getDeletedCandidate().getSha2(), "t");
        } finally {
            executor.shutdownNow();
            executor.awaitTermination(10000, TimeUnit.MILLISECONDS);
        }
    }

    public void testMssqlThreads() {
        garbageCollector = new GarbageCollector(gcProvider, securityService, binaryService, DbType.MSSQL,
                GarbageCollectorStrategy.TRASH_AND_BINARIES, garbageCollectionDatastoreService);
        Assert.assertEquals(garbageCollector.numberOfThreads, 1);

        Arrays.stream(DbType.values())
                .filter(dbType -> dbType != DbType.MSSQL)
                .forEach(dbType -> {
                    garbageCollector = new GarbageCollector(gcProvider, securityService, binaryService, DbType.MYSQL,
                            GarbageCollectorStrategy.TRASH_AND_BINARIES, garbageCollectionDatastoreService);
                    Assert.assertEquals(garbageCollector.numberOfThreads,
                            ConstantValues.gcNumberOfWorkersThreads.getInt());
                });
    }

    public void testWaitForWorkersFailed() throws InterruptedException {
        List<Future<DeletedGCCandidate>> futures = new ArrayList<>();
        GCProvider gcProviderMockFirst = mock(GCProvider.class);
        when(gcProviderMockFirst.getAction()).thenReturn((candidate, result) -> true);
        GCCandidate gcCandidate = new GCCandidate(null, null, "t", null, 0);
        GarbageCollectorInfo info = new GarbageCollectorInfo(GarbageCollectorStrategy.TRASH_AND_BINARIES);
        garbageCollector.numberOfThreads = 20;
        ThreadPoolExecutor executor = getExecutorService(garbageCollector);
        try {
            futures.add(executor.submit(() -> garbageCollector.executeTask(gcProviderMockFirst, gcCandidate, info)));
            GCProvider gcProviderMockSecond = mock(GCProvider.class);
            when(gcProviderMockSecond.getAction()).thenReturn((candidate, result) -> false);
            GCCandidate gcCandidateTwo = new GCCandidate(null, null, "y", null, 0);
            futures.add(
                    executor.submit(() -> garbageCollector.executeTask(gcProviderMockSecond, gcCandidateTwo, info)));

            GCProvider gcProviderMockThird = mock(GCProvider.class);

            GCProviderWrapper gcProviderWrapper = new GCProviderWrapper(gcProviderMockThird);
            for (int i = 0; i < 10; i++) {
                futures.add(
                        executor.submit(() -> garbageCollector.executeTask(gcProviderWrapper, gcCandidateTwo, info)));
            }
            info.candidatesForDeletion = 12;
            List<DeletedGCCandidate> resultsDeletedGcCandidates = garbageCollector
                    .waitForWorkers(gcProvider, futures,
                            info);
            Assert.assertEquals(resultsDeletedGcCandidates.size(), 1);
            Assert.assertEquals(info.failures, 10);
            Assert.assertEquals(info.status, GarbageCollectListenerResultStatus.UNCOMPLETED);
            Assert.assertEquals(resultsDeletedGcCandidates.get(0).getDeletedCandidate().getSha2(), "t");
        } finally {
            executor.shutdownNow();
            executor.awaitTermination(10000, TimeUnit.MILLISECONDS);
        }


    }

    private ThreadPoolExecutor getExecutorService(GarbageCollector garbageCollector) {
        return new ThreadPoolExecutor(garbageCollector.numberOfThreads, garbageCollector.numberOfThreads, 5L,
                TimeUnit.MINUTES,
                new LinkedBlockingQueue<>(ConstantValues.trashcanMaxSearchResults.getInt() * 2),
                new ThreadFactoryBuilder().setDaemon(true)
                        .setNameFormat("art-GC-worker-%s")
                        .build());
    }
}