package org.artifactory.storage.db.version.converter;

import org.artifactory.storage.db.itest.DbBaseTest;
import org.jfrog.converter.DbConverterExecutionException;
import org.jfrog.converter.DbVersion;
import org.jfrog.converter.model.ChangeLogContext;
import org.mockito.Mock;
import org.testng.SkipException;
import org.testng.annotations.AfterClass;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.Test;

import java.sql.SQLException;
import java.util.List;

import static org.jfrog.converter.util.PreDefinedConditions.dbSpecific;
import static org.jfrog.storage.DbType.POSTGRESQL;
import static org.mockito.MockitoAnnotations.openMocks;
import static org.testng.Assert.assertEquals;

/**
 * @author Hussein Alshqerat
 */

public class SafeIndexConcurrentCreatorDbTest extends DbBaseTest {
    private ChangeLogContext context;
    private SafeIndexConcurrentCreator safeIndexConcurrentlyCreator;
    private AutoCloseable autoCloseable;
    @Mock
    private DbVersion dbVersion;
    private static final String IS_INVALID_INDEX_QUERY = "SELECT count(1) FROM pg_class, pg_index WHERE " +
            "pg_index.indisvalid = false AND pg_index.indexrelid = pg_class.oid and relname = 'simple_index'";

    private static final String MAKE_INDEX_INVALID_QUERY = "update pg_index set indisvalid = false where " +
            "indexrelid = (select oid from pg_class where relname = 'simple_index')";
    private static final String CREATE_INDEX_QUERY = "CREATE INDEX simple_index ON nodes(node_type)";
    private static final String DROP_INDEX_QUERY = "DROP INDEX IF EXISTS simple_index";


    @BeforeClass
    void setup() throws SQLException {
        if (dbProperties.isPostgres()) {
            autoCloseable = openMocks(this);
            context = new ChangeLogContext(jdbcHelper.getDataSource().getConnection(),
                    POSTGRESQL, dbVersion);
            safeIndexConcurrentlyCreator = new SafeIndexConcurrentCreator("simple_index", "nodes",
                    List.of("node_type"), dbSpecific(POSTGRESQL));
        }
    }

    @AfterClass
    void clear() throws Exception {
        if (dbProperties.isPostgres()) {
            autoCloseable.close();
            context.close();
        }
    }

    @AfterMethod
    void dropIndex() throws SQLException {
        if (dbProperties.isPostgres()) {
            jdbcHelper.executeUpdate(DROP_INDEX_QUERY);
        }
    }

    void shouldSkipTest() {
        if (!dbProperties.isPostgres()) {
            String message = "This test is for Postgresql only, found " + dbProperties.getDbType() + ". Skipping test.";
            logger.info(message);
            throw new SkipException(message);
        }
    }

    @Test
    void testIndexExistAndValid() throws SQLException, DbConverterExecutionException {
        shouldSkipTest();
        jdbcHelper.executeUpdate(CREATE_INDEX_QUERY);
        safeIndexConcurrentlyCreator.executeConverter(context);
        assertEquals(jdbcHelper.executeSelectCount(IS_INVALID_INDEX_QUERY), 0);
    }

    @Test
    void testIndexExistAndInvalid() throws SQLException, DbConverterExecutionException {
        shouldSkipTest();
        jdbcHelper.executeUpdate(CREATE_INDEX_QUERY);
        jdbcHelper.executeUpdate(MAKE_INDEX_INVALID_QUERY);
        safeIndexConcurrentlyCreator.executeConverter(context);
        assertEquals(jdbcHelper.executeSelectCount(IS_INVALID_INDEX_QUERY), 0);
    }

    @Test
    void testIndexNotExists() throws DbConverterExecutionException, SQLException {
        shouldSkipTest();
        safeIndexConcurrentlyCreator.executeConverter(context);
        assertEquals(jdbcHelper.executeSelectCount(IS_INVALID_INDEX_QUERY), 0);
    }

    @Test
    void testIndexNotExistsAndCreatedInvalid() throws SQLException, DbConverterExecutionException {
        shouldSkipTest();
        safeIndexConcurrentlyCreator.executeConverter(context);
        jdbcHelper.executeUpdate(MAKE_INDEX_INVALID_QUERY);
        safeIndexConcurrentlyCreator.executeConverter(context);
        assertEquals(jdbcHelper.executeSelectCount(IS_INVALID_INDEX_QUERY), 0);
    }
}