package org.artifactory.rest.resource.federation;

import org.apache.http.HttpStatus;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.descriptor.repo.RepoDescriptor;
import org.artifactory.federation.FederatedRepoStatusInfo;
import org.artifactory.federation.MirrorLagInfo;
import org.artifactory.federation.FederatedStatusService;
import org.artifactory.federation.UnavailableMirrorInfo;
import org.artifactory.rest.exception.NotFoundException;
import org.artifactory.repo.service.InternalRepositoryService;
import org.artifactory.event.migration.EventTableRepoKeyMigrationImpl;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import javax.annotation.security.RolesAllowed;
import javax.ws.rs.GET;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;


import java.util.List;

import static org.apache.http.HttpStatus.SC_OK;

@Component
@Scope(BeanDefinition.SCOPE_PROTOTYPE)
@Path("federation/status")
@RolesAllowed({AuthorizationService.ROLE_ADMIN})
public class FederatedStatusResource {

    private static final Logger log = LoggerFactory.getLogger(FederatedStatusResource.class);
    private FederatedStatusService federatedStatusService;
    private InternalRepositoryService internalRepositoryService;
    private EventTableRepoKeyMigrationImpl eventTableRepoKeyMigrationImpl;

    @Autowired
    public void setService(FederatedStatusService federatedStatusService) {
        this.federatedStatusService = federatedStatusService;
    }

    @Autowired
    public void setInternalRepositoryService(InternalRepositoryService internalRepositoryService) {
        this.internalRepositoryService = internalRepositoryService;
    }

    @Autowired
    public void setEventTableRepoKeyMigrationImpl(EventTableRepoKeyMigrationImpl eventTableRepoKeyMigrationImpl) {
        this.eventTableRepoKeyMigrationImpl = eventTableRepoKeyMigrationImpl;
    }

    @GET
    @Path("mirrorsLag")
    @Produces(MediaType.APPLICATION_JSON)
    public Response getMirrorsLag() {
        if (!eventTableRepoKeyMigrationImpl.getIfEventTableRepoKeyMigrationFinished()) {
            return Response.status(HttpStatus.SC_INTERNAL_SERVER_ERROR)
                    .entity("Monitoring is unavailable, migration in progress").build();
        }
        List<MirrorLagInfo> mirrorLagInfos = federatedStatusService.getMirrorsLag();
        return Response.status(SC_OK).entity(mirrorLagInfos).build();
    }

    @GET
    @Path("unavailableMirrors")
    @Produces(MediaType.APPLICATION_JSON)
    public Response getUnavailableMirrors() {
        if (!eventTableRepoKeyMigrationImpl.getIfEventTableRepoKeyMigrationFinished()) {
            return Response.status(HttpStatus.SC_INTERNAL_SERVER_ERROR)
                    .entity("Monitoring is unavailable, migration in progress").build();
        }
        List<UnavailableMirrorInfo> notSyncMirrors =
                federatedStatusService.getUnavailableMirrorsByTypeAndDifferentProvidedStatus();
        return Response.status(SC_OK).entity(notSyncMirrors).build();
    }

    @GET
    @Path("repo/{repoKey}")
    @Produces(MediaType.APPLICATION_JSON)
    public Response getFederatedRepoStatus(
            @PathParam("repoKey") String repoKey) {
        long currentTime = System.currentTimeMillis();
        RepoDescriptor descriptor = this.internalRepositoryService.localFederatedOrCachedRepoDescriptorByKey(repoKey);
        if (descriptor == null) {
            throw new NotFoundException(String.format("Repo key: %s not found", repoKey));
        }
        if (!eventTableRepoKeyMigrationImpl.getIfEventTableRepoKeyMigrationFinished()) {
            return Response.status(HttpStatus.SC_INTERNAL_SERVER_ERROR)
                    .entity("Monitoring is unavailable, migration in progress").build();
        }
        FederatedRepoStatusInfo federatedRepoStatusInfo =
                federatedStatusService.getFederatedRepoStatus(repoKey, currentTime);
        return Response.status(SC_OK).entity(federatedRepoStatusInfo).build();
    }

}

